// basis_fn.h

#if !defined( _basis_fn_h )
#define _basis_fn_h

#include <iostream>
#include <fstream>
#include <iomanip>
#include <cmath>
#include <cstdlib>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <gsl/gsl_sf_gamma.h>

#ifndef M_PI
#define M_PI 3.14159265358979323846264338327950288419716939937510
#endif
using namespace std;
///

enum BasisType{SPLINE_BASIS,FOURIER_BASIS};//constants to be enumerated to indicate the different basis types

class Basis_Fn

{
  public:
   ///
   Basis_Fn();
   ///
   Basis_Fn( double* v, int terms, int* cmp, int* pw, int* sgn, int* actv_dim = NULL ); // creates a terms-way interaction basis fn with knots at v, on data components cmp with powers pw and signs sn.
   ///
   Basis_Fn( double d, int cmp, int pw, int sgn, int* actv_dim = NULL  ); // creates a 1-way interaction basis fn with knot at d, on data component cmp with power pw and sign sn.
   ///
   virtual ~Basis_Fn();
   ///
   void defaults();// initialises the basis to have 1 linear term at zero on first component (0).

   void set_dimensions();

   static int get_Basis_Fn_count();// returns the current number of Basis_Fn objects.

   static int* get_Basis_Fn_count_pointer(){ return &Basis_Fn_count; }

   static void increment_count( int component );

   static void decrement_count( int component );

   static void initialize_component_Basis_Fn_counts( int num_component_dimensions );

   static void destroy_Basis_Fn_counts();

   static int get_component_Basis_Fn_count( int comp );
   
   static int get_component_Basis_Fn_count_products_up_until( int comp );
   
   static int get_component_Basis_Fn_count_products_after( int comp );
   
   double* get_basis_fn_location(){ return t; }

   int* get_basis_fn_power(){ return power; }

   int* get_basis_fn_component(){ return component; }

   int* get_basis_fn_sign(){ return sign; }

   int get_num_interactions(){ return num_interactions; }

   bool is_basis_standardized(){ return basis_standardized; }

   virtual double evaluate( double* x, int* actv_dim = NULL, int order_integral = 0, int integral_component = 0 ) = 0;//evaluate the order_integral'th integral of the basis fn in the intergral_component direction for the gsl::vector x.

   void set_mean_sd( double new_mean, double new_sd ){ mean = new_mean; sd = new_sd; basis_standardized = true; }

   void set_scale( double new_scale ){ scale = new_scale; }

   void set_coeff( double d ){ coeff_store = coeff; coeff = d;}

   void revert_coeff(){ coeff = coeff_store; }

   virtual double get_coeff(){ return coeff; }

   BasisType get_basis_type(){ return basis_type; }

 protected:

   int num_interactions;

   double* t;

   int* component;

   int* power; // if -1, taken as an Indicator function at the point t.

   int* sign; // for spline bases: -1 has basis fn directed to the left, 1 to the right and 0 both ways; for Fourier bases, 0=cosine, 1=sine.

   int* active_dimension;//almost always this will be null, but may be used so that basis function can be switched off unless its active_dimension is active.

   double scale;// used to sclae the basis function.
   
   double mean, sd; // used to normalize basis function.

   bool basis_standardized; //true if the parameters above have been set.

   double coeff, coeff_store;//the first is the coefficient of this basis function, the second is a storage variable to hold a previous value.

   BasisType basis_type;//the type of this basis function

   static int Basis_Fn_count;   // (Static) Number of basis function objects.

   static int* component_Basis_Fn_counts;   // (Static) Number of basis function objects for each component.

   static int dim_component_Basis_Fn_counts;//the dimension of component_Basis_Fn_counts.

 private:

};   

#endif

