
#if !defined( _basis_fn_regression_h )
#define _basis_fn_regression_h

#include "mcmc.h"
#include "spline_basis_fn.h"
#include "fourier_basis_fn.h"

class Basis_Fn;

class Basis_Fn_Regression : public MCMC
{
  public:
   ///
   Basis_Fn_Regression( int argc, char **argv );
   ///
   Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1 );//uses the matrix, matrix constructor of regression.h.
   ///
   Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses the matrix, matrix, matrix constructor of regression.h
   ///

   ~Basis_Fn_Regression();

   void preliminary_basis_fn_regression_initialisation();
   
   virtual void create_design_matrix();//overwriting base class version, sets a full_1way_model design matrix.

   virtual double calculate_log_likelihood() = 0;

   void add_basis_function( Basis_Fn* basis_fn_pointer, int* position = NULL );

   Basis_Fn* add_basis_function( double* v, int terms, int* cmp, int* pw, int* sgn, int* actv_dim = NULL );//adds the corresponding basis fn of basis_fn.h to the design matrix D. Returns the newly created basis fn.

   Basis_Fn* add_basis_function( double d, int cmp, int pw, int sgn, int* actv_dim= NULL );//adds a 1-way interaction basis fn to the design matrix D; Returns the newly created basis fn.
   
   void add_random_basis_function();//randomly creates a new basis fn, and adds it to the model.

   void remove_basis_function( int Dcol, bool permanently = true ); //Removes the basis function corresponding to Dcol of design matrix D. If not permanently, leaves it in Removed_Basis to be restored or deleted later.

   void destroy_removed_basis(){ zap( Removed_Basis ); }
   
   void remove_all_basis_functions();

   void remove_random_basis_function();
 
   int random_basis_fn_to_remove();//propose a basis_fn number to delete.

   int get_number_of_basis_fns(){ return Basis_Fn_number_to_pointer.size(); }
   
   int get_number_of_basis_fn_deletions_possible();//returns the number of basis fns we would be allowed to delete - can't leave D empty.

   void find_knot_point_candidates();//identify the possible spline locations.
   
   void create_full_1way_model(); // Create a design matrix D with all possible 1-way interaction, both-way splines at data points.
   Basis_Fn* get_Basis_Fn_ptr( int basis_num );//return a pointer to the basis_num'th basis function

   virtual double function_evaluate( double* v, int dim_v = 1 ); // evaluates the spline function for the array v - assumes coefficients for the basis fns have been filled in.

   virtual double function_evaluate_bases( double* v ); // evaluates the function for the array v which has as its i'th component the evaluation of the i'th basis function for some input vector - assumes coefficients for the basis fns have been filled in.
      
   virtual double integral_function_evaluate( double* v, int integral_order = 0, int integral_component = 0 );//returns the integral_order'th integral of the spline function in the intergral_component direction evaluated at v.

   double integral_absolute_function_evaluate( double* v, int integral_order = 0 );//evaluates the integral_order'th integral of the absolute of the spline funciton at v - so far only defined for positive integral_order and assuming basis fn powers of 0 or 1.

   Basis_Fn* get_first_unremovable_basis_fn( int comp );//returns the first unremovable basis fn in the (component) dimension comp.
   
   Basis_Fn* get_first_basis_fn_to_the_right( Basis_Fn* current_basis );//finds the nearest basis function to the right of current_basis, used for finding the order of the bases for e.g. absolute integration.

   Basis_Fn* get_first_basis_fn_to_the_right( double t, int comp0 );//finds the nearest basis function to the right of one which would have location t.
   
   Basis_Fn* get_first_basis_fn_to_the_left( Basis_Fn* current_basis );//finds the nearest basis function to the left of current_basis.

   Basis_Fn* get_first_basis_fn_to_the_left( double t );//finds the nearest basis function to the left of one which would have location t.

   Basis_Fn* get_leftmost_removable_basis_fn();//finds the basis function furthest to the left.

   int position_in_basis_number_to_pointer( Basis_Fn* bas );//returns the position of bas in Basis_Fn_number_to_pointer;
   
   double average_function_squared_at_bases_evaluate();// evaluates the spline function at each of the locations of the basis functions, and return the sum of squares of these values.
   
   void set_basis_coefficients( gsl_matrix* m );//sets the basis coefficients to the elements of m.

   void set_basis_coefficients( double* v );//sets the basis coefficients to the elements of v.
   
   void set_basis_coefficients( double d );//sets all of the basis coefficients to take the value d.
   
   void set_basis_coefficients();//sets the basis coefficients to the elements of beta->column(0).

   void revert_basis_coefficients();//returns the basis coefficients to their previous stored value.

   double get_absolute_sum_basis_type_coefficients( BasisType basis_type = FOURIER_BASIS, int* component = NULL, int power_upper_bound = INT_MAX );//returns the sum of the absolute coefficient values of any basis functions of type basis_fn_type.
   
   void print_basis_fns();//print some details of the basis fns to screen.
   
   void draw_random_basis_fn();//draws a random basis function from Basis_Fn_ptr and stores it in moved_basis.
   
   void draw_random_basis_coefficient_from_prior();//randomly picks a basis function, then draws it a new coefficient from the prior.
   
   void draw_basis_coefficient_from_prior( Basis_Fn* bas );//draws a new coefficient for bas from the prior, storing the old value in coeff_store.

   double calculate_basis_coefficient_log_prior( Basis_Fn* bas );//returns the log prior density for the coefficient of bas, calling (virtual) fn below.

   virtual double calculate_basis_coefficient_log_prior( double basis_coeff );//returns the log prior density for the basis coefficient, assumed to be N(0,coeff_prior_precision^-1).

   double calculate_basis_coefficients_log_prior();//calculates the log prior density of all the basis coefficients.
   
   virtual void create_design_point( double* v );// creates the design point of v by evaluating v accross the basis fns.

   void test_for_monotonicity();//see if the current function is isotonic or antitonic.

   bool room_for_new_basis_fn();//if basis_separation > 0, query if there is room for a new basis function.

   virtual void set_active_dimension( int activ_dim );//overwrites virtual function from regression class, to include change of basis types for each dimension.

   virtual double calculate_log_prior();

   double draw_continuous_basis_location( int cmp0 );//draw a legitimate, continuous-valued basis location for component cmp0, perahps checking whether there are no redundancies.

   void update_knot_point_frequencies();//increment any interval of the test_grid_points that contain a knot point.

   void write_knot_point_frequencies( ofstream *any_stream = NULL );//writes the knot_point_frequencies matrix to any_stream or by default to Knot_Point_Frequency_Filename.   
  
   
 private:

   static long power_min; // minimum order of spline basis function allowed.

   static long power_max; // maximum order of spline basis function allowed.

   static long max_fourier_basis; // maximum order of fourier basis function allowed.

   long** basis_powers;// matrix where row i contains power_min, power_max, max_fourier_basis for x-dimension i.

   bool two_sided_splines;

   long full_1way_model_sign;

   bool start_with_full_model;//if true, start with all possible basis functions in model.

   typedef map< int, Basis_Fn*, less< int > > Basis_Fn_numtoptr;

   Basis_Fn_numtoptr Basis_Fn_number_to_pointer;

   Basis_Fn** first_unremovable_basis_fns;//for each x dimension, provides a pointer to the first unremovable basis fn ever added.

  protected:

   bool discrete_knot_points;//if true, then knot points can only be chosen from a discrete range.

   bool check_data_between_bases;// if non-discrete knot points, then this option checks that there is at least one datum between each basis location.

   long knot_point_grid;// if non zero, have a grid of knot_point_grid equally spaced knot points.

   long* knot_point_grids;// a vector for holding knot_point_grid for each dimension, useful if p>1.

   int* num_knot_point_candidates;//array holding the number of candidate splines available in each dimension.

   int total_num_knot_point_candidates;

   double **knot_point_candidates;//array holding the candidate splines.

   double basis_separation;//if continuous basis fns, require they are at least this distance apart.

   unsigned int* num_1way_basis_fns;//the number of single component basis functions in each predictor dimension.

   unsigned int total_num_1way_basis_fns;//total number of single component basis functions across all dimensions.

   int** knot_point_frequencies;//counts of the number of iterations where there is a knot point between each of the test_grid_points.

   int num_knot_point_frequency_samples;//the number of times the knot point locations have been counted.

   double** distances_to_last_knot;//measures the distance to the nearest knot point to the left for each of the test_grid_points.

   double coeff_prior_precision, coeff_prior_st_dev;// prior precision and standard deviation for spline coefficients.

   double coeff_prior_precision1;// alternative prior precision for spline coefficients.

   bool storing_basis_function_coefficients;//true if model requires storage of basis function coefficients, i.e. they can't be integrated out.

   int number_unremovable_basis_functions;// the first this many basis functions can't me deleted from the spline model.

   int number_removable_basis_fns;

   long max_num_splines, max_num_splines_possible; // maximum number of splines allowed/possible.
   
   Basis_Fn *New_Basis, *Basis_to_the_left, *Basis_to_the_right;//the most newly created basis function, and those to the immediate left and right of it.

   Basis_Fn *Removed_Basis, *Moved_Basis;//pointers for holding a basis fn. marked for deletion/movement.

   int removed_basis_position, moved_basis_position;//column position in matrix of above bases.

   bool isotonic, antitonic;

   bool standardize_design_matrix;// if true, normalize each basis function to have zero mean and s.d. 1 on the observed predictors X;

   bool include_spline_bases;// if true, include spline bases in construction of design matrix.

   bool include_fourier_bases;// if true, include fourier bases in construction of design matrix.

   BasisType current_basis_type;

   double geometric_p;//parameter for geometric prior on # basis fns.

   double poisson_mu, log_poisson_mu;//rate parameter for Poisson prior on # basis fns.

   double log_geometric_p, log_1minusgeometric_p;

   double basis_proposal_width;//if proposing a new basis centred around a deleted one, use this proposal width uniformly around it.

   double** design_matrix_basis_fn_coeffs;//a matrix containing the coefficients of each basis fn making up the design matrix columns (say if we've used gram schmidt); if empty, assumed to be the identity, so design column i corresponds to weight 1 on basis fn i.

   double** basis_fn_design_matrix_coeffs;//a matrix containing the coefficientsof each design matrix column making up the basis fns; if empty, assumed to be the identity, so design column i corresponds to weight 1 on basis fn i.

   char Basis_Powers_Filename[FILENAMELENGTH];//File holding the basis function powers for each experiment.

   char Knot_Point_Frequency_Filename[FILENAMELENGTH];//File for writing the knot_point_frequencies matrix.

};

#endif
