#include "cluster_glm_mvn.h"

int Cluster_GLM_MVN::cluster_count = 0;
int Cluster_GLM_MVN::number_mergeable = 0;
int Cluster_GLM_MVN::number_splitable = 0;
bool Cluster_GLM_MVN::split_off_one_subject = 0;
GLM_MVN_Clustering* Cluster_GLM_MVN::Rptr = 0;
gsl_matrix_int* Cluster_GLM_MVN::temp_num_obsns = 0;
gsl_matrix* Cluster_GLM_MVN::temp_yty = 0;
gsl_matrix* Cluster_GLM_MVN::temp_ytd = 0;
gsl_matrix* Cluster_GLM_MVN::temp_dtd = 0;
string* Cluster_GLM_MVN::temp_label = 0;
Cluster_GLM_MVN* Cluster_GLM_MVN::root_cluster = 0;
map<string,Cluster_GLM_MVN*>* Cluster_GLM_MVN::label_to_root_cluster = 0;
bool Cluster_GLM_MVN::two_means_separation = false;
double Cluster_GLM_MVN::gibbs_sweep_log_prob = 0;
bool Cluster_GLM_MVN::using_types = false;

Cluster_GLM_MVN::Cluster_GLM_MVN( gsl_matrix_int* num_obsns_new, gsl_matrix* yty_new, gsl_matrix* ytd_new, gsl_matrix* dtd_new, bool distinct_new, bool mergeable_new, list<int>& subjects_allocated_new, double closeness_contrib_new )
{
   mergeable = mergeable_new;
   initialise();
   num_obsns = gsl_matrix_int_alloc( num_obsns_new->size1, num_obsns_new->size2 );
   for( unsigned int i = 0; i < num_obsns->size1; i++ )
      for( unsigned int j = 0; j < num_obsns->size2; j++ )
         gsl_matrix_int_set( num_obsns, i, j, gsl_matrix_int_get( num_obsns_new, i, j ) );
   yty = ytd = dtd = 0;
   yty_sum = 0;
   if( yty_new )
   {
      yty = gsl_matrix_alloc( yty_new->size1, yty_new->size2 );
      gsl_matrix_memcpy( yty, yty_new );
      yty_sum = .5 * sum_gsl_matrix( yty );
   }
   if( ytd_new )
   {
      ytd = gsl_matrix_alloc( ytd_new->size1, ytd_new->size2 );
      gsl_matrix_memcpy( ytd, ytd_new );
   }
   if( dtd_new )
   {
      dtd = gsl_matrix_alloc( dtd_new->size1, dtd_new->size2 );
      gsl_matrix_memcpy( dtd, dtd_new );
   }
   distinct = distinct_new;
   subjects_allocated = subjects_allocated_new;
   num_subjects_allocated = subjects_allocated.size();
   find_smallest_and_largest_subjects_allocated();
   subject_join = subjects_allocated.begin();
   if( num_subjects_allocated > 1 && mergeable && !distinct )
   {
      splitable = 1;
      number_splitable++;
   }
   else
      splitable = 0;
   closeness_contribution = closeness_contrib_new;
   label = 0;
   closest_cluster = 0;
   second_closest_cluster = 0;
   parent_cluster = 0;
   if( Rptr->euclidean_or_correlation_clustering )
      set_y_averages();
   if( Rptr->store_targets_in_cluster )
      set_targets();
}

Cluster_GLM_MVN::Cluster_GLM_MVN( gsl_matrix_int* num_obsns_new, gsl_matrix* yty_new, gsl_matrix* ytd_new, gsl_matrix* dtd_new, bool distinct_new, bool mergeable_new, list<int>& subjects_allocated_new, double closeness_contrib_new, string label_new )
{
   mergeable = mergeable_new;
   initialise();
   num_obsns = gsl_matrix_int_alloc( num_obsns_new->size1, num_obsns_new->size2 );
   for( unsigned int i = 0; i < num_obsns->size1; i++ )
      for( unsigned int j = 0; j < num_obsns->size2; j++ )
         gsl_matrix_int_set( num_obsns, i, j, gsl_matrix_int_get( num_obsns_new, i, j ) );
   yty = ytd = dtd = 0;
   yty_sum = 0;
   if( yty_new )
   {
      yty = gsl_matrix_alloc( yty_new->size1, yty_new->size2 );
      gsl_matrix_memcpy( yty, yty_new );
      yty_sum = .5 * sum_gsl_matrix( yty );
   }
   if( ytd_new )
   {
      ytd = gsl_matrix_alloc( ytd_new->size1, ytd_new->size2 );
      gsl_matrix_memcpy( ytd, ytd_new );
   }
   if( dtd_new )
   {
      dtd = gsl_matrix_alloc( dtd_new->size1, dtd_new->size2 );
      gsl_matrix_memcpy( dtd, dtd_new );
   }
   distinct = distinct_new;
   subjects_allocated = subjects_allocated_new;
   num_subjects_allocated = subjects_allocated.size();
   find_smallest_and_largest_subjects_allocated();
   subject_join = subjects_allocated.begin();
   if( num_subjects_allocated > 1  && mergeable )
   {
      splitable = 1;
      number_splitable++;
   }
   else
      splitable = 0;
   closeness_contribution = closeness_contrib_new;
   label = new string( label_new );
   closest_cluster = 0;
   second_closest_cluster = 0;
   parent_cluster = 0;
   if( Rptr->euclidean_or_correlation_clustering )
      set_y_averages();
   if( Rptr->store_targets_in_cluster )
      set_targets();
}

Cluster_GLM_MVN::Cluster_GLM_MVN( list<Cluster_GLM_MVN*> clusters_allocated_new )
{
   clusters_allocated = clusters_allocated_new;
   mergeable = 1;//??
   initialise();
   num_obsns = 0;
   yty = 0;
   ytd = 0;
   dtd = 0;
   yty_sum = 0;
   distinct = 0;
   splitable = 1;
   label = 0;
   closest_cluster = 0;
   second_closest_cluster = 0;
   parent_cluster = 0;
   subject_join = subjects_allocated.begin();
   Cluster_GLM_MVN* cl;
   num_subjects_allocated = 0;
   smallest_subject = INT_MAX;
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      cl = *iter;
      /*
      if( cl->label )
      {
         if( !label )
            label = new string( *(cl->label) );
         else
            if( *label != *(cl->label) )
               *label +=  "; " + *(cl->label);
      }
      */
      num_subjects_allocated += cl->num_subjects_allocated;
      if( cl->smallest_subject < smallest_subject )
         smallest_subject = cl->smallest_subject;
      cl->set_parent_cluster( this );
      if( !distinct && cl->distinct )
         distinct = 1;
      yty_sum += cl->get_yty_sum();
   }
   closeness_contribution = Rptr->calculate_closeness_contribution( this );
}


Cluster_GLM_MVN::Cluster_GLM_MVN( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2, bool temporary )
{
   mergeable = 1;
   initialise();
   num_obsns = 0;
   yty = 0;
   ytd = 0;
   dtd = 0;
   if( cl1->distinct || cl2->distinct )
      distinct = 1;
   else
      distinct = 0;
   splitable = 1;
   label = 0;
   closest_cluster = 0;
   second_closest_cluster = 0;
   parent_cluster = 0;
   num_subjects_allocated = cl1->num_subjects_allocated + cl2->num_subjects_allocated;
   v_index = Rptr->optimal_v_index;
   
   if( temporary )
   {
      clusters_allocated.push_back( cl1 );
      clusters_allocated.push_back( cl2 );
      return;
   }
   
   Cluster_GLM_MVN* cl1s_old_parent_cluster = cl1->parent_cluster;
   cl1->set_parent_cluster( this );
   cl2->set_parent_cluster( this );
   set_parent_cluster( cl1s_old_parent_cluster );//new cluster takes the parent of first child if it had one.

   unsigned int i, j;
   if( Rptr->store_data_throughout_tree && cl1->num_obsns && cl2->num_obsns )
   {
      num_obsns = gsl_matrix_int_alloc( cl1->num_obsns->size1, cl1->num_obsns->size2 );
      for( i = 0; i < num_obsns->size1; i++ )
         for( j = 0; j < num_obsns->size2; j++ )
         gsl_matrix_int_set( num_obsns, i, j, gsl_matrix_int_get( cl1->num_obsns, i, j ) );
      gsl_matrix_int_add( num_obsns, cl2->num_obsns );
      yty = gsl_matrix_alloc( cl1->yty->size1, cl1->yty->size2 );
      gsl_matrix_memcpy( yty, cl1->yty );
      gsl_matrix_add( yty, cl2->yty );
      ytd = gsl_matrix_alloc( cl1->ytd->size1, cl1->ytd->size2 );
      gsl_matrix_memcpy( ytd, cl1->ytd );
      gsl_matrix_add( ytd, cl2->ytd );
      if( !Rptr->inputs_repeated_every )
      {
         dtd = gsl_matrix_alloc( cl1->dtd->size1, cl1->dtd->size2 );
         gsl_matrix_memcpy( dtd, cl1->dtd );
         gsl_matrix_add( dtd, cl2->dtd );
      }
      if( cl1->y_averages && cl2->y_averages )
      {
         y_averages = gsl_matrix_alloc( cl1->y_averages->size1, cl1->y_averages->size2 );
         for( j = 0; j < cl1->y_averages->size2; j++ ) 
            for( i = 0; i < cl1->y_averages->size1; i++ )
               gsl_matrix_set( y_averages, i, j, ( gsl_matrix_get( cl1->y_averages, i, j ) * cl1->num_subjects_allocated + gsl_matrix_get( cl2->y_averages, i, j ) * cl2->num_subjects_allocated ) / (double)( cl1->num_subjects_allocated + cl2->num_subjects_allocated) );
      }
      if( cl1->ys && cl2->ys )
         ys = matrix_join( cl1->ys, cl2->ys, false, true );
   }
   
   if( root_cluster == cl1 || root_cluster == cl2 ) //if one of the clusters allocated was the root cluster, take over that position.
      set_root_cluster( this );
   if( label_to_root_cluster && Rptr && Rptr->cluster_by_labels && get_label() )
   {
      map< string, Cluster_GLM_MVN* >::iterator iter = label_to_root_cluster->find( *temp_label );
      if( iter != label_to_root_cluster->end() && ( iter->second == cl1 || iter->second == cl2 ) )
         set_root_cluster( *temp_label, this );
   }

   subject_join = subjects_allocated.begin();
   if( Rptr->heteroskedastic )
   {
      closeness_contribution = cl1->closeness_contribution + cl2->closeness_contribution + cl1->get_cluster_closeness( cl2 );
      if( !Rptr->euclidean_or_correlation_clustering )
         closeness_contribution -= Rptr->get_change_log_prior_cluster_size( cl1->num_subjects_allocated, cl2->num_subjects_allocated, false );
   }
   else
      closeness_contribution = cl1->closeness_contribution + cl2->closeness_contribution + Rptr->proposed_ytd_vstar_dty_sum - Rptr->current_ytd_vstar_dty_sum;
   cl1->closeness_map.clear();
   cl2->closeness_map.clear();//clear out their closeness maps to save space.
   if( cl1->smallest_subject < cl2->smallest_subject )
   {
      smallest_subject = cl1->smallest_subject;
      if( Rptr->closeness_matrix && Rptr->reduce_closeness_matrix )
         zap_vector( Rptr->closeness_matrix[ cl2->smallest_subject ] );
   }
   else
   {
      smallest_subject = cl2->smallest_subject;
      if( Rptr->closeness_matrix && Rptr->reduce_closeness_matrix )
         zap_vector( Rptr->closeness_matrix[ cl1->smallest_subject ] );
   }
   if( Rptr->order_cluster_subjects )
      order_clusters_allocated();
   if( Rptr->keep_vstars == 2 )
   {
      if( !Rptr->correlate_experiments )
      {
         gsl_matrix_int* cl_n = cl1->get_num_obsns();
         for( j = 0; j < cl_n->size2; j++ )
            Rptr->delete_vstar( gsl_matrix_int_get( cl_n, 0, j ) );
         cl_n = cl2->get_num_obsns();
         for( j = 0; j < cl_n->size2; j++ )
            Rptr->delete_vstar( gsl_matrix_int_get( cl_n, 0, j ) );
      }
      else
      {
         Rptr->delete_vstar( cl1->get_total_num_obsns() );
         Rptr->delete_vstar( cl2->get_total_num_obsns() );
      }
   }
 }

void Cluster_GLM_MVN::initialise()
{
   cluster_count++;
   if( mergeable )
   {
      number_mergeable++;
      if( Rptr->preserve_tree )
         Rptr->all_clusters_in_tree.push_back( this );
   }
   old_closeness_contribution = 0;
   heller_prior_d = 0;
   score = 1;
   last_iteration_updated = 1;
   v_index = 0;
   y_averages = 0;
   ys = 0;
   cluster_error_covariances = 0;
   cluster_error_sqrt_inverse_covariances = 0;
   type = 0;
}

Cluster_GLM_MVN::~Cluster_GLM_MVN()
{
   cluster_count--;
   if( num_obsns )
      gsl_matrix_int_free( num_obsns );
   if( yty )
      gsl_matrix_free( yty );
   if( ytd )
      gsl_matrix_free( ytd );
   if( dtd )
      gsl_matrix_free( dtd );
   if( y_averages )
      gsl_matrix_free( y_averages );
   if( ys )
      gsl_matrix_free( ys );
   if( cluster_error_covariances )
      gsl_matrix_free( cluster_error_covariances );
   if( cluster_error_sqrt_inverse_covariances )
      gsl_matrix_free( cluster_error_sqrt_inverse_covariances );
   zap( label );
   zap( old_closeness_contribution );
   zap( heller_prior_d );
   if( parent_cluster )
      parent_cluster->clusters_allocated.remove( this );
   if( mergeable && Rptr && Rptr->preserve_tree )
      Rptr->remove_from_all_clusters_deque( this );
}

void Cluster_GLM_MVN::kill_static_variables()
{
   if( temp_num_obsns )
      gsl_matrix_int_free( temp_num_obsns );
   if( temp_yty )
      gsl_matrix_free( temp_yty );
   if( temp_ytd )
      gsl_matrix_free( temp_ytd );
   if( temp_dtd )
      gsl_matrix_free( temp_dtd );
   zap( temp_label );
}

ostream& operator<< ( ostream& os, Cluster_GLM_MVN* & cl )
{
   os.setf( ios::fixed );
   os << (void*)cl;
   os << " (" << cl->get_num_subjects_allocated() << " subjects)";
   if( Cluster_GLM_MVN::using_types )
      os << " [Type=" << cl->type << "]";
   os << endl;
   if( cl )
   {
      list< int > temp_subjects_allocated = cl->get_subjects_allocated();
      for( list<int>::iterator iter = temp_subjects_allocated.begin(); iter != temp_subjects_allocated.end(); iter++ )
         os << *iter << endl;
//      os << cl->get_targets();
   }
   return os;
}

void Cluster_GLM_MVN::order_clusters_allocated()
{
   if( clusters_allocated.size() != 2 )
      return;
   list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
   Cluster_GLM_MVN* cl1 = *iter;
   iter++;
   Cluster_GLM_MVN* cl2 = *iter;
   if( cl1->clusters_allocated.size() != 2 && cl2->clusters_allocated.size() != 2 )
      return;

   Cluster_GLM_MVN* cl11 = 0;
   Cluster_GLM_MVN* cl12 = cl1;
   Cluster_GLM_MVN* cl21 = cl2;
   Cluster_GLM_MVN* cl22 = 0;
   if( cl1->clusters_allocated.size() == 2 )
   {
      iter = cl1->clusters_allocated.begin();
      cl11 = *iter;
      cl12 = *(++iter);
   }
   if( cl2->clusters_allocated.size() == 2 )
   {
      iter = cl2->clusters_allocated.begin();
      cl21 = *iter;
      cl22 = *(++iter);
   }
   int clsnss_argmax = find_best_cluster_order( cl11, cl12, cl21, cl22, cl1, cl2 );
   switch( clsnss_argmax )
   {
      case 1222://switch cl22 to the front of cl2's children
//         cl2->clusters_allocated.remove( cl22 );
//         cl2->clusters_allocated.push_front( cl22 );
         cl2->reverse_clusters_allocated();
         break;
      case 1121://switch cl11 to the back of cl1's children
//         cl1->clusters_allocated.remove( cl11 );
//         cl1->clusters_allocated.push_back( cl11 );
         cl1->reverse_clusters_allocated();
         break;
      case 1122://switch cl22 to the front of cl2's children
         //and switch cl12 to the back of cl1's children
//         cl2->clusters_allocated.remove( cl22 );
//         cl2->clusters_allocated.push_front( cl22 );
//         cl1->clusters_allocated.remove( cl11 );
//         cl1->clusters_allocated.push_back( cl11 );
         clusters_allocated.reverse();
         break;
   }
}

void Cluster_GLM_MVN::order_clusters_allocated( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 )
{
   if( cl1->subject_join == cl1->subjects_allocated.begin() && cl2->subject_join == cl2->subjects_allocated.begin() )
      return;
   int number_of_models_evaluated_store = Rptr->number_of_models_evaluated;
   Cluster_GLM_MVN* cl11 = 0;
   Cluster_GLM_MVN* cl12 = cl1;
   Cluster_GLM_MVN* cl21 = cl2;
   Cluster_GLM_MVN* cl22 = 0;
   if( cl1->subject_join != cl1->subjects_allocated.begin() )
   {
      cl11 = Rptr->create_temporary_cluster_for_subjects( cl1->subjects_allocated, cl1->subjects_allocated.begin(), cl1->subject_join );
      cl12 = Rptr->create_temporary_cluster_for_subjects( cl1->subjects_allocated, cl1->subject_join, cl1->subjects_allocated.end() );
   }
   if( cl2->subject_join != cl2->subjects_allocated.begin() )
   {
      cl21 = Rptr->create_temporary_cluster_for_subjects( cl2->subjects_allocated, cl2->subjects_allocated.begin(), cl2->subject_join );
      cl22 = Rptr->create_temporary_cluster_for_subjects( cl2->subjects_allocated, cl2->subject_join, cl2->subjects_allocated.end() );
   }
   int clsnss_argmax = find_best_cluster_order( cl11, cl12, cl21, cl22, cl1, cl2 );
   if( cl11 )
   {
      delete cl11;
      delete cl12;
   }
   if( cl22 )
   {
      delete cl21;
      delete cl22;
   }
   switch( clsnss_argmax )
   {
      case 1222://switch cl22 to the front of cl2's children
//         cl2->switch_subjects_about_join();
         cl2->subjects_allocated.reverse();
         break;
      case 1121://switch cl11 to the back of cl1's children
//         cl1->switch_subjects_about_join();
         cl1->subjects_allocated.reverse();
         break;
      case 1122://switch cl22 to the front of cl2's children
         //and switch cl12 to the back of cl1's children
//         cl1->switch_subjects_about_join();
//         cl2->switch_subjects_about_join();
         cl1->subjects_allocated.reverse();
         cl2->subjects_allocated.reverse();
         break;
   }
   Rptr->number_of_models_evaluated = number_of_models_evaluated_store;
}

int Cluster_GLM_MVN::find_best_cluster_order( Cluster_GLM_MVN* cl11, Cluster_GLM_MVN* cl12, Cluster_GLM_MVN* cl21, Cluster_GLM_MVN* cl22, Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 )
{
   bool new_ordering_method = 1;
   int clsnss_argmax = 1221;
   if( !new_ordering_method )
   {
      double clsnss1221 = Rptr->calculate_corrected_closeness( cl12, cl21 );
      double max_clsnss = clsnss1221;
      // first compare cl12 with cl2's children clusters
      if( cl22 )
      {
         double clsnss1222 = Rptr->calculate_corrected_closeness( cl12, cl22 );
         if( clsnss1222 > max_clsnss )
         {
            max_clsnss = clsnss1222;
            clsnss_argmax = 1222;
         }
      }
      if( cl11 )
      {
         double clsnss1121 = Rptr->calculate_corrected_closeness( cl11, cl21 );
         if( clsnss1121 > max_clsnss )
         {
            max_clsnss = clsnss1121;
            clsnss_argmax = 1121;
         }
         if( cl22 )
         {
            double clsnss1122 = Rptr->calculate_corrected_closeness( cl11, cl22 );
            if( clsnss1122 > max_clsnss )
            {
               max_clsnss = clsnss1122;
               clsnss_argmax = 1122;
            }
         }
      }
   }
   else
   {
      if( cl11 && Rptr->calculate_corrected_closeness( cl11, cl2 ) > Rptr->calculate_corrected_closeness( cl12, cl2 ) )
         clsnss_argmax = 1121;
      if( cl22 && Rptr->calculate_corrected_closeness( cl1, cl22 ) > Rptr->calculate_corrected_closeness( cl1, cl21 ) )
      {
         if( clsnss_argmax == 1121 )
            clsnss_argmax = 1122;
         else
            clsnss_argmax = 1222;
      }
   }
   return clsnss_argmax;
}

void Cluster_GLM_MVN::reverse_clusters_allocated()
{
   if( clusters_allocated.empty() )
      return;
   clusters_allocated.reverse();
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
      (*iter)->reverse_clusters_allocated();
}

void Cluster_GLM_MVN::switch_subjects_about_join()
{
   list<int> cl_subs;
   for( list<int>::iterator iter = subjects_allocated.begin(); iter != subject_join; iter++ )
      cl_subs.push_back( *iter );
   subjects_allocated.erase( subjects_allocated.begin(), subject_join );
   subject_join = subjects_allocated.end();
   subjects_allocated.splice( subject_join, cl_subs );
}

string* Cluster_GLM_MVN::get_label()
{
   if( label || clusters_allocated.empty() )
   {
      if( temp_label && label )
         *temp_label = *label;
      return label;
   }
   string* st_ptr = 0;
   string* st_ptr2 = 0;
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      st_ptr2 = (*iter)->get_label();
      if( st_ptr2 )
      {
         if( !st_ptr )
            st_ptr = new string( *st_ptr2 );
         else
         {
            if( *st_ptr != *st_ptr2 )
               *st_ptr +=  "; " + *(st_ptr2);
         }
      }
   }
   if( st_ptr )
   {
      if( !temp_label )
         temp_label = new string;
      *temp_label = *st_ptr;
      delete st_ptr;
   }
   else
      zap( temp_label );
   return temp_label;
}

void Cluster_GLM_MVN::set_label( string new_label )
{
   if( !label )
      label = new string;
   *label = new_label;
}

void Cluster_GLM_MVN::set_y_averages( gsl_matrix* new_y_averages )
{
   if( y_averages )
      gsl_matrix_free( y_averages );
   if( new_y_averages )
      y_averages = new_y_averages;
   y_averages = Rptr->get_target_averages( subjects_allocated );
}

gsl_matrix* Cluster_GLM_MVN::get_y_averages()
{
   if( !y_averages )
      y_averages = Rptr->get_target_averages( subjects_allocated );
   return y_averages;
}

void Cluster_GLM_MVN::set_targets()
{
   if( ys )
   {
      gsl_matrix_free( ys );
      ys = 0;
   }
   ys = get_targets();
}

gsl_matrix* Cluster_GLM_MVN::get_targets()
{
   if( ys )
      return ys;
   list<int> temp_subjects_allocated = get_subjects_allocated();
   if( !Rptr->store_targets_in_cluster )
      return Rptr->get_targets( temp_subjects_allocated );
   ys = Rptr->get_targets( temp_subjects_allocated );
   return ys;
}

gsl_matrix_int* Cluster_GLM_MVN::get_num_obsns( bool recursive_call )
{
   if( !Rptr->preserve_tree || !clusters_allocated.size() )
      return num_obsns;
   if( num_obsns && ( Rptr->mcmc_sampling || Rptr->finished_mcmc_sampling ) )
   {
      gsl_matrix_int_free( num_obsns );//safest to recalculate when mcmc sampling with tree.
      num_obsns = 0;
   }
   if( num_obsns )
      return num_obsns;
   if( !recursive_call )
   {
      if( !temp_num_obsns )
         temp_num_obsns = gsl_matrix_int_calloc( 1, Rptr->num_experiments );
      else
         gsl_matrix_int_set_zero( temp_num_obsns );
   }
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      gsl_matrix_int* iter_num_obsns = (*iter)->get_num_obsns( true );
      if( iter_num_obsns != temp_num_obsns )
         gsl_matrix_int_add( temp_num_obsns, iter_num_obsns );
   }
   return temp_num_obsns;
}

int Cluster_GLM_MVN::get_total_num_obsns()
{
   if( Rptr->preserve_tree && clusters_allocated.size() && ( Rptr->mcmc_sampling || Rptr->finished_mcmc_sampling ) )
      get_num_obsns();
   int sum = 0;
   if( num_obsns )
      for( unsigned int i = 0; i < num_obsns->size1; i++ )
         for( unsigned int j = 0; j < num_obsns->size2; j++ )
            sum += gsl_matrix_int_get( num_obsns, i, j );
   else
      for( unsigned int i = 0; i < temp_num_obsns->size1; i++ )
         for( unsigned int j = 0; j < temp_num_obsns->size2; j++ )
            sum += gsl_matrix_int_get( temp_num_obsns, i, j );
   return sum;
}

int Cluster_GLM_MVN::get_num_experiments_represented()
{
   if( num_obsns )
      return count_non_zero_entries( num_obsns );
   get_num_obsns();
   return count_non_zero_entries( temp_num_obsns );
}

gsl_matrix* Cluster_GLM_MVN::get_yty( bool recursive_call )
{
   if( !Rptr->preserve_tree )
      return yty;
   if( yty && Rptr->preserve_tree && clusters_allocated.size() && ( Rptr->mcmc_sampling || Rptr->finished_mcmc_sampling ) )
   {
      gsl_matrix_free( yty );//safest to recalculate when mcmc sampling with tree.
      yty = 0;
   }
   if( yty )
      return yty;
   if( !recursive_call )
   {
      if( !temp_yty )
         temp_yty = gsl_matrix_calloc( Rptr->p_y, Rptr->yty_cols );
      else
         gsl_matrix_set_zero( temp_yty );
   }
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      gsl_matrix* iter_yty = (*iter)->get_yty( true );
      if( iter_yty != temp_yty )
         gsl_matrix_add( temp_yty, iter_yty );
   }
   return temp_yty;
}

gsl_matrix* Cluster_GLM_MVN::get_ytd( bool recursive_call )
{
   if( !Rptr->preserve_tree )
      return ytd;
   if( ytd && Rptr->preserve_tree && clusters_allocated.size() && ( Rptr->mcmc_sampling || Rptr->finished_mcmc_sampling ) )
   {
      gsl_matrix_free( ytd );//safest to recalculate when mcmc sampling with tree.
      ytd = 0;
   }
   if( ytd )
      return ytd;
   if( !recursive_call )
   {
      if( !temp_ytd )
         temp_ytd = gsl_matrix_calloc( Rptr->p_y, Rptr->ytd_cols );
      else
         gsl_matrix_set_zero( temp_ytd );
   }
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      gsl_matrix* iter_ytd = (*iter)->get_ytd( true );
      if( iter_ytd != temp_ytd )
         gsl_matrix_add( temp_ytd, iter_ytd );
   }
   return temp_ytd;
}

gsl_matrix* Cluster_GLM_MVN::get_dtd( bool recursive_call )
{
   if( !Rptr->preserve_tree )
      return dtd;
   if( dtd && Rptr->preserve_tree && clusters_allocated.size() && ( Rptr->mcmc_sampling || Rptr->finished_mcmc_sampling ) )
   {
      gsl_matrix_free( dtd );//safest to recalculate when mcmc sampling with tree.
      dtd = 0;
   }
   if( dtd )
      return dtd;
   if( !recursive_call )
   {
      if( !temp_dtd )
         temp_dtd = gsl_matrix_calloc( Rptr->num_design_cols, Rptr->dtd_cols );
      else
         gsl_matrix_set_zero( temp_dtd );
   }
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      gsl_matrix* iter_dtd = (*iter)->get_dtd( true );
      if( iter_dtd != temp_dtd )
         gsl_matrix_add( temp_dtd, iter_dtd );
   }
   return temp_dtd;
}

double Cluster_GLM_MVN::get_number_of_possible_clusterings()
{
   if( clusters_allocated.empty() )
      return 1;
   double d = 1;
   list<Cluster_GLM_MVN*>::iterator iter2;
   iter2 = clusters_allocated.begin();
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); ++iter )
      d *= (*iter)->get_number_of_possible_clusterings();
   return d + 1;
}

void Cluster_GLM_MVN::get_random_size_n_cluster_clustering( int size_n, map<Cluster_GLM_MVN*,double>& map_of_clusters )
{
   if( size_n == 1 )
   {
      map_of_clusters[ this ] = 0;
      return;
   }
   else if( get_num_subjects_allocated() < size_n )
   {
      cout << "Not enough subjects to create " << size_n << " clusters" << endl;
      exit(1);
   }
   deque<Cluster_GLM_MVN*> deque_of_clusters;
   Cluster_GLM_MVN* cl;
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
      deque_of_clusters.push_back(*iter);
   for( int i = 0; i < size_n - 2; i++ )
   {
      random_shuffle( deque_of_clusters.begin(), deque_of_clusters.end() );
      cl = *deque_of_clusters.begin();
      deque_of_clusters.pop_front();
      while( !cl->splitable )
      {
         deque_of_clusters.push_back( cl );
         cl = *deque_of_clusters.begin();
      }
      for( list<Cluster_GLM_MVN*>::iterator iter = cl->clusters_allocated.begin(); iter != cl->clusters_allocated.end(); iter++ )
         deque_of_clusters.push_back(*iter);
   }
   for( deque<Cluster_GLM_MVN*>::iterator deque_iter = deque_of_clusters.begin(); deque_iter != deque_of_clusters.end(); deque_iter++ )
      map_of_clusters[ *deque_iter ] = 0;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::get_sibling_cluster() const
{
   if( !parent_cluster )
      return NULL;
   list<Cluster_GLM_MVN*>::iterator iter = parent_cluster->clusters_allocated.begin();
   if( *iter == this )
      iter++;
   if( iter != parent_cluster->clusters_allocated.end() )
      return *iter;
   return NULL;
}

double Cluster_GLM_MVN::get_sibling_closeness( Cluster_GLM_MVN* cl_sibling )
{
   if( !cl_sibling )
      cl_sibling = get_sibling_cluster();
   if( !cl_sibling  )
      return -DBL_MAX;
   return get_cluster_closeness( cl_sibling );
//   return sibling_closeness;
}

double Cluster_GLM_MVN::get_sibling_merge_prob( Cluster_GLM_MVN* cl_sibling )
{
   double d = get_sibling_closeness( cl_sibling );
   if( cl_sibling && Rptr )
   {
      int cl1s = num_subjects_allocated;
      int cl2s = cl_sibling->num_subjects_allocated;
      d += (Rptr->log_prior_cluster_size)[ cl1s + cl2s ] - (Rptr->log_prior_cluster_size)[ cl1s ] - (Rptr->log_prior_cluster_size)[ cl2s ];
      if( Rptr->conditional_tree_distribution )
         d += Rptr->get_uniform_tree_log_prior_change( cl1s, cl2s );
   }
   return d;
}

double Cluster_GLM_MVN::get_childrens_closeness()
{
   if( clusters_allocated.size() != 2 )
      return DBL_MAX;
   list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
   Cluster_GLM_MVN* cl1 = *iter;
   Cluster_GLM_MVN* cl2 = *(++iter);
   return cl1->get_cluster_closeness( cl2 );
}

double Cluster_GLM_MVN::get_childrens_merge_prob()
{
   if( clusters_allocated.size() != 2 )
      return DBL_MAX;
   list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
   Cluster_GLM_MVN* cl1 = *iter;
   Cluster_GLM_MVN* cl2 = *(++iter);
   return cl1->get_sibling_merge_prob( cl2 );
}

list<int> Cluster_GLM_MVN::get_subjects_allocated()
{
   if( clusters_allocated.empty() )
      return subjects_allocated;
   list< int > temp_subjects_allocated = subjects_allocated;
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      list< int > temp_subjects_allocated2 = (*iter)->get_subjects_allocated();
      temp_subjects_allocated.splice( temp_subjects_allocated.end(), temp_subjects_allocated2 );
   }
   return temp_subjects_allocated;
}

int Cluster_GLM_MVN::get_num_subjects_allocated()
{
   if( clusters_allocated.empty() )
      return num_subjects_allocated;
   int temp_num_subjects_allocated = 0;
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
      temp_num_subjects_allocated += (*iter)->get_num_subjects_allocated();
   return temp_num_subjects_allocated;
}

void Cluster_GLM_MVN::find_smallest_and_largest_subjects_allocated()
{
   smallest_subject = INT_MAX;
   largest_subject = -INT_MAX;
   if( clusters_allocated.empty() )
   {
      for( list<int>::iterator iter = subjects_allocated.begin(); iter != subjects_allocated.end(); iter++ )
      {
         if( (*iter) < smallest_subject )
            smallest_subject = *iter;
         if( (*iter) > largest_subject )
            largest_subject = *iter;
      }
   }
   else
   {
      list<int> temp_subjects_allocated = get_subjects_allocated();
      for( list<int>::iterator iter = temp_subjects_allocated.begin(); iter != temp_subjects_allocated.end(); iter++ )
      {
         if( (*iter) < smallest_subject )
            smallest_subject = *iter;
         if( (*iter) > largest_subject )
            largest_subject = *iter;
      }
   }
}

int Cluster_GLM_MVN::get_first_subject_allocated()
{
   list<int> temp_subjects_allocated = get_subjects_allocated();
   return *(temp_subjects_allocated.begin());
}

int Cluster_GLM_MVN::get_last_subject_allocated()
{
   list<int> temp_subjects_allocated = get_subjects_allocated();
   return *(temp_subjects_allocated.rbegin());
}

void Cluster_GLM_MVN::get_first_and_last_subjects_allocated( int& first_subject, int& last_subject )
{
   list<int> temp_subjects_allocated = get_subjects_allocated();
   first_subject = *(temp_subjects_allocated.begin());
   last_subject = *(temp_subjects_allocated.rbegin());
}

list<pair <int,int> > Cluster_GLM_MVN::get_merger_details()
{
   list<pair <int,int> > merge_details;
   if( !clusters_allocated.empty() )
   {
      list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
      int sub1 = (*iter)->get_smallest_subject_allocated();
      int sub2 = (*++iter)->get_smallest_subject_allocated();
      if( sub1 < sub2 )
         merge_details.push_front( pair<int,int>(sub1, sub2) );
      else
         merge_details.push_front( pair<int,int>(sub2, sub1) );
      for( iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
      {
         list<pair <int,int> > temp_merge_details = (*iter)->get_merger_details();
         merge_details.splice( merge_details.end(), temp_merge_details );
      }
   }
   return merge_details;
}

double Cluster_GLM_MVN::get_closeness_contribution()
{
   if( clusters_allocated.size() && Rptr && Rptr->preserve_tree && ( Rptr->mcmc_sampling || Rptr->finished_mcmc_sampling ) )
      closeness_contribution = Rptr->calculate_closeness_contribution( this );
   return closeness_contribution;
}

void Cluster_GLM_MVN::set_closeness_contribution( double closeness_contrib_new )
{
   if( !old_closeness_contribution && Rptr && Rptr->mcmc_sampling && Rptr->variable_v )
      old_closeness_contribution = new double;
   if( old_closeness_contribution )
      *old_closeness_contribution = closeness_contribution;
   if( !heller_prior_d && Rptr && Rptr->heller_algorithm )
      heller_prior_d = new double;
   if( heller_prior_d )
      *heller_prior_d = Rptr->heller_d;
   closeness_contribution = closeness_contrib_new;
}

bool Cluster_GLM_MVN::cluster_contains_subject( int subject )
{
   if( clusters_allocated.empty() )
   {
      if( find( subjects_allocated.begin(), subjects_allocated.end(), subject ) != subjects_allocated.end() )
         return 1;
      return 0;
   }
   list< int > temp_subjects_allocated = get_subjects_allocated();
   if( find( temp_subjects_allocated.begin(), temp_subjects_allocated.end(), subject ) != temp_subjects_allocated.end() )
      return 1;
   return 0;
}

bool Cluster_GLM_MVN::cluster_has_subjects( list<int>& cl_subjects )
{
   list<int> temp_subjects_allocated = get_subjects_allocated();
   if( temp_subjects_allocated.size() != cl_subjects.size() )
      return false;
   list<int>::iterator iter, iter2;
   for( iter = cl_subjects.begin(); iter != cl_subjects.end(); iter++ )
   {
      iter2 = find( temp_subjects_allocated.begin(), temp_subjects_allocated.end(), *iter );
      if( iter2 ==  temp_subjects_allocated.end() )
         return false;
      else
         temp_subjects_allocated.erase( iter2 );
   }
   return true;
}

bool Cluster_GLM_MVN::cluster_has_child_with_subjects( list<int>& cl_subjects )
{
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
      if( (*iter)->cluster_has_subjects( cl_subjects ) )
         return true;
   return false;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::find_sub_cluster_containing_subjects( list<int> splitting_subjects )
{
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      bool all_in = 1;
      list<int>::iterator iter2 = splitting_subjects.begin();
      while( all_in && iter2 != splitting_subjects.end() )
      {
         if( !(*iter)->cluster_contains_subject( *iter2 ) )
            all_in = 0;
         iter2++;
      }
      if( all_in )
         return *iter;
   }
   return NULL;
}

void Cluster_GLM_MVN::print_subjects_allocated()
{
   if( clusters_allocated.empty() )
   {
      for( list<int>::iterator iter = subjects_allocated.begin(); iter != subjects_allocated.end(); iter++ )
         cout << *iter << endl;
   }
   else
   {
      list< int > temp_subjects_allocated = get_subjects_allocated();
      for( list<int>::iterator iter = temp_subjects_allocated.begin(); iter != temp_subjects_allocated.end(); iter++ )
         cout << *iter << endl;
   }
}

void Cluster_GLM_MVN::get_clusters_allocated( Cluster_GLM_MVN*& cl1, Cluster_GLM_MVN*& cl2 )
{
   cl1 = cl2 = NULL;
   if( !clusters_allocated.size() )
      return;
   list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
   cl1 = *(iter++);
   if( iter == clusters_allocated.end() )
      return;
   cl2 = *(iter);
}

map<Cluster_GLM_MVN*, double> Cluster_GLM_MVN::get_single_subject_clusters_allocated( bool deleting_tree )
{
   map<Cluster_GLM_MVN*, double> map_of_clusters;
   if( Rptr->preserve_tree )
   {
      if( !clusters_allocated.empty() )
         get_single_subject_clusters_allocated( map_of_clusters, deleting_tree );
      else if( mergeable )
         map_of_clusters[ this ] = -DBL_MAX;
   }
   else
   {
      for( list<int>::iterator iter = subjects_allocated.begin(); iter != subjects_allocated.end(); iter++ )
         map_of_clusters[ Rptr->make_cluster_containing_subject( *iter ) ] = -DBL_MAX;
   }
   return map_of_clusters;
}

void Cluster_GLM_MVN::get_single_subject_clusters_allocated( map<Cluster_GLM_MVN*, double>& map_of_clusters, bool deleting_tree )
{
   int num_clusters_allocated = clusters_allocated.size();
   list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
   for( int i = 0; i < num_clusters_allocated; i++ )
   {
      if( !(*iter)->clusters_allocated.empty() && ( !Rptr || !Rptr->is_cluster_in_model( *iter ) ) )
      {
//         if( deleting_tree )
//            for( list<Cluster_GLM_MVN*>::iterator iter2 = (*iter)->clusters_allocated.begin(); iter2 != (*iter)->clusters_allocated.end(); iter2++ )
//            {
//               (*iter2)->parent_cluster = NULL;
//            }
         (*iter)->get_single_subject_clusters_allocated( map_of_clusters, deleting_tree );
	 if( deleting_tree )
         {
	    delete *iter;
            iter = clusters_allocated.begin();
         }
         else
            ++iter;
      }
      else
      {
         if( (*iter)->mergeable )
            map_of_clusters[ *iter ] = -DBL_MAX;
         if( deleting_tree )
            (*iter)->parent_cluster = NULL;
         ++iter;
      }
   }
}

int Cluster_GLM_MVN::get_num_single_subject_clusters_allocated( bool top )
{
   if( clusters_allocated.empty() || ( Rptr && !top && Rptr->is_cluster_in_model( this ) ) )
      return 1;
   int num_clusters = 0;
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
      if( !(*iter)->clusters_allocated.empty() && ( !Rptr || !Rptr->is_cluster_in_model( *iter ) ) )
         num_clusters += (*iter)->get_num_single_subject_clusters_allocated( false );
      else
	 num_clusters++;
   return num_clusters;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::draw_random_single_subject_cluster_allocated()
{
   map<Cluster_GLM_MVN*, double> map_of_clusters = get_single_subject_clusters_allocated();
   int num_single_subject_clusters = map_of_clusters.size();
   if( !num_single_subject_clusters )
      return NULL;
   int clnum = (int)gsl_rng_uniform_int( Rptr->rn, num_single_subject_clusters );
   map<Cluster_GLM_MVN*, double>::iterator iter = map_of_clusters.begin();
   for( int i = 0; i < clnum; i++ )
      iter++;
   return iter->first;
}

map<Cluster_GLM_MVN*, double> Cluster_GLM_MVN::get_all_clusters_allocated()
{
   map<Cluster_GLM_MVN*, double> map_of_clusters;
   get_single_subject_clusters_allocated( map_of_clusters );
   return map_of_clusters;
}

void Cluster_GLM_MVN::get_all_clusters_allocated( map<Cluster_GLM_MVN*, double>& map_of_clusters )
{
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      map_of_clusters[ *iter ] = -DBL_MAX;
      if( !(*iter)->clusters_allocated.empty() )
         (*iter)->get_single_subject_clusters_allocated( map_of_clusters );
   }
}

Cluster_GLM_MVN* Cluster_GLM_MVN::draw_random_cluster_from_all_allocated()
{
   map<Cluster_GLM_MVN*, double> map_of_clusters = get_all_clusters_allocated();
   int num_clusters = map_of_clusters.size();
   if( !num_clusters )
      return NULL;
   int clnum = (int)gsl_rng_uniform_int( Rptr->rn, num_clusters );
   map<Cluster_GLM_MVN*, double>::iterator iter = map_of_clusters.begin();
   for( int i = 0; i < clnum; i++ )
      iter++;
   return iter->first;
   
}

void Cluster_GLM_MVN::delete_down_to_subject_clusters()
{
   int num_clusters_allocated = clusters_allocated.size();
   list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin();
   for( int i = 0; i < num_clusters_allocated; i++ )
      if( (*iter)->clusters_allocated.size() )
      {
         (*iter)->delete_down_to_subject_clusters();
         delete *iter;
         iter = clusters_allocated.begin();
      }
      else
      {
         (*iter)->parent_cluster = NULL;
         ++iter;
      }
}

void Cluster_GLM_MVN::set_cluster_closeness( Cluster_GLM_MVN* cl, const double d )
{
   if( !Rptr->closeness_matrix )
      closeness_map[ cl ] = d;
   else
   {
      if( smallest_subject < cl->smallest_subject )
         Rptr->closeness_matrix[ cl->smallest_subject ][ smallest_subject ] = d;
      else
         Rptr->closeness_matrix[ smallest_subject ][ cl->smallest_subject ] = d;
   }
}

double Cluster_GLM_MVN::get_cluster_closeness( Cluster_GLM_MVN* cl, bool look_up )
{
  //   if( ( !Rptr->preserve_tree && !Rptr->heteroskedastic ) || !Rptr->mcmc_sampling )
   if( look_up && ( Rptr->storing_cluster_closenesses && ( !Rptr->mcmc_sampling || ( !clusters_allocated.size() && !cl->clusters_allocated.size() ) ) ) )
   {
      if( !Rptr->closeness_matrix )
      {
         cluster_ptrtodbl::iterator iter = closeness_map.find( cl );
         if( iter != closeness_map.end() )
            return iter->second;
         iter = cl->closeness_map.find( this );
         if( iter != cl->closeness_map.end() )
            return iter->second;
      }
      else{
         if( smallest_subject < cl->smallest_subject )
            return Rptr->closeness_matrix[ cl->smallest_subject ][ smallest_subject ];
         else
            return Rptr->closeness_matrix[ smallest_subject ][ cl->smallest_subject ];
      }
   }
   double d = Rptr->calculate_closeness( this, cl );
   if( !Rptr->euclidean_or_correlation_clustering )
      d += Rptr->get_change_log_prior_cluster_size( get_num_subjects_allocated(), cl->get_num_subjects_allocated(), false );

   if( Rptr->storing_cluster_closenesses )
   {
      set_cluster_closeness( cl, d );
      if( !Rptr->closeness_matrix )
         cl->set_cluster_closeness( this, d );//by symmetry
   }
   return d;
}

double Cluster_GLM_MVN::get_heller_prior_d()
{
   if( heller_prior_d )
      return *heller_prior_d;
   return Rptr->log_heller_dirichlet_alpha;
}

void Cluster_GLM_MVN::erase_from_closeness_map( Cluster_GLM_MVN* cl )
{
   closeness_map.erase( cl );
}

double Cluster_GLM_MVN::find_closest_cluster( map< Cluster_GLM_MVN*, double >* map_of_clusters_ptr )
{
   double closeness, max_closeness = -DBL_MAX, second_max_closeness = -DBL_MAX;
   closest_cluster = 0;
   second_closest_cluster = 0;
   if( !Rptr->closeness_matrix )
      for( cluster_ptrtodbl::iterator iter = closeness_map.begin(); iter != closeness_map.end(); iter++ )
      {
         closeness = iter->second;
         if( max_closeness < closeness && (iter->first)->mergeable && ( !using_types || type == iter->first->type ) )
         {
            second_max_closeness = max_closeness;
            second_closest_cluster = closest_cluster;
            max_closeness = closeness;
            closest_cluster = iter->first;
         }
         else if( second_max_closeness < closeness && (iter->first)->mergeable && ( !using_types || type == iter->first->type ) )
         {
            second_max_closeness = closeness;
            second_closest_cluster = iter->first;
         }
      }
   else
   {
      if( !map_of_clusters_ptr )
         map_of_clusters_ptr = &Rptr->cluster_map;
      for( cluster_ptrtodbl::iterator iter = map_of_clusters_ptr->begin(); iter != map_of_clusters_ptr->end(); ++iter )
      {
         Cluster_GLM_MVN* cl2 = iter->first;
         if( cl2 != this && ( !using_types || type == cl2->type ) )
         {
            if( smallest_subject < cl2->smallest_subject )
               closeness = Rptr->closeness_matrix[ cl2->smallest_subject ][ smallest_subject ];
            else
               closeness = Rptr->closeness_matrix[ smallest_subject ][ cl2->smallest_subject ];
            if( max_closeness < closeness && cl2->mergeable )
            {
               second_max_closeness = max_closeness;
               second_closest_cluster = closest_cluster;
               max_closeness = closeness;
               closest_cluster = cl2;
            }
            else if( second_max_closeness < closeness && cl2->mergeable )
            {
               second_max_closeness = closeness;
               second_closest_cluster = cl2;
            }
         }
      }
   }
   return max_closeness;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::get_closest_cluster()
{
   if( Rptr->finished_mcmc_sampling && Rptr->preserve_tree )
   {
      closest_cluster = get_sibling_cluster();
//       if( closest_cluster )
//       {
//          if( Rptr->is_cluster_in_model( closest_cluster ) )
//          {
//             Rptr->cluster_map[ this ] = Rptr->calculate_closeness( this, closest_cluster );
//             if( !Rptr->euclidean_or_correlation_clustering )
//                Rptr->cluster_map[ this ] += Rptr->get_change_log_prior_cluster_size( num_subjects_allocated, closest_cluster->num_subjects_allocated, false );
//          }
//          else
//          {
//             closest_cluster = NULL;
//             Rptr->cluster_map[ this ] = -DBL_MAX;
//          }
//       }
   }
   return closest_cluster;
}

map< Cluster_GLM_MVN*, double, less <Cluster_GLM_MVN*> > Cluster_GLM_MVN::get_closest_clusters( int k )
{
  cluster_ptrtodbl map_k;
  cluster_ptrtodbl::iterator iter;
  Cluster_GLM_MVN* cl = NULL;
  double max_closeness;
  map_k[ closest_cluster ] = closeness_map[ closest_cluster ];
  for( int i = 0; i < k - 1; i++ )
  {
     max_closeness = -DBL_MAX;
     for( iter = closeness_map.begin(); iter != closeness_map.end(); iter++ )
        if( max_closeness < iter->second && (iter->first)->mergeable && map_k.find( iter->first ) == map_k.end() )
           cl = iter->first;
     map_k[ cl ] = closeness_map[ cl ];
  }
  return map_k;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::merge_with_nearest_cluster()
{
   return merge_cluster( closest_cluster );
}

Cluster_GLM_MVN* Cluster_GLM_MVN::merge_cluster( Cluster_GLM_MVN*& cl, bool calc_closeness )
{
   if( !mergeable || !cl->mergeable )
   {
      cout << "Error: One of these clusters can't be merged" << endl;
      exit(1);
   }
   if( using_types && type != cl->type )
   {
      cout << "Error: These clusters are of different types [" << type << "," << cl->type << "] so can't be merged" << endl;
      cout << this << endl;
      cout << cl << endl;
      exit(1);
   }
     
   if( Rptr->preserve_tree )
      return new Cluster_GLM_MVN( this, cl );

   int i, j;
   v_index = Rptr->optimal_v_index;
   
   if( calc_closeness && !Rptr->euclidean_or_correlation_clustering && !Rptr->average_complete_single_link)
   {
      if( Rptr->heteroskedastic )
      {
         if( !Rptr->heller_algorithm )
         {
            closeness_contribution += cl->closeness_contribution + get_cluster_closeness( cl );
            closeness_contribution -= Rptr->get_change_log_prior_cluster_size( num_subjects_allocated, cl->num_subjects_allocated, false );
         }
         else
	 {
	   //            closeness_contribution = Rptr->new_contribution;
	    closeness_contribution = Rptr->heller_current_contribution;
	    *heller_prior_d = Rptr->heller_d;
	 }
      }
      else
      {
         closeness_contribution += cl->closeness_contribution + Rptr->proposed_ytd_vstar_dty_sum - Rptr->current_ytd_vstar_dty_sum;
         Rptr->current_ytd_vstar_dty_sum = Rptr->proposed_ytd_vstar_dty_sum;
      }
   }

   if( calc_closeness && Rptr->order_cluster_subjects )
      order_clusters_allocated( this, cl );

   if( Rptr->keep_vstars == 2 )
   {
      if( !Rptr->correlate_experiments )
         for( j = 0; j < Rptr->num_experiments; j++ )
         {
            Rptr->delete_vstar( gsl_matrix_int_get( num_obsns, 0, j ) );
            Rptr->delete_vstar( gsl_matrix_int_get( cl->num_obsns, 0, j ) );
         }
      else
      {
         Rptr->delete_vstar( get_total_num_obsns() );
         Rptr->delete_vstar( cl->get_total_num_obsns() );
      }
   }
 
   number_mergeable--;
   splitable = 1;
   gsl_matrix_int_add( num_obsns, cl->num_obsns );
   if( yty )
      gsl_matrix_add( yty, cl->yty );
   if( ytd )
      gsl_matrix_add( ytd, cl->ytd );
   if( dtd )
      gsl_matrix_add( dtd, cl->dtd );
   yty_sum += cl->yty_sum;
   
   if( y_averages )
   {
      for( j = 0; j < Rptr->p_y; j++ ) 
         for( i = 0; i < Rptr->data_block_size; i++ )
            gsl_matrix_set( y_averages, i, j, ( gsl_matrix_get( y_averages, i, j ) * num_subjects_allocated + gsl_matrix_get( cl->y_averages, i, j ) * cl->num_subjects_allocated ) / (double)( num_subjects_allocated + cl->num_subjects_allocated) );
   }
   if( ys && cl->ys )
      matrix_join( ys, cl->ys, false, false );

   if( cl->get_distinct() )
      distinct = 1;
   if( cl->label )
   {
      if( !label )
         label = new string( *(cl->label) );
      else
         if( ( !Rptr->cluster_by_labels || *label != *(cl->label) ) && !Rptr->mcmc_sampling )
            *label +=  "; " + *(cl->label);
   }
   if( splitable && cl->splitable )
      number_splitable--;
   else if( num_subjects_allocated < 2 && cl->num_subjects_allocated < 2 )
      number_splitable++;
   if( ( calc_closeness && Rptr->order_cluster_subjects ) || smallest_subject < cl->smallest_subject )
      subjects_allocated.splice( subjects_allocated.end(), cl->subjects_allocated );
   else
      subjects_allocated.splice( subjects_allocated.begin(), cl->subjects_allocated );
   if( calc_closeness && ( Rptr->euclidean_or_correlation_clustering || Rptr->average_complete_single_link ) )
      closeness_contribution = Rptr->calculate_closeness_contribution( this );

   if( calc_closeness && Rptr->order_cluster_subjects )
   {
      if( num_subjects_allocated < cl->num_subjects_allocated )
      {
         subject_join = subjects_allocated.begin();
         for( i = 0; i < num_subjects_allocated; i++ )
            ++subject_join;
      }
      else
      {
         subject_join = subjects_allocated.end();
         for( i = 0; i < cl->num_subjects_allocated; i++ )
            --subject_join;
      }
   }
   num_subjects_allocated += cl->num_subjects_allocated;
   int other_smallest_subject = cl->smallest_subject;
   if( other_smallest_subject < smallest_subject )
      switch_values( smallest_subject, other_smallest_subject );
   if( Rptr->closeness_matrix && Rptr->reduce_closeness_matrix )
      zap_vector( Rptr->closeness_matrix[ other_smallest_subject ] );
   if( cluster_error_covariances )
      update_cluster_error_covariances();
   closeness_map.erase( cl );
   if( !Rptr->retain_singleton_clusters || cl->num_subjects_allocated > 1 )
      delete cl;
   return this;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::split_cluster( list<int>* cl_subjects_ptr )
{
   list<int> cl_subjects;
   list<int>::iterator iter;
   bool snap_subjects = false;
   if( !splitable || !mergeable )
   {
      cout << "Error: Cannot split this cluster" << endl;
      exit(1);
   }
   if( two_means_separation && !split_off_one_subject && Rptr && Rptr->mcmc_sampling && Rptr->map_allocation_sweep == 0 )
   { // if cl_subjects non-empty, then assume a merge move is taking place and this call is to find the prob of jumping back, splitting the subjects_allocated to leave cl_subjects in one cluster and the rest in the other.
      list<int> cl_subjects2;
      initialise_two_means_clustering( cl_subjects,  cl_subjects2 );
      tempered_gibbs_sweep( cl_subjects, cl_subjects2, cl_subjects_ptr );
   }
   else if( !cl_subjects_ptr )
      if( split_off_one_subject )
         cl_subjects.push_back( find_random_subject() );
      else
      {
         if( snap_subjects )
         {
            int i, cl_subjects_size = 0;
//            cl_subjects_size = 1 + static_cast<int>( gsl_ran_flat( Rptr->rn, 0, num_subjects_allocated - 1 ) );
            while( !cl_subjects_size || cl_subjects_size == num_subjects_allocated )
               cl_subjects_size = gsl_ran_binomial( Rptr->rn, .5, num_subjects_allocated );
            int* subjects_array = new int[ num_subjects_allocated ];
            iter = subjects_allocated.begin();
            for( i = 0; i < num_subjects_allocated; i++ )
            {
               subjects_array[ i ] = Rptr->inverse_dendrogram[ *iter ];
               ++iter;
            }
            qsort( subjects_array, num_subjects_allocated, sizeof(int), unequal_integer_compare );
            int start = gsl_rng_uniform_int( Rptr->rn, num_subjects_allocated - cl_subjects_size + 1 );
//               static_cast<int>( gsl_ran_flat( Rptr->rn, 0, num_subjects_allocated - cl_subjects_size + 1 ) );
            for( i = start; i < start + cl_subjects_size; i++ )
               cl_subjects.push_back( Rptr->dendrogram[ subjects_array[ i ] ] );
            delete[] subjects_array;
         }
         else
         {
            int cl_subjects_size = 0;
            while( !cl_subjects_size || cl_subjects_size == num_subjects_allocated )
            {
               cl_subjects_size = 0;
               cl_subjects.erase( cl_subjects.begin(), cl_subjects.end() );
               for( iter = subjects_allocated.begin(); iter != subjects_allocated.end(); iter++ )
                  if( gsl_ran_bernoulli( Rptr->rn, .5 ) )
                  {
                     cl_subjects.push_back( *iter );
                     cl_subjects_size++;
                  }
            }
         }
      }
   else
   {
      if( cl_subjects_ptr->size() == subjects_allocated.size() )
         return NULL;
      if( !Rptr->order_cluster_subjects )
         cl_subjects = *cl_subjects_ptr;
      else
        for( iter = subjects_allocated.begin(); iter != subjects_allocated.end(); iter++ )
           if( find( cl_subjects_ptr->begin(), cl_subjects_ptr->end(), *iter ) != cl_subjects_ptr->end() )
               cl_subjects.push_back( *iter );
   }
   
   gsl_matrix_int* cl_num_obsns = gsl_matrix_int_calloc( 1, num_obsns->size2 );
   gsl_matrix *cl_yty = 0, *cl_ytd = 0, *cl_dtd = 0;
   if( yty )
      cl_yty =  gsl_matrix_calloc( yty->size1, yty->size2 );
   if( ytd )
      cl_ytd = gsl_matrix_calloc( ytd->size1, ytd->size2 );

   if( dtd )
      cl_dtd = gsl_matrix_calloc( dtd->size1, dtd->size2 );
   for( iter = cl_subjects.begin(); iter != cl_subjects.end(); iter++ )
   {
      if( yty )
         gsl_matrix_add( cl_yty, Rptr->get_yty_for_subject( *iter ) );
      if( ytd )
         gsl_matrix_add( cl_ytd, Rptr->get_ytd_for_subject( *iter ) );
      if( cl_dtd )
         gsl_matrix_add( cl_dtd, Rptr->get_dtd_for_subject( *iter ) );
      gsl_matrix_int_add( cl_num_obsns, Rptr->get_nob_for_subject( *iter ) );
   }

   for( iter = cl_subjects.begin(); iter != cl_subjects.end(); iter++ )
      subjects_allocated.remove( *iter );

   subject_join = subjects_allocated.begin();

   gsl_matrix_int_sub( num_obsns, cl_num_obsns );
   if( yty )
      gsl_matrix_sub( yty, cl_yty );
   if( ytd )
      gsl_matrix_sub( ytd, cl_ytd );
   if( dtd )
      gsl_matrix_sub( dtd, cl_dtd );
   if( subjects_allocated.size() < 2 )
   {
      splitable = 0;
      number_splitable--;
   }
   bool cl_distinct = 0;
   if( distinct )
      cl_distinct = 1;
   Cluster_GLM_MVN* cl;
   if( !label )
      cl = new Cluster_GLM_MVN( cl_num_obsns, cl_yty, cl_ytd, cl_dtd, cl_distinct, 1, cl_subjects, -DBL_MAX );
   else
      cl = new Cluster_GLM_MVN( cl_num_obsns, cl_yty, cl_ytd, cl_dtd, cl_distinct, 1, cl_subjects, -DBL_MAX, *label );
   num_subjects_allocated -= cl->num_subjects_allocated;
   yty_sum -= cl->get_yty_sum();
   cl->find_smallest_and_largest_subjects_allocated();
   if( cl->smallest_subject == smallest_subject )
      find_smallest_and_largest_subjects_allocated();
   if( y_averages )
   {
      set_y_averages();
      cl->set_y_averages();
   }
   if( ys )
   {
      set_targets();
      cl->set_targets();
   }
   cl->type = type;
   double temp_closeness = closeness_contribution;
   int old_v_index = v_index;
   if( !Rptr->splitting_to_optimal || Rptr->recalculate_whilst_splitting_to_optimal )
   {
      closeness_contribution = Rptr->calculate_closeness_contribution( this );
      v_index = Rptr->optimal_v_index;
      cl->closeness_contribution = Rptr->calculate_closeness_contribution( cl );
      cl->v_index = Rptr->optimal_v_index;
      if( Rptr->storing_cluster_closenesses || !Rptr->heteroskedastic )
      {
         temp_closeness -= closeness_contribution + cl->closeness_contribution;
         if( !Rptr->heteroskedastic )
         {
            Rptr->proposed_ytd_vstar_dty_sum = Rptr->current_ytd_vstar_dty_sum - temp_closeness;
            if( Rptr->storing_cluster_closenesses )
            {
               temp_closeness -= (Rptr->a_star_lookup)[ 0 ] * ( log( Rptr->yty_total - Rptr->current_ytd_vstar_dty_sum ) - log( Rptr->yty_total - Rptr->proposed_ytd_vstar_dty_sum ) );
               gsl_matrix_int* combined_num_obsns = gsl_matrix_int_alloc(  num_obsns->size1, num_obsns->size2);
               gsl_matrix_int_memcpy( combined_num_obsns, num_obsns );
               gsl_matrix_int_add( combined_num_obsns, cl_num_obsns );
               Rptr->set_v_index( old_v_index );
               temp_closeness += Rptr->half_log_det_vstar( combined_num_obsns );
               gsl_matrix_int_free( combined_num_obsns );
               Rptr->set_v_index( v_index );
               temp_closeness -= Rptr->half_log_det_vstar( num_obsns );
               Rptr->set_v_index( cl->v_index );
               temp_closeness -= Rptr->half_log_det_vstar( cl_num_obsns );
            }
         }
         if( !Rptr->euclidean_or_correlation_clustering )
            temp_closeness += Rptr->get_change_log_prior_cluster_size(num_subjects_allocated, cl->num_subjects_allocated, false );
         if( Rptr->storing_cluster_closenesses )
         {
            set_cluster_closeness( cl, temp_closeness );
            cl->set_cluster_closeness( this, temp_closeness );
         }
      }
   }
   gsl_matrix_int_free( cl_num_obsns );
   if( cl_yty )
      gsl_matrix_free( cl_yty );
   if( cl_ytd )
      gsl_matrix_free( cl_ytd );
   if( cl_dtd )
      gsl_matrix_free( cl_dtd );
   if( cluster_error_covariances )
   {
      update_cluster_error_covariances();
      cl->update_cluster_error_covariances();
   }
   return cl;
}

Cluster_GLM_MVN* Cluster_GLM_MVN::split_cluster( int cl_subject )
{
   list<int> cl_subjects;
   cl_subjects.push_back( cl_subject );
   return split_cluster( &cl_subjects );
}

int Cluster_GLM_MVN::find_random_subject()
{
    int sub_ind = gsl_rng_uniform_int( Rptr->rn, num_subjects_allocated );
    list<int>* subjects_allocated_ptr = &subjects_allocated;
    list<int> temp_subjects_allocated;
    if( !clusters_allocated.empty() )
    {
       temp_subjects_allocated = get_subjects_allocated();
       subjects_allocated_ptr = &temp_subjects_allocated;
    }
    list<int>::iterator iter = subjects_allocated_ptr->begin();
    for( int i = 0; i < sub_ind; i++ )
      ++iter;
    return *iter;
}

void Cluster_GLM_MVN::find_two_random_subjects( int& subj1, int& subj2 )
{
   int sub_ind1 = gsl_rng_uniform_int( Rptr->rn, num_subjects_allocated );
   int sub_ind2 = gsl_rng_uniform_int( Rptr->rn, num_subjects_allocated - 1 );
   if( sub_ind2 == sub_ind1 )
      sub_ind2 = num_subjects_allocated - 1;
   list<int>::iterator iter = subjects_allocated.begin();
   if( sub_ind1 > sub_ind2 )
      switch_values( sub_ind1, sub_ind2 );
   int i;
   for( i = 0; i < sub_ind1; i++ )
      iter++;
   subj1 = *iter;
   for( i = 0; i < sub_ind2 - sub_ind1; i++ )
      iter++;
   subj2 = *iter;
}

void Cluster_GLM_MVN::find_most_distant_subjects( int& subj1, int& subj2 )
{
   double dist, max_dist = -DBL_MAX;
   list<int>::iterator iter1, iter2;
   iter1 = subjects_allocated.begin();
   while( iter1 != subjects_allocated.end() )
   {
      iter2 = iter1;
      while( ++iter2 != subjects_allocated.end() )
      {
         dist = Rptr->calculate_squared_euclidean_or_correlation_distance( *iter1, *iter2 );
         if( dist > max_dist )
         {
            max_dist = dist;
            subj1 = *iter1;
            subj2 = *iter2;
         }
      }
      iter1++;
   }
}

void Cluster_GLM_MVN::initialise_two_means_clustering( list<int>& cl_subjects )
{
   list<int> cl_subjects2;
   initialise_two_means_clustering( cl_subjects, cl_subjects2 );
}

void Cluster_GLM_MVN::initialise_two_means_clustering( list<int>& cl_subjects, list<int>& cl_subjects2 )
{
   int subj1 = 0, subj2 = 0;
   double dist1, dist2;
   if( Rptr->deterministic_initial_split )
      find_most_distant_subjects( subj1, subj2 );
   else
      find_two_random_subjects( subj1, subj2 );
   cl_subjects.push_back( subj1 );
   cl_subjects2.push_back( subj2 );
   for( list<int>::iterator iter = subjects_allocated.begin(); iter != subjects_allocated.end(); iter++ )
      if( *iter != subj1 && *iter!= subj2 )
      {
         bool in_first_list = false; 
         if( Rptr->deterministic_initial_split )
         {
             list<int> current_subject;
             current_subject.push_back( *iter );
             dist1 = Rptr->calculate_squared_euclidean_or_correlation_distance( current_subject, cl_subjects );
             dist2 = Rptr->calculate_squared_euclidean_or_correlation_distance( current_subject, cl_subjects2 );
             in_first_list = ( dist1 < dist2 );
          }
         else if( gsl_ran_bernoulli( Rptr->rn, .5 ) )
            in_first_list = true;
         if( in_first_list )
            cl_subjects.push_back( *iter );
         else
            cl_subjects2.push_back( *iter );
      }
   tempered_gibbs_sweep( cl_subjects, cl_subjects2, NULL, Rptr->intermediate_gibbs_scans, false ); 
}

void Cluster_GLM_MVN::tempered_gibbs_sweep( list<int>& cl_subjects,  list<int>& cl_subjects2, list<int>* cl_subjects_ptr, int number_of_scans, bool calculate_probabilities )
{
   gibbs_sweep_log_prob = 0;
   int cl1_size = cl_subjects.size();
   int cl2_size = cl_subjects2.size();
   if( ( cl1_size <= 1 && cl2_size <= 1 ) || !number_of_scans )
      return;
   Cluster_GLM_MVN* cl1 = Rptr->create_temporary_cluster_for_subjects( cl_subjects );
   Cluster_GLM_MVN* cl2 = Rptr->create_temporary_cluster_for_subjects( cl_subjects2 );
   Cluster_GLM_MVN* cl3 = NULL;
   double log_p1, log_p2, max_logp, normaliser, p1, p2, contrib_13 = 0, contrib_23 = 0;
   int subject_num;

   bool two_means_separation_store = two_means_separation, in_cl_subjects_ptr = true, in_cl1;
   two_means_separation = false;//turn this off so we don't recursively call when we call split below
   for( int i = 0; i < number_of_scans; i++ )
   {
      list<int>::iterator iter = subjects_allocated.begin();
      bool cl1_corresponds_to_cl_subjects_ptr = true;
      if( cl_subjects_ptr )
      {
         in_cl_subjects_ptr = ( find( cl_subjects_ptr->begin(), cl_subjects_ptr->end(), *iter ) != cl_subjects_ptr->end() );
         in_cl1 = ( find( cl_subjects.begin(), cl_subjects.end(), *iter ) != cl_subjects.end() );
         if( !( in_cl_subjects_ptr ^ in_cl1 ) )
            cl1_corresponds_to_cl_subjects_ptr = true;
         else
            cl1_corresponds_to_cl_subjects_ptr = false;
      }
      while( ++iter != subjects_allocated.end() )//want to fix the first subject where it is
      {
         bool keep_looking_for_subject_to_move = true;
         do{
            subject_num = *iter;
            in_cl1 = cl1->cluster_contains_subject( subject_num );
            if( ( ( in_cl1 && cl1_size == 1 ) || ( !in_cl1 && cl2_size == 1 ) ) && subject_num != subjects_allocated.back() )
            {
               iter++;
               subjects_allocated.remove( subject_num );
               subjects_allocated.push_back( subject_num );
            }
            else
               keep_looking_for_subject_to_move = false;
         }while( keep_looking_for_subject_to_move );
         
         if( in_cl1 )
         {
            cl1_size--;
            contrib_13 = cl1->closeness_contribution;
            if( cl1_size )
               cl3 = cl1->split_cluster( subject_num );
         }
         else
         {
            cl2_size--;
            contrib_23 = cl2->closeness_contribution;
            if( cl2_size )
               cl3 = cl2->split_cluster( subject_num );
         }
         if( !cl1_size && in_cl1 )
         {
            p1 = 1;
            p2 = 0;
         }
         else if( !cl2_size && !in_cl1 )
         {
            p1 = 0;
            p2 = 1;
         }
         else
         {
            if( Rptr->gibbs_sweep_power )
            {
               if( in_cl1 )
                  contrib_23 = Rptr->calculate_closeness_contribution( cl2, cl3 );
               else
                  contrib_13 = Rptr->calculate_closeness_contribution( cl1, cl3 );
               log_p1 = contrib_13 - cl1->closeness_contribution + Rptr->get_change_log_prior_cluster_size( cl1_size, 1, false );//effectively, log_p1 = Rptr->calculate_log_predictive_probability( cl3, cl1 );
               log_p2 = contrib_23 - cl2->closeness_contribution + Rptr->get_change_log_prior_cluster_size( cl2_size, 1, false );//effectively, log_p2 = Rptr->calculate_log_predictive_probability( cl3, cl2 );
               if( log_p2 > log_p1 )
                  max_logp = log_p2;
               else
                  max_logp = log_p1;                 
               p1 = exp( Rptr->gibbs_sweep_power * ( log_p1 - max_logp ) );
               p2 = exp( Rptr->gibbs_sweep_power * ( log_p2 - max_logp ) );
               normaliser = p1 + p2;
               p1 /= normaliser;
               p2 = 1 - p1;
            }
            else
               p1 = p2 = .5;
         }
         zap( cl3 );
         if( cl_subjects_ptr )
            in_cl_subjects_ptr = ( find( cl_subjects_ptr->begin(), cl_subjects_ptr->end(), subject_num ) != cl_subjects_ptr->end() );
         if( cl2_size && ( !cl1_size || ( !cl_subjects_ptr && Rptr->with_probability( p1 ) ) || ( cl_subjects_ptr && !( in_cl_subjects_ptr ^ cl1_corresponds_to_cl_subjects_ptr ) ) ) )
         {
            if( cl1_size )
            {
               cl1->add_subject( subject_num );
               cl1->closeness_contribution = contrib_13;
//               cl1->merge_cluster( cl3 );
            }
            cl1_size++;
            if( calculate_probabilities )
               gibbs_sweep_log_prob += log( p1 );
         }
         else
         {
            if( cl2_size )
            {
               cl2->add_subject( subject_num );
               cl2->closeness_contribution = contrib_23;
//               cl2->merge_cluster( cl3 );
            }
            cl2_size++;
            if( calculate_probabilities )
               gibbs_sweep_log_prob += log( p2 );
         }
      }
   }
   cl_subjects = cl1->get_subjects_allocated();
   cl_subjects2 = cl2->get_subjects_allocated();
   delete cl1;
   delete cl2;
   two_means_separation = two_means_separation_store;
}

void Cluster_GLM_MVN::add_subject( int subject, string* cl_label )
{
   subjects_allocated.push_back( subject );
   num_subjects_allocated++;
   if( subject < smallest_subject )
      smallest_subject = subject;
   gsl_matrix_add( yty, Rptr->get_yty_for_subject( subject ) );
   gsl_matrix_add( ytd, Rptr->get_ytd_for_subject( subject ) );
   if( dtd )
      gsl_matrix_add( dtd, Rptr->get_dtd_for_subject( subject ) );
   gsl_matrix_int_add( num_obsns, Rptr->get_nob_for_subject( subject ) );
   yty_sum = .5 * sum_gsl_matrix( yty );
   if( !splitable && num_subjects_allocated > 1 && mergeable && !distinct )
   {
      splitable = 1;
      number_splitable++;
   }
   if( cl_label )
   {
      if( !label )
         label = new string( *(cl_label) );
      else
         if( *label != *(cl_label) )
            *label +=  "; " + *(cl_label);
   }
}

void Cluster_GLM_MVN::set_clusters_allocated( list<Cluster_GLM_MVN*> new_clusters_allocated )
{
   clusters_allocated = new_clusters_allocated;
}

void Cluster_GLM_MVN::set_clusters_allocated( Cluster_GLM_MVN* cl1, Cluster_GLM_MVN* cl2 )
{
   clusters_allocated.clear();
   clusters_allocated.push_back( cl1 );
   clusters_allocated.push_back( cl2 );
}

void Cluster_GLM_MVN::remove_allocated_cluster(  Cluster_GLM_MVN* cl )
{
   clusters_allocated.remove( cl );
}

void Cluster_GLM_MVN::set_parent_cluster( Cluster_GLM_MVN* new_parent_cluster )
{
   if( parent_cluster )
      parent_cluster->clusters_allocated.remove( this );
   parent_cluster = new_parent_cluster;
   if( parent_cluster )
   {
      parent_cluster->clusters_allocated.push_back( this );
      if( Rptr->order_cluster_subjects )
         parent_cluster->order_clusters_allocated();
   }
   /*
   Cluster_GLM_MVN* sibling_cluster = get_sibling_cluster();
   if( sibling_cluster )
   {
      sibling_closeness = get_cluster_closeness( sibling_cluster );
      sibling_cluster->sibling_closeness = sibling_closeness;
   }
   if( parent_cluster )
   {
      sibling_cluster = parent_cluster->get_sibling_cluster();
      if( sibling_cluster )
      {
	 parent_cluster->sibling_closeness = parent_cluster->get_cluster_closeness( sibling_cluster );
	 sibling_cluster->sibling_closeness = parent_cluster->sibling_closeness;
      }
   }
   */
}

void Cluster_GLM_MVN::set_my_root_cluster( Cluster_GLM_MVN* new_root_cluster )
{
   if( Rptr->cluster_by_labels && get_label() )
      if( label )
         set_root_cluster( *label, new_root_cluster );
      else
         set_root_cluster( *temp_label, new_root_cluster );
   else
      set_root_cluster( new_root_cluster );
}

void Cluster_GLM_MVN::set_root_cluster( Cluster_GLM_MVN* new_root_cluster )
{
   if( Rptr->preserve_tree )
   {
      if( root_cluster && root_cluster->mergeable )
         Rptr->all_clusters_in_tree.push_back( root_cluster );
      if( new_root_cluster && new_root_cluster->mergeable )
         Rptr->remove_from_all_clusters_deque( new_root_cluster );
   }
   root_cluster = new_root_cluster;
   root_cluster->set_parent_cluster( NULL );
}

void Cluster_GLM_MVN::set_root_cluster( string cl_label, Cluster_GLM_MVN* new_root_cluster)
{
   if( !label_to_root_cluster )
      label_to_root_cluster = new map<string, Cluster_GLM_MVN*>;
   if( Rptr->preserve_tree )
   {
      if( (*label_to_root_cluster).find( cl_label ) != (*label_to_root_cluster).end() && (*label_to_root_cluster)[ cl_label ]->mergeable )
         Rptr->all_clusters_in_tree.push_back( (*label_to_root_cluster)[ cl_label ] );
      if( new_root_cluster && new_root_cluster->mergeable)
         Rptr->remove_from_all_clusters_deque( new_root_cluster );
   }
   (*label_to_root_cluster)[ cl_label ] = new_root_cluster;
   new_root_cluster->set_parent_cluster( NULL );
}

Cluster_GLM_MVN* Cluster_GLM_MVN::get_my_root_cluster()
{
   if( Rptr->cluster_by_labels && get_label() )
   {
      if( label )
         return get_root_cluster( *label );
      else
         return get_root_cluster( *temp_label );
   }
   return get_root_cluster();
}

Cluster_GLM_MVN* Cluster_GLM_MVN::get_root_cluster( string& cl_label )
{
   if( label_to_root_cluster )
   {
      map<string, Cluster_GLM_MVN*>::iterator iter = (*label_to_root_cluster).find( cl_label );
      if( iter != (*label_to_root_cluster).end() )
         return iter->second;
   }
   return NULL;
}

list<Cluster_GLM_MVN*> Cluster_GLM_MVN::get_root_clusters()
{
   list<Cluster_GLM_MVN*> root_list;
   if( root_cluster )
      root_list.push_back( root_cluster );
   if( label_to_root_cluster )
   {
      map< string, Cluster_GLM_MVN* >::iterator iter;
      for( iter = label_to_root_cluster->begin(); iter != label_to_root_cluster->end(); iter++ )
         if( iter->second )
            root_list.push_back( iter->second );
   }
   return root_list;
}

int Cluster_GLM_MVN::get_number_of_root_clusters()
{
   int number_of_root_clusters = 0;
   if( root_cluster )
      number_of_root_clusters = 1;
   if( label_to_root_cluster )
      for( map< string, Cluster_GLM_MVN* >::iterator iter = label_to_root_cluster->begin(); iter != label_to_root_cluster->end(); iter++ )
         if( iter->second )
            number_of_root_clusters++;
   return number_of_root_clusters;
}

void Cluster_GLM_MVN::destroy_tree()
{
   list<Cluster_GLM_MVN*> root_list = get_root_clusters();
   for( list<Cluster_GLM_MVN*>::iterator iter = root_list.begin(); iter != root_list.end(); iter++ )
   {
      for( list<Cluster_GLM_MVN*>::iterator iter2 = (*iter)->clusters_allocated.begin(); iter2 != (*iter)->clusters_allocated.end(); iter2++ )
         ;
      if( (*iter)->subjects_allocated.empty() )
         delete *iter;
   }
   root_cluster = NULL;
   zap( label_to_root_cluster );
}

void Cluster_GLM_MVN::destroy_clusters_allocated()
{
   for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
   {
      (*iter)->destroy_clusters_allocated();
      if( (*iter)->subjects_allocated.empty() )
      {
         clusters_allocated.remove( *iter );
         delete *iter;
      }
   }
}

void Cluster_GLM_MVN::empty_parent_data_matrices_from_top()
{
  list<Cluster_GLM_MVN*> root_list = get_root_clusters();
  for( list<Cluster_GLM_MVN*>::iterator iter = root_list.begin(); iter != root_list.end(); iter++ )
     (*iter)->empty_parent_data_matrices();
}

void Cluster_GLM_MVN::empty_parent_data_matrices()
{
   if( clusters_allocated.size() )
   {
      if( num_obsns )
      {
         gsl_matrix_int_free( num_obsns );
         num_obsns = 0;
      }
      if( yty )
      {
         gsl_matrix_free( yty );
         yty = 0;
      }
      if( ytd )
      {
         gsl_matrix_free( ytd );
         ytd = 0;
      }
      if( dtd )
      {
         gsl_matrix_free( dtd );
         dtd = 0;
      }
      if( y_averages )
      {
         gsl_matrix_free( y_averages );
         y_averages = 0;
      }
      for( list<Cluster_GLM_MVN*>::iterator iter = clusters_allocated.begin(); iter != clusters_allocated.end(); iter++ )
         (*iter)->empty_parent_data_matrices();
   }
}

gsl_matrix* Cluster_GLM_MVN::update_cluster_error_covariances( bool bayesian_estimate )
{
   if( cluster_error_covariances )
      gsl_matrix_free( cluster_error_covariances );
   list< int > temp_subjects_allocated = get_subjects_allocated();
   int cl_num_subjects =  temp_subjects_allocated.size();
   cluster_error_covariances = Rptr->Regression::get_target_covariances( temp_subjects_allocated, false, (bool)Rptr->residuals );
   int ys_cols = Rptr->p_y * Rptr->data_block_sizes_sum * Rptr->num_experiments / Rptr->num_data_block_sizes;
   if( bayesian_estimate || cl_num_subjects <= ys_cols )
   {
      gsl_matrix_scale( cluster_error_covariances, cl_num_subjects );
      if( Rptr->scaled_error_covariances_prior_mean )
         gsl_matrix_add( cluster_error_covariances, Rptr->scaled_error_covariances_prior_mean );
//      else
         for( int i = 0; i < Rptr->error_cov_dim; i++ )
            *gsl_matrix_ptr( cluster_error_covariances, i, i ) += .0001 * (Rptr->error_cov_dim + 1);// regularise the cov matrix
      gsl_matrix_scale( cluster_error_covariances, 1.0 / cl_num_subjects );
   }
   get_cluster_error_sqrt_inverse_covariances();
   return cluster_error_covariances;
}

gsl_matrix* Cluster_GLM_MVN::get_cluster_error_sqrt_inverse_covariances()
{
   if( !cluster_error_covariances )
      update_cluster_error_covariances( true );
   if( cluster_error_sqrt_inverse_covariances )
      gsl_matrix_free( cluster_error_sqrt_inverse_covariances );
   cluster_error_sqrt_inverse_covariances = gsl_matrix_alloc( Rptr->error_cov_dim, Rptr->error_cov_dim );
   gsl_matrix *temp = gsl_matrix_alloc( Rptr->error_cov_dim, Rptr->error_cov_dim );
   gsl_matrix_memcpy( cluster_error_sqrt_inverse_covariances, cluster_error_covariances );
   gsl_matrix_memcpy( temp, cluster_error_covariances );
   gsl_matrix_memcpy( cluster_error_sqrt_inverse_covariances, temp );
   gsl_linalg_cholesky_decomp( temp );
   my_invert( temp, cluster_error_sqrt_inverse_covariances );
   gsl_matrix_free( temp );
   return cluster_error_sqrt_inverse_covariances;
}
        
bool Cluster_GLM_MVN::up_to_date()
{
   return Rptr->accept_since_iteration( last_iteration_updated );
}

int count_non_zero_entries( gsl_matrix_int* m )
{
   int num_rows = m->size1;
   int num_cols = m->size2;
   int non_zero_entries = 0;
   for( int i = 0; i < num_rows; i++ )
      for( int j = 0; j < num_cols; j++ )
         if( gsl_matrix_int_get( m, i, j ) )
            non_zero_entries++;
   return non_zero_entries;
}

