#include "glm.h"

double GLM::delta1 = 1;//.001;
double GLM::delta2 = 10;//.1;
bool GLM::g_prior = 0;
bool GLM::calc_empir_v = 0;
bool GLM::diagonal_v = false;
bool GLM::diagonal_vstar = false;
bool GLM::diagonalise_empir_v = 0;
double GLM::v_multiplier = 1000;
double GLM::v_multiplier_gamma_shape_parameter = 1e-2;
double GLM::v_multiplier_gamma_scale_parameter = 1e-2;

GLM::GLM( int argc, char **argv )
      : Basis_Fn_Regression( argc, argv )
{
   preliminary_glm_initialisation();
   int i;
   for( i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
         cout << "gprior=" << g_prior << "   (If non-zero, will use the g-prior for the prior variance matrix for the splines)" << endl;
         cout << "calculateempiricalv=" << calc_empir_v << "   (If non-zero, will calculate an empirical prior variance matrix for the splines, otherwise will use gammaparameter (see below) times the identity matrix)" << endl;
         cout << "diagonaliseempiricalv=" << diagonalise_empir_v << "   (If non-zero, then if calculating empirical V, set all the off diagonal elements of V to 0)" << endl;
         cout << "empiricalvmultiplier=" << v_multiplier << "   (A multiplier for the above matrix)" << endl;
         cout << "vmultipliershapeparameter=" << v_multiplier_gamma_shape_parameter << "   (Gamma distribution shape parameter for prior of V multiplier)" << endl;
         cout << "vmultiplierscaleparameter=" << v_multiplier_gamma_scale_parameter << "   (Gamma distribution scale parameter for prior of V multiplier)" << endl;
         cout << "priorvariancefile=" << V_Filename << "   (A file to write the above matrix to)" << endl;
         cout << "priorvvaluesfile=" <<  Prior_V_Values_Filename << "   (A file to read/write the values of the prior on V_)" << endl;
         cout << "gammaparameter=" << a_ << "   (Shape parameter for Inverse Gamma prior on observation error variance)" << endl;
         cout << "gammaparameterb=" << b_ << "   (Scale parameter for Inverse Gamma prior on observation error variance)" << endl;
         cout << "gammaparametera1=" << a_1 << "   (Alternative shape parameter for Inverse Gamma prior on observation error variance)" << endl;
         cout << "gammaparameterb1=" << b_1 << "   (Alternative scale parameter for Inverse Gamma prior on observation error variance)" << endl;
         cout << "delta1=" << delta1 << "   (Shape parameter of Inverse Gamma distribution for V multilplier if considered random )" << endl;
         cout << "delta2=" << delta2 << "   (Scale parameter of Inverse Gamma distribution for V multilplier if considered random )" << endl;
         cout << "gammahyperparameterg=" << b_g << endl;
         cout << "gausslaguerrev=" << gauss_laguerre_v << "   (If true, integrate out V using gauss laguerre integration.)" << endl;
         cout << "vmixture=" << mixture_of_vs << "   (If 2, then at each stage of clustering, choose V from the list of number_of_vs possible candidates.  If 1, average over them. Otherwise do clustering for fixed V, and repeat for the number_of_vs candidates)" << endl;
         cout << "correlateduncorrelatedmixture" << correlated_uncorrelated_mixture << "   (If true, then have additional mixture on V_ w.p. 1/2 of having experimental correlations)" << endl;
         cout << "vgridsize=" << number_of_vs << "   (The size of the grid over which to searh for optimal V)" << endl;
         cout << "vrange=" << v_range<< "   (The maximum multiplicative factor for the grid of  V)" << endl;
         return;
      }
      if ( InitialStringMatch( argv[i], "gprior=" ) )
         StringToBool( &g_prior, argv[i] );
      if ( InitialStringMatch( argv[i], "calculateempiricalv=" ) )
         StringToBool( &calc_empir_v, argv[i] );
      if ( InitialStringMatch( argv[i], "diagonaliseempiricalv=" ) )
         StringToBool( &diagonalise_empir_v, argv[i] );
      if ( InitialStringMatch( argv[i], "empiricalvmultiplier=" ) )
         StringToDouble( &v_multiplier, argv[i] );
      if ( InitialStringMatch( argv[i], "vmultipliershapeparameter=" ) )
         StringToDouble( &v_multiplier_gamma_shape_parameter, argv[i] );
      if ( InitialStringMatch( argv[i], "vmultiplierscaleparameter=" ) )
         StringToDouble( &v_multiplier_gamma_scale_parameter, argv[i] );
      if ( InitialStringMatch( argv[i], "priorvariancefile=" ) )
         StringToString( V_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "priorvvaluesfile=" ) )
         StringToString( Prior_V_Values_Filename, argv[i] );
      if ( InitialStringMatch( argv[i], "gammaparameter=" ) )
      {
         StringToDouble( &a_, argv[i] );
         b_ = a_;
      }
      if( InitialStringMatch( argv[i], "gammaparametera1=" ) )
         StringToDouble( &a_1, argv[i] );
      if( InitialStringMatch( argv[i], "gammaparameterb1=" ) )
         StringToDouble( &b_1, argv[i] );
      if( InitialStringMatch( argv[i], "gammahyperparameterg=" ) )
         StringToDouble( &b_g, argv[i] );
      if( InitialStringMatch( argv[i], "delta1=" ) )
         StringToDouble( &delta1, argv[i] );
      if( InitialStringMatch( argv[i], "delta2=" ) )
         StringToDouble( &delta2, argv[i] );
      if( InitialStringMatch( argv[i], "gausslaguerrev=" ) )
         StringToBool( &gauss_laguerre_v, argv[i] );
      if( InitialStringMatch( argv[i], "vmixture" ) )
         StringToLong( &mixture_of_vs, argv[i] );
      if( InitialStringMatch( argv[i], "correlateduncorrelatedmixture=" ) )
         StringToBool( &correlated_uncorrelated_mixture, argv[i] );
      if( InitialStringMatch( argv[i], "vgridsize=" ) )
         StringToLong( &number_of_vs, argv[i] );
      if( InitialStringMatch( argv[i], "vrange=" ) )
         StringToDouble( &v_range, argv[i] );
   }
   for( i = 1; i < argc; i++ )
      if ( InitialStringMatch( argv[i], "gammaparameterb=" ) )
         StringToDouble( &b_, argv[i] );
   
   if( !calc_empir_v )
      v_multiplier = 1 / coeff_prior_precision;
//   random_effects_precision = coeff_prior_precision;

   if( calculate_pairwise_variances )//then use these to automatically set b_ and b_1
   {
       b_ = ( a_ + 1 ) * LQ_pairwise_variance;
       b_1 = ( a_1 + 1 ) * median_pairwise_variance;
   }
   initialise_glm();
   if( print_to_screen )
   {
       cout << "Error variance ~ IG(" << a_ << "," << b_ << ") or IG(" << a_1 << "," << b_1 << ") " << endl;
       if( !mixture_of_vs || ( correlated_uncorrelated_mixture && number_of_vs == 2 ) )
           cout << "Basis function coefficient variance = " << v_multiplier << endl;
       else if( mixture_of_vs == 1 && delta1 > 0 && delta2 > 0 && ( !correlated_uncorrelated_mixture || number_of_vs > 2 ) )
           cout << "Basis function coefficient variance ~ IG(" << delta1 << "," << delta2 << ")   (using Guass-Laguerre quadrature on " << dim_gl << " points.)" << endl;
       if( random_effects_model )
       {
          if( !random_effect_variances )
             cout << "Random effects variance = " << 1 / random_effects_precision << endl;
          else
             cout << "Random effects variances read from file " << endl;
       }
   }

   if( !calc_empir_v && strlen(V_Filename) )
   {
      V_ = scan_data( V_Filename );
      V_inv = gsl_matrix_alloc( dim_V, dim_V );
      my_invert( V_, V_inv );//calculates V_inv to be the inverse of V_.
   }
}

GLM::GLM( gsl_matrix* X1, gsl_matrix* Y1 )
        : Basis_Fn_Regression( X1, Y1 )
{
   preliminary_glm_initialisation();
   initialise_glm();
}

GLM::GLM( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 )
        : Basis_Fn_Regression( X1, Y1, D1 )
{
   preliminary_glm_initialisation();
   initialise_glm();
}

GLM::~GLM()
{
   zap_vector( eta );
   if( E_y )
      gsl_matrix_free( E_y );
   if( m_star )
      gsl_matrix_free( m_star );
   if( b_star )
      gsl_matrix_free( b_star );
   if( b_matrix )
      gsl_matrix_free( b_matrix );
   if( V_inv_sqrt )
      gsl_matrix_free( V_inv_sqrt );
   if( beta_hat_mx )
      gsl_matrix_free( beta_hat_mx );
   for( int i = 0; i < number_of_vs; i++ )
   {
      GLM::set_v_index( i );
      if( V_ )
         gsl_matrix_free( V_ );
      if( V_inv )
         gsl_matrix_free( V_inv );
   }
   delete [] half_logdet_v_inv;
   delete [] log_prior_v;
   delete [] V_array;
   delete [] V_inv_array;
   if( error_covariances )
      gsl_matrix_free( error_covariances );
   if( global_error_covariances )
      gsl_matrix_free( global_error_covariances );
   if( scaled_error_covariances_prior_mean )
      gsl_matrix_free( scaled_error_covariances_prior_mean );
}

void GLM::preliminary_glm_initialisation()
{
   a_ = 1e-2;//DBL_MIN;
   b_ = a_;
   a_1 = a_;
   b_1 = b_ * 100;
   b_g = .2;
   b_h = 10;
   gauss_laguerre_v = 0;
   mixture_of_vs = 1;
   correlated_uncorrelated_mixture = false;
   number_of_vs = 1;
   v_matrices_differ = false;
   v_grid_size = 1;
   v_range = 100;
   V_ = 0;
   V_inv = 0;
   V_inv_sqrt = 0;
   *V_Filename = 0;
   *Prior_V_Values_Filename = 0;
   variable_v = 0;
}

void GLM::initialise_glm()
{
   int i, j;
   num_design_cols = D->size2;
   dim_V = num_design_cols;
   if( !V_inv )// use reference prior
   {
      V_inv = gsl_matrix_alloc( dim_V, dim_V );
      gsl_matrix_set_identity( V_inv );
      gsl_matrix_scale( V_inv, 1.0 / v_multiplier );
      diagonal_v = true;
   }
   //   if( !experiment_numbers )
   //      create_experiment_numbers_vector();   
   if( standardize_experiments )
   {
      find_experiment_sds();
      standardize_targets_by_experiment();
   }
   eta = 0;
   E_y = 0;
   m_star = 0;
   b_star = 0;
   error_covariances = NULL;
   global_error_covariances = NULL;
   scaled_error_covariances_prior_mean = NULL;
   if( p_y > 1 )
   {
      b_matrix = gsl_matrix_calloc( p_y, p_y );
      for( i = 0; i < p_y; i++ )
         gsl_matrix_set( b_matrix, i, i, b_ );
   }
   else
      b_matrix = 0;
   double b_h_scaler = y_max[ 0 ] - y_min[ 0 ];
   if( normalise_targets && y_scale[ 0 ] > 0 )
      b_h_scaler /= y_scale[ 0 ];
   b_h /= b_h_scaler * b_h_scaler;
   
   if( p_y == 1 )
   {
      lambda_prior_normalising_constant = a_ * log( b_ ) - gsl_sf_lngamma( a_ );
      lambda_prior_normalising_constant1 = a_1 * log( b_1 ) - gsl_sf_lngamma( a_1 );
   }
   else
   {
//      lambda_prior_normalising_constant = ( a_ + p_y - 1 ) / 2 * my_lndet( b_matrix ) - gsl_sf_lngamma( ( a_ + p_y - 1 ) / 2 ) - .25 * p_y * ( p_y - 1 ) * LOG_PI;
      lambda_prior_normalising_constant = ( a_ + p_y - 1 ) / 2 * p_y * log( b_ ) - gsl_sf_lngamma( ( a_ + p_y - 1 ) / 2 ) - .25 * p_y * ( p_y - 1 ) * LOG_PI;
      for( j = 1; j < p_y; j++ )
         lambda_prior_normalising_constant -= gsl_sf_lngamma( ( a_ + p_y - 1 ) / 2 - .5 * j );
      lambda_prior_normalising_constant = ( a_1 + p_y - 1 ) / 2 * p_y * log( b_1 ) - gsl_sf_lngamma( ( a_1 + p_y - 1 ) / 2 ) - .25 * p_y * ( p_y - 1 ) * LOG_PI;
      for( j = 1; j < p_y; j++ )
         lambda_prior_normalising_constant1 -= gsl_sf_lngamma( ( a_1 + p_y - 1 ) / 2 - .5 * j );
   }
   lambda_prior_normalising_constant_store = lambda_prior_normalising_constant;
//   lambda = draw_lambda_from_prior();
// draw_betas_from_prior();
   beta_hat_mx = 0;
   if( !gauss_laguerre_v && v_range <= 0 )
      number_of_vs = 1;
   if( !correlate_experiments && correlated_uncorrelated_mixture )
       calc_empir_v= true;
   if( gauss_laguerre_v )
   {
      mixture_of_vs = 1;
      gauss_laguerre_multiplier = 1;
//      gauss_laguerre_multiplier = 1.0 / v_multiplier;
//      if( !normalise_targets )
//         gauss_laguerre_multiplier = pow( 10, exp( log_ave_y_scale ) );
      create_gauss_laguerre_matrices();
      number_of_vs = dim_gl;
      double x_gl_00 = gsl_matrix_get( x_gl, 0, 0 );
      if( V_inv )
         gsl_matrix_scale( V_inv, v_multiplier / x_gl_00 );
      if( V_ )
	gsl_matrix_scale( V_, x_gl_00 / v_multiplier );
      v_multiplier = x_gl_00;
   }
   if( correlated_uncorrelated_mixture && mixture_of_vs )
   {
      number_of_vs *= 2;
      diagonal_v = false;
   }
   if( number_of_vs == 1 )
      mixture_of_vs = 0;
   v_multiplier_store = v_multiplier;
   if( !mixture_of_vs )
   {
      v_grid_size = number_of_vs;
      number_of_vs = 1;
      correlated_uncorrelated_mixture = false;
   }
   if( number_of_vs > 1 && !V_ && V_inv )
   {
      V_ = gsl_matrix_alloc( dim_V, dim_V );
      my_invert( V_inv, V_ );
   }
   V_array = new gsl_matrix*[ number_of_vs ];
   V_inv_array = new gsl_matrix*[ number_of_vs ];
   half_logdet_v_inv  = new double[ number_of_vs ];
   log_prior_v = new double[ number_of_vs ];
   current_v_index = 0;
   optimal_v_index = 0;
   for( i = 1; i < number_of_vs; i++ )
   {
      if( V_ )
         V_array[ i ] = gsl_matrix_alloc( dim_V, dim_V );
      else
         V_array[ i ] = 0;
      if( V_inv )
         V_inv_array[ i ] = gsl_matrix_alloc( dim_V, dim_V );
      else
         V_inv_array[ i ] = 0;
   }
   if( !strlen( Prior_V_Values_Filename ) )
      update_v_arrays( true );
   else
   {
      update_v_arrays( false );
//      scan_data( Prior_V_Values_Filename );
      scan_array_data( Prior_V_Values_Filename, log_prior_v, number_of_vs );
      for( i = 0; i < number_of_vs; i++ )
         log_prior_v[ i ] = log( log_prior_v[ i ] );
   }
   if( diagonal_v && gram_schmidt && !random_effects_model )
      diagonal_vstar = true;
   if( random_effect_variances )
      process_random_effect_variances( b_/(a_+1) );
}

double GLM::get_lambda()
{
   return lambda;
}

void GLM::set_lambda( double d )
{
   if( d < 0 )
   {
      cout << "Lambda = " << d << "not allowed, must be >0, setting to DBL_MIN " << endl;
      lambda = DBL_MIN;
   }
   else
      lambda = d;
}

double GLM::calculate_log_likelihood()
{
   log_likelihood = calculate_non_conjugate_log_likelihood();
   return log_likelihood;
}

double GLM::calculate_non_conjugate_log_likelihood()
{
   if( E_y )
   {
      gsl_matrix_free( E_y );
      E_y = 0;
   }
   E_y = inverse_link();
   log_likelihood = 0;
   for( int i = 0; i < n; i++ )
      for( int j = 0; j < p_y; j++ )
         log_likelihood += likelihood_fn( i, j );
   return log_likelihood;
}

double GLM::likelihood_fn( int rw, int cl )
{
   if( static_cast<int>( gsl_matrix_get( Y, rw, cl ) ) == 0 )
      return log( 1 - gsl_matrix_get( E_y, rw, 0 ) );
   else
      return log( gsl_matrix_get( E_y, rw, 0 ) );
}

void GLM::calculate_posterior_mean_parameters()
{
}

void GLM::calculate_linear_predictor()
{
   if( !eta )
      eta = new double[ n ];
   for( int i = 0; i < n; i++ )
   {
      double eta_i = 0;
      for( int j = 0; j < dim_V; j++ )
         eta_i += gsl_matrix_get( D, i, j ) * gsl_matrix_get( beta, i, 0 );
      eta[ i ] = eta_i;
   }
}

gsl_matrix* GLM::inverse_link()
{ 
  calculate_linear_predictor();
  gsl_matrix* inv_lnk = gsl_matrix_alloc( n, 1 );
  for( int i = 0; i < n; i++ )
    //    for ( int j = 0; j < num_cols; j++ )
      gsl_matrix_set( inv_lnk, i, 0, inverse_link_fn( eta[ i ] ) );
  return inv_lnk;
}

gsl_matrix* GLM::link( gsl_matrix* m )
{ 
  int num_rows = m->size1, num_cols = m->size2;
  gsl_matrix* lnk = gsl_matrix_alloc( num_rows, num_cols );
  for( int i = 0; i < num_rows; i++ )
    for ( int j = 0; j < num_cols; j++ )
      gsl_matrix_set( lnk, i, j, link_fn( gsl_matrix_get( m,  i, j ) ) );
  return lnk;
}

double GLM::inverse_link_fn( double d )
{
  return 1.0 / ( 1.0 + exp( - d ) );
  //  return d;
}

double GLM::link_fn( double d )
{
  return log( d / ( 1.0 - d ) );
  //  return d;
}

void GLM::draw_betas_from_prior()
{
   if( !beta )
      beta = gsl_matrix_alloc( dim_V, 1 );
   double *bs_orthog = new double[ dim_V ];
   int i, j;
   if( V_ && ! V_inv_sqrt )
   {
      V_inv_sqrt = gsl_matrix_alloc( dim_V, dim_V );
      gsl_matrix_memcpy( V_inv_sqrt, V_inv );
      gsl_linalg_cholesky_decomp( V_inv_sqrt );
      for( i = 0; i < dim_V; i++ )
         for( j = i + 1; j < dim_V; j++ )
            gsl_matrix_set( V_inv_sqrt, i, j, 0 );
   }
   for( i = 0; i < dim_V; i++ )
      bs_orthog[ i ] = gsl_ran_gaussian( rn, lambda );
//    bs_orthog[ i ] = ran_dble_exponential( rn, lambda );
   for( i = 0; i < dim_V; i++ )
      if( V_ )
      {
         double beta_i = 0;
         for( j = 0; j < dim_V; j++ )
            beta_i += gsl_matrix_get( V_inv_sqrt, i, j ) * bs_orthog[ j ];
         gsl_matrix_set( beta, i, 0, beta_i );
      }
      else
         gsl_matrix_set( beta, i, 0, bs_orthog[ i ] );
   delete [] bs_orthog;
}

double GLM::draw_lambda_from_prior()
{
  return gsl_ran_gamma( rn, a_, 1 / b_ ); // ~Gamma(a_,b_)
}

double GLM::get_lambda_density( double lambda_val )
{
  return lambda_prior_normalising_constant - b_ * lambda_val + ( a_ - 1 ) * log( lambda_val ); // ~Gamma(a_,b_)
}

/*
double GLM::draw_lambda_from_posterior()
{
  double sum_mod_betas = 0;
  for( int i = 0; i < dim_V; i++ )
    sum_mod_betas += abs( gsl_matrix_get( beta, i, 0 ) );
  return gsl_ran_exponential( rn, dim_V ) / sum_mod_betas; // ~Gamma(DBL_MIN,DBL_MIN)
}
*/

double GLM::draw_error()
{
    return gsl_ran_gaussian( rn, lambda );
}

void GLM::calculate_empirical_V()
{
   set_v_index( 0 );
   if( !V_ )
      V_ = gsl_matrix_alloc( dim_V, dim_V );
   if( ! V_inv )
      V_inv = gsl_matrix_alloc( dim_V, dim_V );
   gsl_matrix_set_identity( V_ );
   int num_repeats = dim_V / num_design_cols;
   if( num_data_blocks > 1 )
   {
      if( !beta_hat_mx )
         calculate_beta_hat_matrix();
      gsl_matrix_scale( V_, DBL_MIN );// added in in case any sample covariances are zero.
      gsl_matrix* covs = calculate_covariances( beta_hat_mx );
      for( int i = 0; i < num_design_cols; i++ )
         for( int j = 0; j < num_design_cols; j++ )
         {
            double v_ij = v_multiplier * gsl_matrix_get( covs, i, j );
            for( int k = 0; k < num_repeats; k++ )
               gsl_matrix_set( V_, k * num_design_cols + i, k * num_design_cols + j, v_ij );
         }
      gsl_matrix_free( covs );
      if( diagonalise_empir_v )
         diagonalise( V_ );
      my_invert( V_, V_inv );//calculates V_inv to be the inverse of V_.
   }
   else
   {
      gsl_matrix_scale( V_, v_multiplier );
      gsl_matrix_set_identity( V_inv );
      gsl_matrix_scale( V_inv, 1.0 / v_multiplier );
   }
   write_V_to_file();
}

void GLM::write_V_to_file()
{
   if( !strlen(V_Filename) )
      return;
   num_design_cols = D->size2;
   ofstream V_Stream( V_Filename, ios::out );
   for( int i = 0; i < num_design_cols; i++ )
   {
      for( int j = 0; j < num_design_cols; j++ )
      {
         double v_ij = V_ ? gsl_matrix_get( V_, i, j ) : ( i == j ? v_multiplier : 0 );
         V_Stream << v_ij << "  ";
      }
      V_Stream << endl;
   }
}

double GLM::get_log_prior_v_multiplier()
{
   if( delta1 <= 0 || delta2 <= 0 )
      return 0;
   return log( gsl_ran_gamma_pdf ( 1 / v_multiplier, delta1, 1 / delta2 ) ) - 2 * log( v_multiplier );
}

void GLM::set_v_index( int index )
{
   if( index == current_v_index )
      return;
   V_ = V_array[ index ];
   V_inv = V_inv_array[ index ];
   current_v_index = index;
   if( gauss_laguerre_v )
      v_multiplier = gsl_matrix_get( x_gl, index % dim_gl, 0 );
   else if( V_ )
   {
      v_multiplier = gsl_matrix_get( V_, 0, 0 );
      if( correlate_experiments )
         v_multiplier /= gsl_matrix_get( experiment_covariances, 0, 0 );
   }
   else if( v_range > 0 )
      v_multiplier = v_multiplier_store * pow( v_range, (double)index / ( number_of_vs - 1 ) );
}

void GLM::update_v_arrays( bool update_probabilities )
{
   int i, j, k;
   V_array[ 0 ] = V_;
   V_inv_array[ 0 ] = V_inv;
   if( V_inv )
      half_logdet_v_inv[ 0 ] = .5 * p_y * my_lndet( V_inv );
   else
      half_logdet_v_inv[ 0 ] = -.5 * p_y * dim_V * log( v_multiplier );
   if( correlated_uncorrelated_mixture )
   {
      gsl_matrix_set_zero( V_array[ number_of_vs / 2 ] );
      gsl_matrix_set_zero( V_inv_array[ number_of_vs / 2 ] );
      for( j = 0; j < dim_V; j++ )
      {
         double v_jj = gsl_matrix_get( V_, j, j );
         gsl_matrix_set( V_array[ number_of_vs / 2 ], j, j, v_jj );
         gsl_matrix_set( V_inv_array[ number_of_vs / 2 ], j, j, 1 / v_jj );
      }
      half_logdet_v_inv[ number_of_vs / 2 ] = .5 * p_y * my_lndet( V_inv_array[ number_of_vs / 2 ] );
   }
   double prior_v_multiplier_sum = 0;
   if( update_probabilities )
   {
      v_multiplier_store = v_multiplier;
      log_prior_v[ 0 ] = get_log_prior_v_multiplier();
      double prior_v = exp( log_prior_v[ 0 ] );
      if( correlated_uncorrelated_mixture )
      {
         log_prior_v[ number_of_vs / 2 ] = log_prior_v[ 0 ];
         prior_v_multiplier_sum = 2 * prior_v;
      }
      else
         prior_v_multiplier_sum = prior_v;
   }
   
   int nvs = correlated_uncorrelated_mixture ? number_of_vs / 2 : number_of_vs;
   double v_factor = 1;
   if( nvs > 1 )
      v_factor = pow( v_range, 1.0 / ( number_of_vs - 1 ) );
   double v_quot = 1 / v_factor;
   for( i = 1; i < nvs; i++ )
   {
      if( gauss_laguerre_v )
      {
         v_factor = gsl_matrix_get( x_gl, i, 0 ) / gsl_matrix_get( x_gl, i - 1, 0 );
         v_quot = 1 / v_factor;
      }
      if( V_ || V_inv )
         for( j = 0; j < dim_V; j++ )
            for( k = 0; k < dim_V; k++ )
            {
               if( V_ )
                  gsl_matrix_set( V_array[ i ], j, k, gsl_matrix_get( V_array[ i - 1 ], j, k ) * v_factor );
               if( V_inv )
                  gsl_matrix_set( V_inv_array[ i ], j, k, gsl_matrix_get( V_inv_array[ i - 1 ], j, k ) * v_quot );
            }
      if( V_inv )
         half_logdet_v_inv[ i ] = .5 * my_lndet( V_inv_array[ i ] );
      else
         half_logdet_v_inv[ i ] = half_logdet_v_inv[ i - 1 ] - .5 * p_y * log( v_factor );
      if( correlated_uncorrelated_mixture )
      {
         gsl_matrix_set_zero( V_array[ i + number_of_vs / 2 ] );
         gsl_matrix_set_zero( V_inv_array[ i + number_of_vs / 2 ] );
         for( j = 0; j < dim_V; j++ )
         {
            double v_jj = gsl_matrix_get( V_array[ i ], j, j );
            gsl_matrix_set( V_array[ i + number_of_vs / 2 ], j, j, v_jj );
            gsl_matrix_set( V_inv_array[ i + number_of_vs / 2 ], j, j, 1 / v_jj );
         }
         half_logdet_v_inv[ i + number_of_vs / 2 ] = .5 * p_y * my_lndet( V_inv_array[ i + number_of_vs / 2 ] );
      }
      if( update_probabilities )
      {
         v_multiplier *= v_factor;
         log_prior_v[ i ] = get_log_prior_v_multiplier();
         double prior_v = exp( log_prior_v[ i ] );
         if( correlated_uncorrelated_mixture )
         {
            log_prior_v[ i + number_of_vs / 2 ] = log_prior_v[ i ];
            prior_v_multiplier_sum += 2 * prior_v;
         }
         else
            prior_v_multiplier_sum += prior_v;
      }
   }
   if( update_probabilities )
   {
      v_multiplier = v_multiplier_store;
      if( !gauss_laguerre_v && prior_v_multiplier_sum > 0 )
         for( i = 0; i < number_of_vs; i++ )
            log_prior_v[ i ] -= log( prior_v_multiplier_sum );
   }
   if( !v_matrices_differ )
   {
      i = 0;
      while( ++i < number_of_vs && !v_matrices_differ )
         if( gsl_matrices_differ( V_inv_array[ i ], V_inv_array[ i - 1 ] ) )
            v_matrices_differ = true;
   }
}

double ran_dble_exponential( gsl_rng* r, double d )
{
  if ( gsl_ran_flat( r, 0, 1 ) < .5 )
    return gsl_ran_exponential( r, d );
  else
    return -gsl_ran_exponential( r, d );
}
