#include "glm_mvn.h"

long GLM_MVN::correlation_block = 0;
double GLM_MVN::closeness_coefficient = 50;
double GLM_MVN::closeness_power = 1;
double GLM_MVN::dirac_coefficient = 1;

GLM_MVN::GLM_MVN( int argc, char **argv )
        : GLM( argc, argv )
{
   preliminary_glm_mvn_initialisation();
   for( int i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
         cout << "calculateerrorcovariances" << calculate_error_covariances << "   (If true, estimate the covariances between time points in each cluster, both within and between experiments.)" << endl;
         cout << "correlatedblocksize=" << correlation_block << "   (Every this many data points x_i, x_j will have errors correlated with one another by exp( -(x_i-x_j) * closenesscoefficient ) + I(i=j)*diraccoefficient)" << endl;
         cout << "closenesscoefficient=" <<closeness_coefficient << endl;
         cout << "closenesspower=" <<closeness_power << endl;
         cout << "diraccoefficient=" << dirac_coefficient << endl;
         cout << "numbermontecarlosamples=" << number_monte_carlo_samples << "   (Number of monte carlo samples of curves to be taken for looking at posterior distributions of functions of interest)" << endl;
         cout << "l2distancefile=" << L2Distance_Filename << "   (File for writing monte carlo sample of l2 distances of curves from zero)" << endl;
         cout << "integralabsolutefile=" << Integral_Absolute_Filename << "   (File for writing monte carlo sample of integrals of abolute value of curves)" << endl;
         return;
      }
      if( InitialStringMatch( argv[i], "calculateerrorcovariances" ) )
         StringToBool( &calculate_error_covariances, argv[i] );
      if ( InitialStringMatch( argv[i], "correlatedblocksize=" ) )
         StringToLong( &correlation_block, argv[i] );
      if ( InitialStringMatch( argv[i], "closenesscoefficient=" ) )
         StringToDouble( &closeness_coefficient, argv[i] );
      if ( InitialStringMatch( argv[i], "closenesspower=" ) )
         StringToDouble( &closeness_power, argv[i] );
      if ( InitialStringMatch( argv[i], "diraccoefficient=" ) )
         StringToDouble( &dirac_coefficient, argv[i] );
      if ( InitialStringMatch( argv[i], "numbermontecarlosamples=" ) )
         StringToLong( &number_monte_carlo_samples, argv[i] );
      if( InitialStringMatch( argv[i], "l2distancefile=" ) )
         StringToString( L2Distance_Filename, argv[i] );
      if( InitialStringMatch( argv[i], "integralabsolutefile=" ) )
         StringToString( Integral_Absolute_Filename, argv[i] );
   }
   initialise_glm_mvn();
}

GLM_MVN::GLM_MVN( gsl_matrix* X1, gsl_matrix* Y1 )
      : GLM( X1, Y1 )
{
   preliminary_glm_mvn_initialisation();
   initialise_glm_mvn();
}

GLM_MVN::GLM_MVN( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 )
      : GLM( X1, Y1, D1 )
{
   preliminary_glm_mvn_initialisation();
   initialise_glm_mvn();
}

GLM_MVN::~GLM_MVN()
{
   if( V_star )
      gsl_matrix_free( V_star );
   if( V_star_inv )
      gsl_matrix_free( V_star_inv );
   if( V_star_sqrt )
      gsl_matrix_free( V_star_sqrt );
   zap_array( yty, p_y, p_y );
   if( D_sq )
      gsl_matrix_free( D_sq );
   if( C1 )
      gsl_matrix_free( C1 );
   if( C1D )
      gsl_matrix_free( C1D );
   if( C1Y )
      gsl_matrix_free( C1Y );
   if( yd_vstar_dy )
      gsl_matrix_free( yd_vstar_dy );
}

Regression* GLM_MVN::create_new_regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 )
{
   Regression* Rptr = new GLM_MVN( X1, Y1, D1 );
   return Rptr;
}

void GLM_MVN::preliminary_glm_mvn_initialisation()
{
   *L2Distance_Filename = 0;
   *Integral_Absolute_Filename = 0;
   number_monte_carlo_samples = 1;
   if( !strlen( PredictionFilename ) )
   {
      char* new_file = PredictionFilename;
      strcpy( new_file, TargetFilename );
      string insert = "_preds_";
      insert_in_string_before_dot( new_file, &insert );
   }
   if( !strlen( ParameterFilename ) )
   {
      char* new_file = ParameterFilename;
      strcpy( new_file, TargetFilename );
      string insert = "_parameters_";
      insert_in_string_before_dot( new_file, &insert );
   }
   calculate_error_covariances = false;
   scaled_error_covariances_prior_mean = NULL;
   error_cov_dim = 0;
}

void GLM_MVN::initialise_glm_mvn()
{
   D_sq = 0;
   D_sqs = 0;
   C1 = 0;
   C1D = 0;
   C1Y = 0;
//   C1D_sq = 0;
   if( correlation_block )
      calculate_closeness_matrix();
   V_star = 0;
   V_star_inv = 0;
   V_star_sqrt = 0;
   yty = 0;
   yd_vstar_dy = 0;

   if( p_y == 1 )
      a_star = a_ + .5 * regression_n;
   else
      a_star = ( a_ + regression_n + p_y - 1 ) / 2;

   if( C1 && !C1D )
      calculate_translated_D();
   
   if( !D_sq )
      calculate_D_sq();
   if( gram_schmidt )
      diagonalise( D_sq, 1e-6 );

   if( g_prior )
   {
      set_v_index( 0 );
      if( !V_ )
         V_ = gsl_matrix_alloc( dim_V, dim_V );
      if( ! V_inv )
         V_inv = gsl_matrix_alloc( dim_V, dim_V );
      gsl_matrix_memcpy( V_inv, D_sq );
      gsl_matrix_scale( V_inv, 1.0 / v_multiplier / gsl_matrix_get( D_sq, 0, 0 ) );
      my_invert( V_inv, V_ );
   }
   else if( calc_empir_v )
   {
      calculate_empirical_V();
      update_v_arrays( true );
   }

   if( ( correlate_experiments || correlate_columns ) && !strlen( Experiment_Locations_Input_Filename ) && !start_experiments_independent )
   {
         //initialise the experiment locations using MDS on correlation between beta_hat's for the different experiments.
      calculate_beta_hat_matrix();
      gsl_matrix* exp_locs = calculate_correlation_by_experiment( beta_hat_mx );
      experiment_covariances_to_locations( exp_locs );
      get_experiment_covariances();
      gsl_matrix_free( exp_locs );
//      if( !experiment_location_parameters )
//         for( int i = 0; i < num_locations; i++ )
//             for( int j = 0; j < experiment_location_dimension; j++ )
//                 experiment_location_means[ i ][ j ] = experiment_locations[ i ][ j ];
   }

   if( correlate_experiments )
   {  
      dim_V *= num_experiments;
      if( !D_sqs && !treat_experiments_as_different_dimensions )
      {
         matrix_block( D_sq, num_experiments );
         gsl_matrix_scale( D_sq, 1.0 / num_experiments );
      }
      else
      {
         for( int i = 1; i < num_data_block_sizes; i++ )
            matrix_block( D_sq, D_sqs[ i ] );
         delete [] D_sqs;
         D_sqs = 0;
      }
      if( beta )
      {
         gsl_matrix_free( beta );
         beta = gsl_matrix_alloc( dim_V, 1 );
      }
      for( int i = 0; i < number_of_vs; i++ )
      {
	 set_v_index( i );
	 matrix_block( V_inv, num_experiments );//expand V_ and D_sq to correct size
	 if( !V_ )
	 {
	    V_ = gsl_matrix_alloc( dim_V, dim_V );
	    my_invert( V_inv, V_ );
	 }
	 else
	    matrix_block( V_, num_experiments );
	 V_array[ i ] = V_;
	 V_inv_array[ i ] = V_inv;
         half_logdet_v_inv[ i ] *= num_experiments;
         if( !correlated_uncorrelated_mixture || i < number_of_vs / 2 )
            find_correlated_experiments_covariances();
      }
      set_v_index( 0 );
      update_v_arrays();
   }
   else if( correlate_columns )
   {
//      lambda_prior_normalising_constant -= ( a_ + p_y - 1 ) / 2 * my_lndet( b_matrix );
      lambda_prior_normalising_constant -= ( a_ + p_y - 1 ) / 2 * p_y * log( b_ );
      lambda_prior_normalising_constant1 -= ( a_1 + p_y - 1 ) / 2 * p_y * log( b_1 );
      get_experiment_correlations();
      for( int i = 0; i < p_y; i++ )
         for( int j = 0; j < p_y; j++ )
            gsl_matrix_set( b_matrix, i, j, b_ * gsl_matrix_get( experiment_covariances, i, j ) );
      double log_det_b_matrix = my_lndet( b_matrix );
      lambda_prior_normalising_constant += ( a_ + p_y - 1 ) / 2 * log_det_b_matrix;
      lambda_prior_normalising_constant1 += ( a_1 + p_y - 1 ) / 2 * (log_det_b_matrix + p_y*(log( b_1 )-log( b_ )));
   }

   calculate_ml_const();
}

void GLM_MVN::calculate_ml_const()
{
   if( !storing_basis_function_coefficients )
   {
      if( number_of_vs == 1 )
         ml_const = half_logdet_v_inv[ 0 ]; //- regression_n/2 * log( PI )
      else
         ml_const = 0;
   }
   else
      ml_const = gsl_sf_lngamma( a_ + regression_n * p_y ) - gsl_sf_lngamma( regression_n * p_y );
   if( C1 )
      ml_const += 0.5 * regression_num_subjects * my_lndet( C1 );
}

void GLM_MVN::calculate_ytys()
{
   create_array( yty, p_y, p_y );
   for( int i = 0; i < p_y; i++ )
      for( int j = 0; j < p_y; j++ )
      {
         yty[ i ][ j ] = 0;
         for( int ii = 0; ii < regression_n; ii++ )
            if( !C1Y )
               yty[ i ][ j ] += gsl_matrix_get( Y, ii, i ) * gsl_matrix_get( Y, ii, j );
            else
               yty[ i ][ j ] += gsl_matrix_get( C1Y, ii, i ) * gsl_matrix_get( C1Y, ii, j );
      }
}

double GLM_MVN::calculate_basis_coefficient_log_prior( double basis_coeff )
{
   return -.5 * ( LOG_2PI + log( v_multiplier ) + basis_coeff * basis_coeff / v_multiplier );
}

void GLM_MVN::create_posterior_parameter_matrices()
{
   if( !m_star )
      m_star = gsl_matrix_alloc( dim_V, p_y );
   if( !b_star )
      b_star = gsl_matrix_alloc( p_y, p_y );
   if( !yd_vstar_dy )//&& p_y > 1 )
      yd_vstar_dy = gsl_matrix_alloc( p_y, p_y );
}

void GLM_MVN::calculate_posterior_mean_parameters()
{
   if( C1 && !C1D )
      calculate_translated_D();
   if( !D_sq )
      calculate_D_sq();
   if( !V_star )
      V_star = gsl_matrix_alloc( dim_V, dim_V );
   if( V_star_inv && (int)V_star_inv->size1 != dim_V )
   {
      gsl_matrix_free( V_star_inv );
      V_star_inv = 0;
   }
   if( V_inv && (int)V_inv->size1 != dim_V )
   {
      gsl_matrix_free( V_inv );
      V_inv = gsl_matrix_alloc( dim_V, dim_V );
      gsl_matrix_set_identity( V_inv );
      gsl_matrix_scale( V_inv, 1.0 / v_multiplier );
   }
   if( !V_star_inv )
      V_star_inv = gsl_matrix_alloc( dim_V, dim_V );
   create_posterior_parameter_matrices();
   if( !yty )
      calculate_ytys();
   int i, j;
   if( V_inv )
      for( i = 0; i < dim_V; i++ )
         for( j = 0; j < dim_V; j++ )
            gsl_matrix_set( V_star_inv, i, j, gsl_matrix_get( V_inv, i, j ) + gsl_matrix_get( D_sq, i, j ) );
   else
      for( i = 0; i < dim_V; i++ )
         for( j = 0; j < dim_V; j++ )
            if( i != j )
               gsl_matrix_set( V_star_inv, i, j, gsl_matrix_get( D_sq, i, j ) );
            else
               gsl_matrix_set( V_star_inv, i, j, 1 / v_multiplier + gsl_matrix_get( D_sq, i, j ) );
   my_invert( *V_star_inv, V_star );
   double** DTY = 0;
   create_zeroed_array( DTY, dim_V, p_y );
   int shift = 0, row_counter = 0;
   for( int i0 = 0; i0 < regression_num_data_blocks; i0++ )
   {
      if( correlate_experiments )
         shift = get_experiment_number( i0 ) * num_design_cols;
      for( i = 0; i < num_design_cols; i++ )
	 for( j = 0; j < p_y; j++ )
         {
            for( int ii = row_counter; ii < row_counter + data_block_size; ii++ )
               if( !C1Y )
                  DTY[ i + shift ][ j ] += gsl_matrix_get( D, ii, i ) * gsl_matrix_get( Y, ii, j );
               else
                  DTY[ i + shift ][ j ] += gsl_matrix_get( C1D, ii, i ) * gsl_matrix_get( C1Y, ii, j );
         }
      row_counter += data_block_size;
   }
   
//   *m_star = (*V_star) * DTY;
   for( i = 0; i < dim_V; i++ )
      for( j = 0; j < p_y; j++ )
      {
         double mstar_ij = 0;
         for( int k = 0; k < dim_V; k++ )
            mstar_ij += gsl_matrix_get( V_star, i, k ) * DTY[ k ][ j ];
         gsl_matrix_set( m_star, i, j, mstar_ij );
      }
//   yd_vstar_dy = DTY.transpose * (*m_star );
   for( i = 0; i < p_y; i++ )
      for( j = 0; j < p_y; j++ )
      {
         double yd_vstar_dy_mx_ij = 0;
         for( int k = 0;  k < dim_V; k++ )
            yd_vstar_dy_mx_ij += DTY[ k ][ i ] * gsl_matrix_get( m_star, k, j );
         gsl_matrix_set( yd_vstar_dy, i, j, yd_vstar_dy_mx_ij );
      }
   zap_array( DTY, dim_V, p_y );
   if( p_y == 1 )
      gsl_matrix_set( b_star, 0, 0, b_ + .5 * ( yty[ 0 ][ 0 ] - gsl_matrix_get( yd_vstar_dy, 0, 0 ) ) );
   else
      for( i = 0; i < p_y; i++ )
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( b_star, i, j, gsl_matrix_get( b_matrix, i, j ) + yty[ i ][ j ] - gsl_matrix_get( yd_vstar_dy, i, j ) );
   e_sigma_sq = gsl_matrix_get( b_star, 0, 0 ) / ( a_star - 1 );
}

gsl_matrix* GLM_MVN::get_b_star_correlations()
{
   gsl_matrix* b_star_corr = gsl_matrix_alloc( p_y, p_y );
   for( int i = 0; i < p_y; i++ )
   {
      for( int j = 0; j < i; j++ )
      {
         double corr_ij = gsl_matrix_get( b_star, i, j ) / sqrt( gsl_matrix_get( b_star, i, i ) * gsl_matrix_get( b_star, j, j ) );
         gsl_matrix_set( b_star_corr, i, j, corr_ij );
         gsl_matrix_set( b_star_corr, j, i, corr_ij );
      }
      gsl_matrix_set( b_star_corr, i, i, 1 );
   }
   return b_star_corr;
}

double GLM_MVN::draw_lambda_from_posterior()
{
   lambda = gsl_ran_gamma( rn, a_star, 1 / gsl_matrix_get( b_star, 0, 0 ) ); // ~Gamma(a_star,b_star)
   return lambda;
}

void GLM_MVN::draw_betas_from_posterior()
{
   if( !beta )
      beta = gsl_matrix_alloc( dim_V, 1 );
   draw_lambda_from_posterior();
   double *bs_orthog = new double[ dim_V ];
   if( V_star_sqrt && (int)V_star_sqrt->size1 != dim_V )
   {
      gsl_matrix_free( V_star_sqrt );
      V_star_sqrt = 0;
   }
   if( !V_star_sqrt )
      V_star_sqrt = gsl_matrix_calloc( dim_V, dim_V );
   int i, j;
   for( i = 0; i < dim_V; i++ )
      for( j = 0; j < dim_V; j++ )
         gsl_matrix_set( V_star_sqrt, i, j, gsl_matrix_get( V_star, i, j ) );
   gsl_linalg_cholesky_decomp( V_star_sqrt );
   for( i = 0; i < dim_V; i++ )
      for( j = i + 1; j < dim_V; j++ )
         gsl_matrix_set( V_star_sqrt, i, j, 0 );
   for( i = 0; i < dim_V; i++ )
      bs_orthog[ i ] = gsl_ran_ugaussian( rn ) / sqrt( lambda ) ;
   for( i = 0; i < dim_V; i++ )
   {
      double beta_i = gsl_matrix_get( m_star, i, 0 );
      for( j = 0; j < dim_V; j++ )
         beta_i += gsl_matrix_get( V_star_sqrt, i, j ) * bs_orthog[ j ];
      gsl_matrix_set( beta, i, 0, beta_i );
      ( get_Basis_Fn_ptr( i ) )->set_coeff( beta_i );
   }
   delete [] bs_orthog;
}

double GLM_MVN::calculate_log_likelihood()
{
   if( !storing_basis_function_coefficients )
   {
      proposed_log_likelihood = ml_const + lambda_prior_normalising_constant - regression_n * p_y * LOG_SQRT_PI;
      double temp_closeness = 0, max_closeness = -DBL_MAX, ave_closeness = 0;
      for( int i = 0; i < number_of_vs; i++ )
      {
         set_v_index( i );
         calculate_posterior_mean_parameters();
         temp_closeness = gsl_sf_lngamma( p_y * a_star ) + .5 * p_y * my_lndet( V_star ) - a_star * my_lndet( b_star );
         if( number_of_vs > 1 )
            temp_closeness += half_logdet_v_inv[ i ];
         if( temp_closeness > max_closeness )
         {
            ave_closeness *= exp( max_closeness - temp_closeness );
            max_closeness = temp_closeness;
            temp_closeness = 0;
            optimal_v_index = i;
         }
         else
            temp_closeness -= max_closeness;
         if( mixture_of_vs == 1 )
            temp_closeness += log_prior_v[ i ];
         if( gauss_laguerre_v )
            temp_closeness += log( gsl_matrix_get( w_gl, current_v_index, 0 ) );
         ave_closeness += exp( temp_closeness );
      }
      ave_closeness = max_closeness + log( ave_closeness );
      if( mixture_of_vs == 1 )
         temp_closeness = ave_closeness;
      else
         temp_closeness = max_closeness;
      set_v_index( optimal_v_index );
      proposed_log_likelihood += temp_closeness;
   }
   else
   { // matrix variate density not available here, as basis fn coefficient is assumed a scalar in basis_fn.h.
      double* d_row = new double[ num_design_cols ];
      int j;
      for( j = 0; j < num_design_cols; j++ )
         d_row[ j ] = gsl_matrix_get( D, 0, j );
      double fn_val = function_evaluate_bases( d_row );
      double error_j, ssq = 0;
      for( j = 0; j < p_y; j++ )
      {
         error_j = gsl_matrix_get( Y, 0, j ) - fn_val;
         ssq += error_j * error_j;
      }
      for( int i = 1; i < regression_n; i++ )
      {
         bool rows_equal = true;
         int col = -1;
         while( ++col < num_design_cols && rows_equal )
            rows_equal = d_row[ col ] == gsl_matrix_get( D, 0, col );
         if( !rows_equal )
         {
            for( int j = 0; j < num_design_cols; j++ )
               d_row[ j ] = gsl_matrix_get( D, i, j );
            fn_val = function_evaluate_bases( d_row );
         }
         for( j = 0; j < p_y; j++ )
         {
            error_j = gsl_matrix_get( Y, i, j ) - fn_val;
            ssq += error_j * error_j;
         }
      }
      delete [] d_row;
      proposed_log_likelihood = ml_const - ( a_ + .5 * regression_n * p_y ) * log( b_ + .5 * ssq );
   }
   if( normalise_targets )
      proposed_log_likelihood -= regression_n * p_y * log_ave_y_scale;
   if( standardize_experiments )
      proposed_log_likelihood -= regression_n * p_y * log_ave_experiments_sds;
   return proposed_log_likelihood;
}

void GLM_MVN::calculate_closeness_matrix()
{
   calculate_C1_matrix();
   if( C1 )
      calculate_translated_Y();
}

void GLM_MVN::calculate_C1_matrix()
{
   if( n % correlation_block )
   {
      cout << "period given for correlations does not divide into the number of observations" << endl;
      exit(1);
   }
   int num_blocks;
   if( correlation_block == inputs_repeated_every )// then blocks in block diagonal will be identical, so less work to do.
   {
      if( !C1 )
         C1 = gsl_matrix_alloc( correlation_block, correlation_block );
      gsl_matrix_set_identity( C1 );
      num_blocks = 1;
   }
   else
   {
      if( !C1 )
         C1 = gsl_matrix_alloc( n, n );
      gsl_matrix_set_identity( C1 );
      num_blocks = n / correlation_block;
   }

   gsl_matrix *temp = gsl_matrix_alloc( correlation_block, correlation_block );
   gsl_matrix *temp2 = gsl_matrix_alloc( correlation_block, correlation_block );
   gsl_matrix_set_identity( temp );
   int i, i2;
   for( int block = 0; block < num_blocks; block++ )
   {
      for( i = 0; i < correlation_block; i++ )
         for( i2 = 0; i2 < correlation_block; i2++ )
         {
            double diff, dist = 0;
            for( int j = 0; j < p; j++ )
            {
               diff = gsl_matrix_get( X, i + block * num_blocks, j ) - gsl_matrix_get( X, i2 + block * num_blocks, j );
               dist += diff * diff;
            }
            gsl_matrix_set( temp, i, i2, exp( -closeness_coefficient * pow( dist, closeness_power ) ) );
            if( i == i2 )
               *gsl_matrix_ptr( temp, i, i2 ) += dirac_coefficient;
         }
      gsl_matrix_memcpy( temp2, temp );
      gsl_linalg_cholesky_decomp( temp );
      my_invert( temp, temp2 );
      gsl_matrix_free( temp );
      for( i = 0; i < correlation_block; i++ )
         for( i2 = 0; i2 < correlation_block; i2++ )
            gsl_matrix_set( C1, i + block*num_blocks, i2 + block*num_blocks, gsl_matrix_get( temp2, i, i2 ) );
      gsl_matrix_free( temp2 );
   }
}

void GLM_MVN::calculate_D_sq()
{
   if( data_block_sizes && !D_sqs && !treat_experiments_as_different_dimensions )
      D_sqs = new gsl_matrix*[ num_data_block_sizes ];

   int start_row = 0;
   for( int bs = 0; bs < (!treat_experiments_as_different_dimensions ? num_data_block_sizes:1); bs++ )
   {
      if( D_sq && ( (int)D_sq->size1 != num_design_cols || (int)D_sq->size2 != num_design_cols ) )
      {
         gsl_matrix_free( D_sq );
         D_sq = 0;
      }
      if( !D_sq )
         D_sq = gsl_matrix_calloc( num_design_cols, num_design_cols);
      int num_d_rows_used = regression_n, num_repeats = 1;
      if( inputs_repeated_every > 0 && ( !C1 || inputs_repeated_every == (int)C1->size1 ) )
      {
         if( !data_block_sizes || data_block_sizes_sum != inputs_repeated_every || treat_experiments_as_different_dimensions )
            num_d_rows_used = inputs_repeated_every;
         else
            num_d_rows_used = data_block_sizes[ bs ];
         num_repeats = regression_n / inputs_repeated_every;
      }
      for( int i = 0; i < num_design_cols; i++ )
      {
         double dsq_ii = 0;
         for( int i2 = start_row; i2 < start_row + num_d_rows_used; i2++ )
         {
            double d_i2i = gsl_matrix_get( D, i2, i );
            dsq_ii += d_i2i * d_i2i;
         }
         gsl_matrix_set( D_sq, i, i, dsq_ii * num_repeats );
         if( !gram_schmidt )
            for( int j = (i+1); j < num_design_cols; j++ )
            {
               double dsq_ij = 0;
               for( int i2 = start_row; i2 < start_row + num_d_rows_used; i2++ )
                  if( !C1D )
                     dsq_ij += gsl_matrix_get( D, i2, i ) * gsl_matrix_get( D, i2, j );
                  else
                     dsq_ij += gsl_matrix_get( C1D, i2, i ) * gsl_matrix_get( C1D, i2, j );
               dsq_ij *= num_repeats;
               gsl_matrix_set( D_sq, i, j, dsq_ij );
               gsl_matrix_set( D_sq, j, i, dsq_ij );
            }
      }
      if( D_sqs )
      {
         D_sqs[ bs ] = D_sq;
         if( bs < num_data_block_sizes - 1 )
         {
            start_row += data_block_sizes[ bs ];
            D_sq = 0;
         }
         else
            D_sq = D_sqs[ 0 ];
      }
   }
}

void GLM_MVN::calculate_translated_D()
{
   num_design_cols = D->size2;
   if( !C1D )
      C1D = gsl_matrix_alloc( n, num_design_cols );
   double temp2;
   int size_C1 = C1->size1; //or cols, matrix is square.
   
   int num_blocks = n / size_C1;
   for( int block = 0; block < num_blocks; block++ )
      for( int i = 0; i < size_C1; i++ )
         for( int j = 0; j < num_design_cols; j++ )
         {
            temp2 = 0;
            for( int i2 = 0; i2 < size_C1; i2++ )
               temp2 += gsl_matrix_get( C1, i, i2 ) * gsl_matrix_get( D, i2 + block * size_C1, j );
            gsl_matrix_set( C1D, i + block * size_C1, j, temp2 );
         }

//  if( !C1D_sq )  /// Use D_sq for either case C1 or !C1
//    C1D_sq = new matrix( num_design_cols, num_design_cols);
//  for( int i = 0; i < num_design_cols; i++ )
//   for( int j = 0; j < num_design_cols; j++ )
//      {
//	(*C1D_sq)( i, j ) = 0;
//	for( int i2 = 0; i2 < size_C1; i2++ )
//	  (*C1D_sq)( i, j ) += (*C1D)( i2, i ) * (*C1D)( i2, j );
//      }
//  (*C1D_sq) *= (double)num_blocks;
}

void GLM_MVN::update_D_dependent_matrices()
{
   num_design_cols = D->size2;
   dim_V = num_design_cols;
   if( correlate_experiments )
      dim_V *= num_experiments;
   if( C1D )
      calculate_translated_D();
   if( D_sq )
      calculate_D_sq();
   for( int v_num = 0; v_num < number_of_vs; v_num++ )
   {
      set_v_index( v_num );
      if( V_inv )
      {
	 if( (int)V_inv->size1 != dim_V )
	 {
	    gsl_matrix_free( V_inv );
	    V_inv_array[ v_num ] = V_inv = gsl_matrix_alloc( dim_V, dim_V );
	 }
	 gsl_matrix_set_identity( V_inv );
	 gsl_matrix_scale( V_inv, 1.0 / v_multiplier );
      }
      half_logdet_v_inv[ v_num ] = -.5 * p_y * dim_V * log( v_multiplier );
   }
   if( beta )
   {
      gsl_matrix_free( beta );
      beta = gsl_matrix_alloc( dim_V, 1 );
   }
//   if( V_star_sqrt )
//   {
//      for( int i = 0; i < dim_V; i++ )
//         for( int j = 0; j < dim_V; j++ )
//            gsl_matrix_set( V_star_sqrt, i, j, gsl_matrix_get( V_star, i, j ) );
//      gsl_linalg_cholesky_decomp( V_star_sqrt );
//   }
   calculate_ml_const();
}

void GLM_MVN::calculate_translated_Y()
{
   if( !C1Y )
      C1Y = gsl_matrix_alloc( n, p_y );
   if( (int)C1->size2 != n )
   {
      double temp2;
      int num_blocks = n / correlation_block;
      for( int block = 0; block < num_blocks; block++ )
         for( int i = 0; i < correlation_block; i++ )
            for( int j = 0; j < p_y; j++ )
	    {
               temp2 = 0;
               for( int i2 = 0; i2 < correlation_block; i2++ )
                  temp2 += gsl_matrix_get( C1, i, i2 ) * gsl_matrix_get( Y, i2 + block * correlation_block, j );
               gsl_matrix_set( C1Y, i + block * correlation_block, j, temp2 );
	    }
   }
   else
      for( int i = 0; i < n; i++ )
         for( int j = 0; j < p_y; j++ )
         {
            double c1y_ij = 0;
            for( int k = 0; k < correlation_block; k++ )
               c1y_ij += gsl_matrix_get( C1, i, k ) * gsl_matrix_get( Y, k, j );
            gsl_matrix_set( C1Y, i, j, c1y_ij );
         }
}

void GLM_MVN::calculate_beta_hat_matrix()
{
   if( !beta_hat_mx )
      beta_hat_mx = gsl_matrix_alloc( regression_num_data_blocks * p_y, num_design_cols );
   if( !D_sq )
      calculate_D_sq();

   gsl_matrix* inv_D_sq = NULL;
   gsl_matrix* inv_D_sq_D = NULL;
   gsl_matrix **inv_D_sqs = NULL, **inv_D_sq_Ds = NULL;
   if( data_block_sizes )
   {
      inv_D_sqs = new gsl_matrix*[ num_data_block_sizes ];
      inv_D_sq_Ds = new gsl_matrix*[ num_data_block_sizes ];
   }

   int i, j, i2, bs;
   int cum_dbs = 0;
   for( bs = 0; bs < num_data_block_sizes; bs++ )
   {
      get_experiment_number( bs );
      if( D_sqs )
	 D_sq = D_sqs[ bs ];
      inv_D_sq = gsl_matrix_alloc( D_sq->size1, D_sq->size2 );
      inv_D_sq_D = gsl_matrix_alloc( num_design_cols, data_block_size );
      if( data_block_sizes )
      {
	 inv_D_sqs[ bs ] = inv_D_sq;
	 inv_D_sq_Ds[ bs ] = inv_D_sq_D;
      }
      my_invert( D_sq, inv_D_sq );
      if( inputs_repeated_every > 0 && ( !C1 || inputs_repeated_every == (int)C1->size1 ) )
	 gsl_matrix_scale( inv_D_sq, num_data_blocks / num_data_block_sizes );
      for( i = 0; i < num_design_cols; i++ )
	for( j = 0; j < data_block_size; j++ )
	{
	   double temp = 0;
	   for( i2 = 0; i2< num_design_cols; i2++ )
	      temp += gsl_matrix_get( inv_D_sq, i, i2 ) * gsl_matrix_get( D, cum_dbs + j, i2 );
	   gsl_matrix_set( inv_D_sq_D, i, j, temp );
	}
      cum_dbs += data_block_size;
   }

   int counter = 0;
   for( i = 0; i < regression_num_data_blocks; i++ )
   {
      if( data_block_sizes )
	 inv_D_sq_D = inv_D_sq_Ds[ get_experiment_number( i ) ];
      for( int ycol = 0; ycol < p_y; ycol++ )
         for( j = 0; j < num_design_cols; j++ )
         {
            double temp = 0;
            for( i2 = 0; i2< data_block_size; i2++ )
                temp += gsl_matrix_get( Y, counter + i2, ycol ) * gsl_matrix_get( inv_D_sq_D, j, i2 );
         //	  temp += gsl_matrix_get( Y, i * data_block_size + i2, ycol ) * gsl_matrix_get( D->gslobj(), j, i2 );
            gsl_matrix_set( beta_hat_mx, i * p_y + ycol, j, temp );
         }
      counter += data_block_size;
   }
   for( bs = 0; bs < num_data_block_sizes; bs++ )
   {
      if( data_block_sizes )
      {
	 inv_D_sq = inv_D_sqs[ bs ];
	 inv_D_sq_D = inv_D_sq_Ds[ bs ];
      }
      gsl_matrix_free( inv_D_sq );
      gsl_matrix_free( inv_D_sq_D );
   }
   if( data_block_sizes )
   {
      zap_vector( inv_D_sqs );
      zap_vector( inv_D_sq_Ds );
      get_experiment_number( 0 );
      if( D_sqs )
         D_sq = D_sqs[ 0 ];
   }
}

double GLM_MVN::function_variance_evaluate( double* v )
{
   if( a_star <= 1 )
      return 0;
   create_design_point( v );
   double mvm = 0, mvm_off_diagonal = 0;
   for( int i = 0; i < design_point_dim; i++ )
   {
       mvm += design_point[ i ] * gsl_matrix_get( V_star, i, i ) * design_point[ i ];
       for( int j = 0; j < i; j++ )
          mvm_off_diagonal += design_point[ i ] * gsl_matrix_get( V_star, i, j ) * design_point[ j ];
   }
   mvm += 2 * mvm_off_diagonal;
   return mvm * e_sigma_sq;
}

void GLM_MVN::output_residual_variogram( gsl_matrix* x_diffs, gsl_matrix* residual_diffs )
{
   ofstream VarioStream( Variogram_Filename, ios::out );
   if ( !VarioStream )
   {
      cout << "ERROR: Couldn't open residual variogram output file"<< endl;
      exit( 1 );
   }
   long ire_store = get_inputs_repeated_every();
   long dbs_store = data_block_size;
   data_block_size = 1;
   GLM_MVN G(x_diffs, residual_diffs);
   G.set_inputs_repeated_every( 0 );
   G.create_full_1way_model();
   G.calculate_D_sq();
   if( G.V_inv )
      gsl_matrix_free( G.V_inv );//dimension probably doesn't match
   G.calculate_log_posterior();
   int num_vbls = G.get_number_of_design_variables();
   gsl_matrix* mean = G.get_m_star();
   for( int i = 0; i < num_vbls; i++ )
      ( G.get_Basis_Fn_ptr( i ) )->set_coeff( gsl_matrix_get( mean, i, 0 ) );
   G.make_grid_predictions( 0 );
   G.write_predictions( &VarioStream );
   G.set_inputs_repeated_every( ire_store );
   data_block_size = dbs_store;
}

double GLM_MVN::monte_carlo_estimate_l2_distance_from_zero()
{
   if( number_monte_carlo_samples < 2 )
      return average_function_squared_at_bases_evaluate();
   ofstream L2Stream( L2Distance_Filename, ios::out );
   double sum_l2 = 0;
   double l2_i;
   for( int i = 0; i < number_monte_carlo_samples; i++ )
   {
      draw_betas_from_posterior();
      l2_i = average_function_squared_at_bases_evaluate();
      if( strlen( L2Distance_Filename ) )
         L2Stream << l2_i << endl;
      sum_l2 += l2_i;
   }
   return sum_l2 / number_monte_carlo_samples;
}

double GLM_MVN::monte_carlo_estimate_absolute_integral()
{
   double sum_ia = 0;
   double* v = new double[ p ];
   for( int i = 0; i < p; i++)
      v[ i ] = 1;
   if( number_monte_carlo_samples < 2 )
      sum_ia = integral_absolute_function_evaluate( v, 1 );
   else
   {
      ofstream AbsIntegralStream( Integral_Absolute_Filename, ios::out );
      double ia_i;
      for( int i = 0; i < number_monte_carlo_samples; i++ )
      {
         draw_betas_from_posterior();
         ia_i = integral_absolute_function_evaluate( v, 1 );
         if( strlen( Integral_Absolute_Filename ) )
            AbsIntegralStream << ia_i << endl;
         sum_ia += ia_i;
      }
   }
   delete [] v;
   return sum_ia / number_monte_carlo_samples;
}

void GLM_MVN::monte_carlo_estimate_l2_distance_from_zero_and_absolute_integral( double& ave_l2, double& ave_ia )
{
   double* v = new double[ p ];
   for( int j = 0; j < p; j++)
      v[ j ] = ( normalise_predictors == 1 ) ? 1 : x_max[ j ];
   if( number_monte_carlo_samples < 2 )
   {
      ave_l2 = average_function_squared_at_bases_evaluate();
      ave_ia = integral_absolute_function_evaluate( v, 1 );
   }
   else
   {
      ave_l2 = ave_ia = 0;
      double l2_i, ia_i;
      ofstream L2Stream( L2Distance_Filename, ios::out );
      ofstream AbsIntegralStream( Integral_Absolute_Filename, ios::out );
      for( int i = 0; i < number_monte_carlo_samples; i++ )
      {
         draw_betas_from_posterior();
         l2_i = average_function_squared_at_bases_evaluate();
         if( strlen( L2Distance_Filename ) )
            L2Stream << l2_i << endl;
         ave_l2 += l2_i;
         ia_i = integral_absolute_function_evaluate( v, 1 );
         if( strlen( Integral_Absolute_Filename ) )
            AbsIntegralStream << ia_i << endl;
         ave_ia += ia_i;
      }
      ave_l2 /= number_monte_carlo_samples;
      ave_ia /= number_monte_carlo_samples;
   }
   delete [] v;
}

void GLM_MVN::find_correlated_experiments_covariances()
{
   if( V_inv && ( (int)V_inv->size1 != dim_V || (int)V_inv->size2 != dim_V ) )
   {
      gsl_matrix_free( V_inv );
      V_inv = 0;
   }
   if( !V_inv )
      V_inv = gsl_matrix_alloc( dim_V, dim_V );
   if( !experiment_location_dimension )
   {
      for( int i = 1; i < num_experiments; i++ )
         square_matrix_stripe( V_, i * num_design_cols, v_multiplier * .5 );
      my_invert( V_, V_inv );
   }
   else
      update_correlated_experiments_covariances();
}

void GLM_MVN::update_correlated_experiments_covariances()
{
   get_experiment_covariances();
   if( correlate_experiments )
   {
      my_invert( experiment_covariances, experiment_precisions );
      for( int i = 0; i < num_experiments; i++ )
         for( int j = i; j < num_experiments; j++ )
         {
            double v_ij = gsl_matrix_get( experiment_covariances, i, j ) * v_multiplier;
            double v_inv_ij = gsl_matrix_get( experiment_precisions, i, j ) / v_multiplier;
            matrix_square_symmetric_strip( V_, i * num_design_cols, j * num_design_cols, num_design_cols, v_ij );
            matrix_square_symmetric_strip( V_inv, i * num_design_cols, j * num_design_cols, num_design_cols, v_inv_ij );
         }
//      my_invert( V_, V_inv );
   }
   else if( correlate_columns )
   {
      for( int i = 0; i < p_y; i++ )
         for( int j = 0; j < p_y; j++ )
            gsl_matrix_set( b_matrix, i, j, b_ * gsl_matrix_get( experiment_covariances, i, j ) );
   }
}

void GLM_MVN::update_correlated_experiments_covariances( int exp_number )
{
   int i, j;
   if( correlate_experiments )
   {
      if( !kriging_experiment_correlation )
      {
         update_experiment_precisions( exp_number );
         my_invert( experiment_covariances, experiment_precisions );
         for( i = 0; i < num_experiments; i++ )
            for( j = i; j < num_experiments; j++ )
            {
               double v_ij = gsl_matrix_get( experiment_covariances, i, j ) * v_multiplier;
               double v_inv_ij = gsl_matrix_get( experiment_precisions, i, j ) / v_multiplier;
               matrix_square_symmetric_strip( V_, i * num_design_cols, j * num_design_cols, num_design_cols, v_ij );
               matrix_square_symmetric_strip( V_inv, i * num_design_cols, j * num_design_cols, num_design_cols, v_inv_ij );
            }
      }
      else
      {
         for( i = 0; i < num_experiments; i++ )
         {
            double v_ij = get_experiment_covariance( i, exp_number ) * v_multiplier;
            matrix_square_symmetric_strip( V_, i * num_design_cols, exp_number * num_design_cols, num_design_cols, v_ij );
         }
         my_invert( experiment_covariances, experiment_precisions );
         for( i = 0; i < num_experiments; i++ )
            for( j = i; j < num_experiments; j++ )
            {
               double v_inv_ij = gsl_matrix_get( experiment_precisions, i, j ) / v_multiplier;
               matrix_square_symmetric_strip( V_inv, i * num_design_cols, j * num_design_cols, num_design_cols, v_inv_ij );
            }

      }
//      my_invert( V_, V_inv );
   }
   else if( correlate_columns )
   {
      get_experiment_covariances();
      for( i = 0; i < p_y; i++ )
         for( j = 0; j < p_y; j++ )
            gsl_matrix_set( b_matrix, i, j, b_ * gsl_matrix_get( experiment_covariances, i, j ) );
   }
}

void GLM_MVN::calculate_global_error_covariances()
{
   if( global_error_covariances )
      gsl_matrix_free( global_error_covariances );
   global_error_covariances = get_target_covariances( false, (bool)residuals );
   error_cov_dim = global_error_covariances->size1;//or size2
   //if using as a prior mean:
   scaled_error_covariances_prior_mean = gsl_matrix_alloc( error_cov_dim, error_cov_dim );
   gsl_matrix_memcpy( scaled_error_covariances_prior_mean, global_error_covariances );
   gsl_matrix_scale( scaled_error_covariances_prior_mean, error_cov_dim + 1 );
}
