
#if !defined( _glm_mvn_mcmc_clustering_h )
#define _glm_mvn_mcmc_clustering_h

#include <set>
#include "glm_mvn_clustering.h"

class GLM_MVN_MCMC_Clustering : public GLM_MVN_Clustering
{
  public:
   ///
   GLM_MVN_MCMC_Clustering( int argc, char **argv );
   ///
   GLM_MVN_MCMC_Clustering( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   GLM_MVN_MCMC_Clustering( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~GLM_MVN_MCMC_Clustering();
   ///
   void initialise_glm_mvn_mcmc_clustering();

   virtual void add_post_burn_in_moves();

   virtual void update_any_stats();//overwritten fn to reset the cluster_deque after merge_until() has been used.

   virtual void ModelPrintOut();//overwritten fn outputting model details to screen.

   virtual void print_model_results();// overwritten fn, prints out the final results specific to model used.

   virtual void initialise_ergodic_averages();//overwritten fn initialising model averages

   virtual void update_ergodic_averages();//overwritten fn updating model averages

   virtual void reset_chain();//overwritten fn resetting the starting values of the MCMC.

   void end_of_chain_processing();//overwritten fn updating at end of the MCMC chain.

   void set_starting_clusters();//to choose what starting cluster configuration.

   void get_subject_cluster_types();//overwritten fn now giving the subject cluster types as their best_clustering memberships if find_map and changing_clustering_in_mcmc.

   void calculate_stirling_numbers();//initialsie the array stirling.

   virtual void move(); // Shuffles the cluster list, randomly chooses from a shift, a birth or a death depedning on what's possible.
   
   void propose_move();// Proposes the chosen move type.

   void execute_move();// Carries out the proposed move (used if move accepted).

   void undo_move();// Cancels the proposed move (used if move not accepted).

   void calculate_number_mergeable_splitable();//finds the number of clusters available for merge/split.

   void find_mate();//used for merge and shift moves, finds a proposed_iter2 to merge/take subjects from *proposed_iter.

   bool deque_iterators_mergeable();//checks whether the the proposed_iter iterators for the cluster_deque are ok for merging.

   bool deque_iterators_mergeable( cluster_ptrs::iterator iter, cluster_ptrs::iterator iter2 );

   void find_cluster_to_alter();//chooses a cluster to split/merge from mergeable_cluster_deque.

   double calculate_tree_score();//when preserving tree, this function calculates the tree score for the current model.

   double add_up_tree_score();//adds up the score for the current tree.

   int is_tree_above_model_optimal( Cluster_GLM_MVN* cl1 = NULL, Cluster_GLM_MVN* cl2 = NULL );//returns 1 if the tree above the clusters in the cluster map is the same as would be created by performing agglomerative clustering on them. If cl1 is non null but cl2 null, replace cl1 with its children; if both non null replace both with their shared parent.

   void calculate_tree_score_normalising_constant( Cluster_GLM_MVN* cl1 = NULL, Cluster_GLM_MVN* cl2 = NULL );//Calculates the normalising constant of the proposed model when using the tree 'score' distribuion. If cl1 is non null but cl2 null, replace cl1 with its childrens cluster sizes; if both non null replace both with their shared parent's cluster size.

   double get_r_k( int k, double* catalans, int size_catalans );// returns the number of trees which would fit the cluster sizes with corresonding catalans and match k optimal agglomerative clustering subtrees.

   double get_r_k( int k, double* catalans, int size_catalans, int* chosen_indices, int index = 0, int start_index = 0 );

   void fit_optimal_tree_around_model();//repace tree with the optimal subtrees T_i* fit around the current model.

   void sample_tree_from_full_conditionals();//for each  subtree, draws a new subtree from its full conditional distribution.

   void sample_subtree_from_full_conditional();//for a randomly chosen subtree, draws a new subtree from its full conditional distribution.
   
   void sample_subtree_from_full_conditional( int tree_number );//for subtree tree_number, draws a new subtree from its full conditional distribution.

   void sample_subtree_above_model_nodes_from_full_conditional( Cluster_GLM_MVN* cl );//draws a new subtree above the model nodes (the model containing cl in case multiple trees) from its full conditional distribution.
   
   void find_proposed_tree_score();//works out the new score for the tree under proposed move.

   bool find_best_sibling_pair_in_set( set<Cluster_GLM_MVN*> cluster_list, Cluster_GLM_MVN*& cl, Cluster_GLM_MVN*& cl_sibling );//returns a boolean reporting whether search was successful.

   void calculate_jump_prob();//calculates the probability of the split move if jump_distn != 0.

   void write_tree_merger_details();//if preserving tree, takes tree and writes it's merger details to the Merger_Details_Filename.

   void reset_experiment_location_parameters();//overwriting the function from Regression class, in case parameters are being integrated out.
   
  private:

   int MERGE, SPLIT, SUBJECT_SHIFT, SINGLE_SUBJECT_SHIFT, SINGLETON_MERGE, SINGLETON_SPLIT, V_SHIFT, PRIOR_V_SHIFT, EXPERIMENT_SHIFT, EXPERIMENT_SIGN_CHANGE, EXPERIMENT_MULTIPLIER_SHIFT, EXPERIMENT_POWER_SHIFT, EXPERIMENT_PARAMETER_SHIFT, TREE_FULL_CONDITIONAL;//constants to be enumerated (dynamically) to indicate the different move types.

   double* stirling; // array holding the log-Striling numbers for num_mergeable_subejcts.

   double** stirling_mx; // matrix with row i holding the log-Striling numbers for i=1-num_mergeable_subejcts.

   int dim_stirling_mx;//number of rows/cols of stirling_mx.
   
   long starting_configuration;//if =0, we start with all subjects in 1 cluster, if 2 all in individual clusters, if =2 maximise BIC first.

   long random_number_of_clusters;//non zero if we allow the number of clusters to change during the MCMC

   long subject_shift_move_included;//non zero if the shifting subjects between clusters move is to be used.

   long single_subject_shift_move_included;//non zero if the shifting a single subject between clusters move is to be used.

   bool shift_clusters_below_model;//if non zero and preserving tree, only shift clusters which lie underneath the current model.

   long singleton_merge_split_moves_included;//non zero if the severing of a single subject into its own new cluster and reverse move are to be used.

   bool changing_clustering_in_mcmc;//true if one of the move types in mcmc changes cluster composition.

   bool v_shift_move_included;//non zero if the prior variance multiplier v_multiplier can be shifted.

   bool prior_v_shift_move_included;//non zero if the prior probabiilites assigned to a mixture of V_s are to be treated as random.

   bool experiment_shift_move_included;//non zero if the experiment locations (abstract locations to induce correlations between experiments) are to be moved.

   long tree_full_conditional_move_included;//non-zero if want to draw subtrees from their full conditional distribution - set 1 if do tree above model before merge and below node before split, 2 if to be done as a whole, separate independent move type.

   double prob_full_conditional;//if tree_full_conditional_move_included, with this probability sample the relevant subtree before merge or split.

   bool metropolis_tree_changes;//if non-zero, still accept tree deformations which don't affect the current model with the usual metropolis acceptance probability.

   bool agglomerative_based_split;//if non-zero, split a cluster according to agglomerative-clustering based rule
   
   long jump_distn;//if zero choose clusters to merge uniformly, if 1 choose according to suitability (only approximately coded in terms of the calculated acceptance rate, which might matter if n small), if -1 choose most suitable (for deterministic use only).

   cluster_ptrs::iterator proposed_iter;//iterator for a pointer to the cluster which move is proposed for (cluster_ptrs defined in previous class).

   cluster_ptrs::iterator proposed_iter2;//iterator fora second such pointer if needed (e.g merge move).

   Cluster_GLM_MVN *proposed_cluster, *proposed_cluster2;//cluster pointers correpsonding to the above two iterators.

   int proposed_subject;//proposed subject to move from one cluster to another.

   int proposed_v_num, proposed_v_num2;//if have a mixture of vs, use these as random indices for altering them.

   bool iter_distinct;//true if proposed_iter cluster is 'distinct'.

   bool iter_splitable;//true if proposed_iter cluster is 'splitable'.

   int number_mergeable, number_splitable;//count the number of clusters available for merge/split.

   string* proposed_iter_label;//label for the cluster pointed to by proposed_iter.

   Cluster_GLM_MVN* holding_cluster;//a cluster which holds say the newly formed cluster from a split whilst it's fate is decided.

   double v_shift_size;

   double prior_v_shift_size;

   double log_prior_v_store, log_prior_v_store2;//variables for holding old values of log_prior_v[ proposed_v_num ] and log_prior_v[ proposed_v_num2 ].

   double ave_v_multiplier;//for calculating average v_multiplier value over mcmc.

   double best_v_multiplier;//for holding the best values of v_multiplier.

   double best_experiment_closeness_power;//for holding the best values of experiment_closeness_power.

   double* ave_prior_v_probs;//for calculating average exp(log_prior_v) values over mcmc.

   double* best_log_prior_v;//array for holding the best values of log_prior_v found.

   int best_num_clusters;

   double experiment_shift_size;

   int proposed_experiment_index, proposed_experiment_index2;

   int proposed_experiment_group;
   
   double experiment_power_shift_size;
   
   double* experiment_location_store;//when proposing a new location for experiment proposed_experiment_index, store the old one here.

   double* experiment_location_parameter_store;//when proposing a new location parameter for experiment group proposed_experiment_group, store the old one here.

   bool integrate_out_experiment_parameters;//set true if experiment location means to be integrated out and not sampled.
   
   double* experiment_location_sum_store;//when proposing a new location for experiment proposed_experiment_index, store the corresponding sum here.
   
   double* experiment_location_shrinkages;// array for storing shrinkage factors for posterior means of abstract spatial locations of experiments for each group in a partition of the experiments.

   double* experiment_location_mean_shrinkages;// array for storing shrinkage factors for posterior means of means of abstract spatial locations of experiments for each group in a partition of the experiments.

   double* experiment_location_posterior_precisions;// array for storing posterior variances of individual abstract spatial locations of experiments for each group in a partition of the experiments.

   double* experiment_location_sum_posterior_precisions;// array for storing posterior variances of sum of abstract spatial locations of experiments for each group in a partition of the experiments.
   
   int* experiment_signs_store;//when proposing new sign for experiment proposed_experiment_index, store the old one here.

   double experiment_var_multiplier_store;//when proposing new variance for experiment proposed_experiment_index, store the old one here.
   
   double experiment_closeness_power_store;//when proposing a new power of distance between locations used for calculating experiment correlation, store the old one here.

   double experiment_variance_likelihood, proposed_experiment_variance_likelihood, proposed_experiment_variance_likelihood_diff, proposed_experiment_variance_likelihood_diff2;//used for keeping track of the posterior contribution from the distributions on the experiment variances.

   double num_configurations, proposed_num_configurations;//number of possible configurations of tree.

   int tree_above_model_score, proposed_tree_above_model_score;//1 if the tree joining the nodes in the cluster map up to the root(s) is the tree that would result from agglomerative hierarchical clustering, 0 otherwise.

   int proposed_cluster_score, proposed_cluster2_score;//variables for holding proposed cluster scores.

};

double log_gamma_density_ratio( const double x_new, const double x_old, const double shape, const double scale );//function for calculating f(x_new)/f(x_old) where f is the density for the gamma(shape,scale) distribution.

#endif
