
#if !defined( _mcmc_h )
#define _mcmc_h

#include "regression.h"
#include <map>
#include <time.h>

#define LOG_TWO 0.6931472
#define LOG_THREE 1.098612

class MCMC : public Regression
{
  public:
   ///
   MCMC( int argc, char **argv );
   ///
   MCMC( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   MCMC( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~MCMC();
   ///
   void initialise();

   virtual double calculate_log_likelihood() = 0;

   virtual double calculate_log_prior();

   virtual void revise_log_posterior(){;}//Compute proposed_log_posterior from log_posterior, perhaps without recalculating everything.

   void propose_a_move();//Propose a candidate MH move, calling move().

   void accept_or_reject_move();//Decide whether to accept MH move, and update model.
   
   virtual void move(); //Generates a candidate MH move.

   void sampler(); // Draws a MH sample from the posterior

   virtual void update_posterior_maximizing_model(){;};//if maximum_log_posterior increases, this virtual function can be overwritten for updating the best model kept.

   virtual void reset_chain(){;}; //Resets the starting values of the MCMC chain when performing multiple runs.

   virtual void end_of_chain_processing(){;}//Does and updating/processing at the end of the MCMC chain.

   bool get_accept(){ return accept; } // Calculates accpetance ratio and draws uniform(0,1) to see if accept/reject.

   //   virtual void UpdateModelAverages() = 0; // Virtual function, updates the averages of the model parameters.
   void print_acceptance_rates();
   
   virtual void ModelPrintOut(){;}// print out fns of interest of the model.

   virtual void print_model_results();// Virtual function, prints out the final results specific to model used.

   void print_out_every_output();// Output printed to screen every PrintOutEvery iterations.

   void print_logacceptanceratio_warning();//prints warning to screen if LogAcceptanceRatio is not a number.
   
   virtual void write_predictions( ofstream *any_stream = NULL );//creates the predictions_mx and writes it to any_stream or by default to PredictionFilename.   
   
   virtual void initialise_ergodic_averages(){;}//initialises the ergodic averages (to zero) for specific model.

   virtual void update_ergodic_averages(){;} // calcuate ergodic averages for specific model.

   void base_update_ergodic_averages();// calcuate ergodic averages for base model.

   void update_average_predictions_mx();// update the ave_predictions_mx by adding predictions mx in the relevant columns.

   void normalize_average_predictions_mx();// For relevant columns of ave_predictions_mx, divide by prediction_count.

   virtual void add_post_burn_in_moves(){;}//adds any moves to be used after the burn in.
   
   int add_move();// add a new move and return the move number it has been assigned.

   int add_move( string new_label);// add a new move with label new_label, and return the move number it has been assigned.

   void add_number_of_moves( int add_number_of_moves );

   void set_number_of_moves( int new_number_of_moves );

   double get_number_of_moves(){ return number_of_moves; }

   void knock_out_move( int move_type_out );// adjust the proposal cdf so that move_type_out has probability 0.

   void draw_move_type();//randomly draw a move type from proposal_cdf.

   virtual void execute_move(){;}//carry the propose move through

   virtual void undo_move(){;}//cancel the proposed move.
   
   virtual void post_move_updates(){;}// Carries out any updates required after the move, accepted or not.

   void set_move_type_label( int mt, string s ){ move_type_labels[ mt ] = s; }

   string get_move_type_label(){ return get_move_type_label( move_type ); }

   string get_move_type_label( int mt );

   void print_move_type(){ cout << get_move_type_label() << endl; }

   string get_alligned_move_type_label(){ return get_alligned_move_type_label( move_type ); }
   
   string get_alligned_move_type_label( int mt );//gets the label for move type mt, alligned for neat output.
   
   double cooling_schedule();//caclulates the cooling schedule for simulated annealing using global variable iteration.

   double cooling_schedule( int it, int max_it );//caclulates the cooling schedule for simulated annealing

   static int get_iteration(){ return iteration; }

   bool is_maximizing_run(){ return posterior_maximizing_run == chain_run; }//return true if maximum log posterior has been increased in this run.

   bool accept_since_iteration( int& reference_iteration );//returns true if a move has been accepted since iteration reference_iteration.

   bool past_burn_in(){ return ( iteration > burn_in ); }

   bool first_iteration_since_burn_in(){ return ( iteration == burn_in + 1); }
   
   bool last_chain_run(){ return ( chain_run == number_of_runs ); }

   bool multiple_runs(){ return number_of_runs > 1; }
   
   double get_current_ergodic_normaliser(){ return static_cast<double>( iteration - burn_in ) / static_cast<double>( sample_every );}
  private:
   
   static long samples;

   static int iteration;
   ///
   static long sample_every;
   ///
   static long burn_in;
   ///
   long int PrintOutEvery;
   ///
   int mixing_print_out_every;//when printing mcmc mixing details to file, do so after every this many samples.

   int number_of_moves;

   int number_of_moves_used; // <= to number_of_moves, this many have non-zero probability.

   int max_label_size; // the largest size of the strings describing the move types, used for allignment.

   bool accept;

   double old_log_likelihood;

   double old_log_posterior;

   double old_log_prior;
   ///

   double log_posterior_ratio, average_log_posterior_ratio;//average of log(proposed_posterior/posterior).
   
   double average_log_proposal_ratio;//average of log(pr(proposing current model)/pr(proposing proposed model)).

   double average_proposed_log_posterior;//Average value of proposed log posterior in MCMC.

   double average_acceptance_prob;//average of log acceptance probability for MH.
   
   double simulated_annealing_parameter;//scale parameter for cooling schedule for simulated annealing.
   
   double simulated_annealing_parameter2;//shift parameter for cooling schedule for simulated annealing.
   
  protected:

   int move_type, original_move_type;

   bool force_accept, force_reject;//used only when we want to force a move to be accepted or rejected.

   bool posterior_cancels_with_proposal;//set true if the change in posterior is not to be automatically included in calculating the MH acceptance probability.

   bool gibbs_sampling;//set non-zero if we wish to use gibbs sampler.

   bool finished_mcmc_sampling;//set non-zero when sampling has completed.

   bool simulated_annealing;
   
   double proposed_log_likelihood, max_log_likelihood;

   double proposed_log_prior, proposed_log_posterior;

   double log_auxiliary_posterior, proposed_log_auxiliary_posterior;//additional posterior terms when using auxiliary variables

   double previous_log_posterior;

   double proposed_BIC, max_BIC;

   double initial_log_posterior, maximum_log_posterior;

   double maximum_run_posterior;//holds either the maximum posterior for that chain, or the maximum of a single calculation done once for each chain, according to usage.

   int posterior_maximizing_iteration;

   int posterior_maximizing_run;

   int last_iteration_accepted;
      
   double ergodic_ave_normaliser, run_ergodic_ave_normaliser;

   long number_of_runs;
   
   static int chain_run;
   
   static bool hill_climbing;// if non-zero, then only accept a move if log(proposalratio) > 0.

   static bool model_averaging;// if zero, then report posterior modal statistics rather than means.

   static bool find_map;//if non zero, then report the maximum a posteriori model.

   bool automatically_calculate_residuals;//if true, then after MCMC find residuals of data from mean - turn off for clustering models, where have different models for different observations and need to be more careful.
   
   double** ave_predictions_mx;// matrix to hold predictions to be written to PredictionFilename.

   int ave_predictions_mx_cols;// number of columns in ave_predictions_mx.
   
   int prediction_count;//number of times predictions are made and added to ave_predictions_mx;

   bool predict_from_model;//if true, the current model can be used for prediction (useful in subsampling with constraints).

   double* proposal_cdf;//gsl::vector of length number_of_moves holding the move probabilities.
   
   double* TotalAcceptanceRB;//gsl::vector of length number_of_moves holding the total Rao Blackwellised acceptances for each move_type.

   int* TotalAcceptance;//gsl::vector of length number_of_moves holding the total number of acceptances for each move_type.

   int* TotalIts;//gsl::vector of length number_of_moves holding the total number of visits to each move type.
   
   map< int, string, std::less<int> > move_type_labels;//map taking move_type to its name.

   list< int > trans_dimensional_move_types;//list containing the move types which switch dimensions of the model.
   
   double LogProposalRatio; // Contribution to acceptance ratio of proposal.
   
   double LogAcceptanceRatio;
   
   char Log_Posterior_Filename[FILENAMELENGTH];

   clock_t time, iteration_time;
};

#endif
