#include "mcmc_basis_fn_regression.h"

MCMC_Basis_Fn_Regression::MCMC_Basis_Fn_Regression( int argc, char **argv )
      : GLM_MVN( argc, argv )
{
   v_shift_move_included = 0;
   rw_on_basis_coeffs = true;
   basis_coeff_step_size = .01 * sqrt( v_multiplier );
   random_basis_fns = true;
   for( int i = 1; i < argc; i++ )
   {
      if ( strstr( argv[i], "help" ) )
      {
//          cout << "randombasisfns=" << random_basis_fns << endl;
//          cout << "vshiftmoveincluded=" << v_shift_move_included << endl;
         
//          cout << "basiscoeffstepsize=" << basis_coeff_step_size << endl;
         return;
      }
      if( InitialStringMatch( argv[i], "randombasisfns=" ) )
         StringToBool( &random_basis_fns, argv[i] );
      if( InitialStringMatch( argv[i], "vshiftmoveincluded=" ) )
         StringToBool( &v_shift_move_included, argv[i] );
      if( InitialStringMatch( argv[i], "basiscoeffstepsize=" ) )
         StringToDouble( &basis_coeff_step_size, argv[i] );
   }
   initialise();
}

MCMC_Basis_Fn_Regression::MCMC_Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1 )
        : GLM_MVN( X1, Y1 )
{
   initialise();
}

MCMC_Basis_Fn_Regression::MCMC_Basis_Fn_Regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 )
        : GLM_MVN( X1, Y1, D1 )
{
   initialise();
}

MCMC_Basis_Fn_Regression::~MCMC_Basis_Fn_Regression()
{
   if( previous_m_star && previous_m_star != m_star )
      gsl_matrix_free( previous_m_star );
   if( previous_V_star && previous_V_star != V_star )
      gsl_matrix_free( previous_V_star );
   if( previous_V_star_sqrt && previous_V_star_sqrt != V_star_sqrt )
      gsl_matrix_free( previous_V_star_sqrt );
   if( previous_b_star && previous_b_star != b_star )
      gsl_matrix_free( previous_b_star );
   zap_vector( num_splines_probs );
}

void MCMC_Basis_Fn_Regression::initialise()
{
   ADD_BASIS = DEL_BASIS = MOVE_BASIS = V_SHIFT = SHIFT_COEFFICIENT = -INT_MAX;
   storing_basis_function_coefficients = false;
   previous_m_star = NULL;
   previous_V_star= NULL;
   previous_V_star_sqrt = NULL;
   previous_b_star= NULL;
   if( random_basis_fns )
   {
      ADD_BASIS = add_move( "Add basis fn." );
      DEL_BASIS = add_move( "Delete basis fn." );
      trans_dimensional_move_types.push_back( ADD_BASIS );
      trans_dimensional_move_types.push_back( DEL_BASIS );
      MOVE_BASIS = add_move( "Move basis fn." );
      integer_parameter_pointers.push_back( Basis_Fn::get_Basis_Fn_count_pointer() );
      if( V_inv )
      {
         gsl_matrix_free( V_inv );
         V_inv = 0;
      }
   }
   if( v_shift_move_included )
   {
      V_SHIFT = add_move( "V shift" );
      parameter_pointers.push_back( &v_multiplier );
      variable_v = 1;
   }
   if( storing_basis_function_coefficients )
      SHIFT_COEFFICIENT = add_move( "Shift basis fn. coefficient" );
   number_removable_basis_fns = get_number_of_basis_fns() - number_unremovable_basis_functions;
   num_splines_probs_length = max_num_splines + 1;
   num_splines_probs = new double[ num_splines_probs_length ];
   for( int i = 0; i < num_splines_probs_length; i++ )
      num_splines_probs[ i ] = 0;
   calculate_log_posterior();
   if( !storing_basis_function_coefficients )
      set_basis_coefficients( m_star );
   make_predictions();
   write_predictions();
}

void MCMC_Basis_Fn_Regression::move()
{
   int num_possible_deletions = get_number_of_basis_fn_deletions_possible();
   int num_bases = get_number_of_basis_fns();
   if( move_type == ADD_BASIS )
   {
      if( num_bases == max_num_splines )
      {
         move_type = DEL_BASIS;
         if( max_num_splines == total_num_knot_point_candidates + 1 )
            LogProposalRatio -= LOG_THREE;
         else
            LogProposalRatio -= LOG_TWO;
      }
      else if( num_bases == max_num_splines - 1 )
         if( max_num_splines == total_num_knot_point_candidates + 1 )
            LogProposalRatio += LOG_THREE;
         else
            LogProposalRatio += LOG_TWO;
      else if( num_possible_deletions == 0 )
         LogProposalRatio -= LOG_THREE;
   }
   else if( move_type == DEL_BASIS )
   {
      if( num_possible_deletions == 0 )
      {
         move_type = ADD_BASIS;
         LogProposalRatio -= LOG_THREE;
      }
      else if( num_possible_deletions == 1 )
         LogProposalRatio += LOG_THREE;
      else if( num_bases == max_num_splines )
      {
         if( max_num_splines == total_num_knot_point_candidates + 1 )
            LogProposalRatio -= LOG_THREE;
         else
            LogProposalRatio -= LOG_TWO;
      }
   }
   else if( move_type == MOVE_BASIS )
   {
      if( num_possible_deletions == 0 )
      {
         move_type = ADD_BASIS;
         LogProposalRatio -= LOG_THREE;
      }
      else if( num_bases == total_num_knot_point_candidates + 1 )
      {
         move_type = DEL_BASIS;
         LogProposalRatio -= LOG_THREE;
      }
   }
   propose_move();
}

void MCMC_Basis_Fn_Regression::propose_move()
{
   if( move_type == ADD_BASIS )
   {
      V_star = V_star_sqrt = 0;
      m_star = 0;
      add_random_basis_function();
      update_D_dependent_matrices();
   }
   else if( move_type == DEL_BASIS )
   {
      V_star = V_star_sqrt = 0;
      m_star = 0;
      remove_random_basis_function();
      update_D_dependent_matrices();
   }
   else if( move_type == MOVE_BASIS )
   {
      V_star = V_star_sqrt = 0;
      m_star = 0;
      remove_random_basis_function();
      add_random_basis_function();
      update_D_dependent_matrices();
   }
   else if( move_type == SHIFT_COEFFICIENT )
   {
      if( !rw_on_basis_coeffs )
         draw_random_basis_coefficient_from_prior();
      else
      {
         draw_random_basis_fn();
         rw_shift_coefficient();
      }
   }
   calculate_log_likelihood();
   proposed_log_posterior = proposed_log_likelihood;
}

void MCMC_Basis_Fn_Regression::execute_move()
{
   if( previous_m_star && previous_m_star != m_star )
      gsl_matrix_free( previous_m_star );
   if( previous_V_star && previous_V_star != V_star )
      gsl_matrix_free( previous_V_star );
   if( previous_V_star_sqrt && previous_V_star_sqrt != V_star_sqrt )
      gsl_matrix_free( previous_V_star_sqrt );
   if( previous_b_star && previous_b_star != b_star )
      gsl_matrix_free( previous_b_star );
   
   previous_b_star = b_star;
   previous_m_star = m_star;
   previous_V_star = V_star;
   previous_V_star_sqrt = V_star_sqrt;
   ml_const_store = ml_const;
}

void MCMC_Basis_Fn_Regression::undo_move()
{
   if( move_type == ADD_BASIS )
   {
      remove_basis_function( get_number_of_basis_fns() - 1 );
      update_D_dependent_matrices();
   }
   else if( move_type == DEL_BASIS )
   {
      add_basis_function( Removed_Basis, &removed_basis_position );
      update_D_dependent_matrices();
   }
   else if( move_type == MOVE_BASIS )
   {
      int col = get_number_of_basis_fns();
      if( removed_basis_position >= col )
         col--;
      add_basis_function( Removed_Basis, &removed_basis_position );
      remove_basis_function( get_number_of_basis_fns() - 1 );//col );
      update_D_dependent_matrices();
   }
   else if( move_type == SHIFT_COEFFICIENT )
      Moved_Basis->revert_coeff();

   if( b_star && previous_b_star != b_star )
      gsl_matrix_free( b_star );
   if( m_star && previous_m_star != m_star )
      gsl_matrix_free( m_star );
   if( V_star && previous_V_star != V_star )
      gsl_matrix_free( V_star );
   if( V_star_sqrt && previous_V_star_sqrt != V_star_sqrt )
      gsl_matrix_free( V_star_sqrt );   
   b_star = previous_b_star;
   m_star = previous_m_star;
   V_star = previous_V_star;
   V_star_sqrt = previous_V_star_sqrt;
   ml_const = ml_const_store;
}

void MCMC_Basis_Fn_Regression::post_move_updates()
{
   if( !storing_basis_function_coefficients )
      set_basis_coefficients( m_star );
//      draw_betas_from_posterior();//automatically sets basis coefficients.
}

void MCMC_Basis_Fn_Regression::initialise_ergodic_averages()
{
   if( !ave_predictions_mx && predictions_mx_rows )
   {
      ave_predictions_mx_cols = predictions_update_cols + 1;
      if( variable_to_predict_over != second_variable_to_predict_over )
         ave_predictions_mx_cols++;
      ave_predictions_mx = new double*[ predictions_mx_rows ];
      for( int i = 0; i < predictions_mx_rows; i++ )
      {
         ave_predictions_mx[ i ] = new double[ ave_predictions_mx_cols ];
          for( int j = 0; j < ave_predictions_mx_cols; j++ )
            ave_predictions_mx[ i ][ j ] = 0;
      }
   }
   if( !storing_basis_function_coefficients )
   {
      previous_m_star = m_star;
      previous_V_star = V_star;
      if( !V_star_sqrt )
      {
         V_star_sqrt = gsl_matrix_alloc( V_star->size1, V_star->size2 );
         gsl_matrix_memcpy( V_star_sqrt, V_star );
         gsl_linalg_cholesky_decomp( V_star_sqrt );
         for( int i = 0; i < dim_V; i++ )
            for( int j = i + 1; j < dim_V; j++ )
               gsl_matrix_set( V_star_sqrt, i, j, 0 );
      }
      previous_V_star_sqrt = V_star_sqrt;
      previous_b_star = b_star;
   }
   ml_const_store = ml_const;
   isotonic_count = antitonic_count = monotonic_count = 0;
}

void MCMC_Basis_Fn_Regression::update_ergodic_averages()
{
   test_for_monotonicity();
   if( isotonic )
      isotonic_count++;
   if( antitonic )
      antitonic_count++;
   if( isotonic || antitonic )
      monotonic_count++;
   num_splines_probs[ get_number_of_basis_fns() ]++;
}

void MCMC_Basis_Fn_Regression::ModelPrintOut()
{
   if( past_burn_in() )
   {
      double normaliser = get_current_ergodic_normaliser();
      if( normaliser )
      {
         cout << "P(Isotonic)  = " << isotonic_count / normaliser << endl;
         cout << "P(Antitonic) = " << antitonic_count / normaliser << endl;
         cout << "P(Monotonic) = " << monotonic_count / normaliser << endl;
      }
   }
   cout << "Number of basis functions = " << get_number_of_basis_fns() << endl;
   if( storing_basis_function_coefficients )
      print_basis_fns();
   if( random_basis_fns )
      draw_histogram( num_splines_probs, num_splines_probs_length );
}

void MCMC_Basis_Fn_Regression::make_predictions()
{
   int last_col = 0;
   int i, j;//, shift = predict_on_data ? 0 : regression_n;
   bool extra_deleted_point = predict_on_data && deleted_x_point;
   bool predict_on_knot_point_candidates = !( n > regression_n || predict_on_data );
   double x_val = 0;

   if( !predictions_mx && ( n > regression_n || predict_on_data || predict_on_knot_point_candidates ) )
   {
      if( p == 1 )
         second_variable_to_predict_over = variable_to_predict_over;
      if( log_targets || storing_basis_function_coefficients )
      {
         prediction_fn_value_col = 1;
         predictions_update_cols = 1;
      }
      else
      {
         predictions_update_cols = 3;
         prediction_fn_value_col = 2;
      }
      if( second_variable_to_predict_over != variable_to_predict_over )
      {
         predictions_update_cols++;
         prediction_fn_value_col++;
      }
      if( predict_on_data )
         predictions_mx_rows = n;
      else if( n > regression_n )
         predictions_mx_rows = n - regression_n;
      else
         predictions_mx_rows = total_num_knot_point_candidates;
      if( extra_deleted_point )
         predictions_mx_rows++;
      predictions_mx_cols = predictions_update_cols + 1;
      predictions_mx = new double*[ predictions_mx_rows ];
      for( i = 0; i < predictions_mx_rows; i++ )
         predictions_mx[ i ] = new double[ predictions_mx_cols ];
      for( i = 0; i < predictions_mx_rows; i++ )
      {
         if( extra_deleted_point && i == delete_point )
         {
            x_val = deleted_x_point[ variable_to_predict_over ];
//            shift--;
         }
         else if( !predict_on_knot_point_candidates )
            x_val = test_grid_points[ i ][ variable_to_predict_over ];
//            x_val = gsl_matrix_get( X, shift + i, variable_to_predict_over );
         else
            x_val = knot_point_candidates[ variable_to_predict_over ][ i ];
         if( normalise_predictors && x_scale[ variable_to_predict_over ] )
         {
            x_val *= x_scale[ variable_to_predict_over ];
            if( normalise_predictors == 1 )
               x_val += x_min[ variable_to_predict_over ];
         }
         predictions_mx[ i ][ 0 ] = x_val;
         if( variable_to_predict_over != second_variable_to_predict_over )
         {
            if( extra_deleted_point && i == delete_point )
               x_val = deleted_x_point[ second_variable_to_predict_over ];
            else if( !predict_on_knot_point_candidates )
               x_val = test_grid_points[ i ][ second_variable_to_predict_over ];
//               x_val = gsl_matrix_get( X, shift + i, second_variable_to_predict_over );
            else
               x_val = knot_point_candidates[ i ][ second_variable_to_predict_over ];
            if( normalise_predictors && x_scale[ second_variable_to_predict_over ] )
            {
               x_val *= x_scale[ second_variable_to_predict_over ];
               if( normalise_predictors == 1 )
                  x_val += x_min[ second_variable_to_predict_over ];
            }
            predictions_mx[ i ][ 1 ] = x_val;
         }
      }
//      if( extra_deleted_point )
//         shift++;
   }

   if( !storing_basis_function_coefficients )
   //put the posterior means in for basis coefficients
      set_basis_coefficients( m_star );

   if( predictions_mx )
   {
      last_col = predictions_mx_cols - 1;
      double fn_val = 0, sd_val = 0, *xx = 0;//new double[ p ];
      for( i = 0; i < predictions_mx_rows; i++ )
      {
         if( extra_deleted_point && i == delete_point )
         {
            for( j = 0; j < p; j++ )
               xx = deleted_x_point;
//               xx[ j ] = deleted_x_point[ j ];
//            shift--;
         }
         else
//            for( j = 0; j < p; j++ )
               if( !predict_on_knot_point_candidates )
                  xx = test_grid_points[ i ];
//                  xx[ j ] = gsl_matrix_get( X, shift + i, j );
               else
                  xx = knot_point_candidates[ i ];
//                  xx[ j ] = knot_point_candidates[ i ][ j ];
         fn_val = function_evaluate( xx );
         if( predictions_update_cols == 3 )
            sd_val = sqrt( function_variance_evaluate( xx ) );
         if( normalise_targets && y_scale[ 0 ] > 0 )
         {
            fn_val = ( fn_val + y_bar[ 0 ] ) * y_scale[ 0 ];
            sd_val *= y_scale[ 0 ];
         }
         if( log_targets )
            fn_val = exp( fn_val ) - 1;
         if( predictions_update_cols == 3 )
         {
            predictions_mx[ i ][ last_col - 2 ] = fn_val - 2 * sd_val;
            predictions_mx[ i ][ last_col - 1 ] = fn_val;
            predictions_mx[ i ][ last_col ] = fn_val + 2 * sd_val;
         }
         else
            predictions_mx[ i ][ last_col ] = fn_val;
      }
//      delete [] xx;
   }
   if( !storing_basis_function_coefficients )
      revert_basis_coefficients();
}

void MCMC_Basis_Fn_Regression::rw_shift_coefficient()
{
   double old_coeff = Moved_Basis->get_coeff();
   double new_coeff = old_coeff + basis_coeff_step_size * gsl_ran_flat( rn, -1, 1 );
   Moved_Basis->set_coeff( new_coeff );
   LogProposalRatio += calculate_basis_coefficient_log_prior( new_coeff ) - calculate_basis_coefficient_log_prior( old_coeff );
}
