
#if !defined( _mcmc_density_estimation_h )
#define _mcmc_density_estimation_h

#include "basis_fn_regression.h"
#include<gsl/gsl_cdf.h>
class MCMC_Density_Estimation : public Basis_Fn_Regression
{
  public:
   ///
   MCMC_Density_Estimation( int argc, char **argv );
   ///
   MCMC_Density_Estimation( gsl_matrix* X1, gsl_matrix* Y1 );// uses matrix,matrix constructor of base class
   ///
   MCMC_Density_Estimation( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );// uses matrix,matrix,matrix constructor of base class
   ///
   ~MCMC_Density_Estimation();
   ///
   virtual Regression* create_new_regression( gsl_matrix* X1, gsl_matrix* Y1, gsl_matrix* D1 );//needed for partition regression.

   void initialise();

   virtual void move();

   void propose_move();// Proposes the chosen move type.

   virtual void execute_move();// Carries out the proposed move (used if move accepted).

   virtual void undo_move();// Cancels the proposed move (used if move not accepted).

   virtual void post_move_updates();// Carries out any updates required after the move, accepted or not.

   virtual void revise_log_posterior();

   virtual void initialise_ergodic_averages();//overwritten fn initialising model averages

   virtual void update_ergodic_averages();//overwritten fn updating model averages
   
   virtual void ModelPrintOut();//overwritten fn outputting model details to screen.

   virtual void make_predictions();//calculates the predictive probs of each unmergeable cluster.
   
   virtual double calculate_log_likelihood();

   virtual void update_posterior_maximizing_model();

   void end_of_chain_processing();//overwritten fn updating at end of the MCMC chain.

   double prior_measure( double left_t, double right_t );//the base measure used in the prior. By default Lebesgue.
   
   void write_posterior_maximizing_model();

   void set_basis_coeffs();//set the basis coefficient for calculating the posterior mean surface.
  
  protected:

   int ADD_BASIS, DEL_BASIS, MOVE_BASIS;//constants to be enumerated (dynamically) to indicate the different move types.

   bool random_basis_fns;//if true, have random number and locations of basis fns.

   bool v_shift_move_included;//non zero if the prior variance multiplier v_multiplier can be shifted.
   
   int max_num_cps;//the maximum number of changepoints allowed.

   double* num_cps_probs;//array holding the probabilities of each possible number of cps.

   int num_cps_probs_length;//length of array num_cps_probs.

   int isotonic_count, antitonic_count, monotonic_count;//keep track of the number of samples satisfying these conditions.

   double basis_coeff_step_size;//the size of random walk step used on basis fn coefficients.

   double ml_const, ml_const_store;

   double alpha, log_alpha;//the dirichlet distn parameter, and its log.

   double dirichlet_normaliser;//normalising constant from dirichlet process.

   double beta_distn_alpha, beta_distn_beta;//prior parameters for a Beta distn base measure.

   list<int> partition_point_count, proposed_partition_point_count;//counts of the number of data points inside each set of the partition.

   list<double> partition_volumes, proposed_partition_volumes;//the volumes of each set of the partition.

   list<double> partition_measures, proposed_partition_measures;//the bases measure of each set of the partition.

   list<int> partition_starting_points, proposed_partition_starting_points;//the index of the first data point inside each set of the partition.

   list<double> partition_posterior_means, proposed_partition_posterior_means;//the posterior mean probabilties of being in each set of the partition.

   list<double> partition_mls, proposed_partition_mls;//the marginal likelihoods of each set of the partition.

   list<double> best_changepoints;//the locations of the most probable changepoints visited.

   list<int> best_components;//the components of the corresponding best_changepoints.

   list<double> best_changepoint_fn_vals;//the fn values at each of the best changepoints.

   list<bool> best_changepoint_unremovable;//the unremovable status of each of the best changepoints.
   
   char Best_Changepoints_Filename[FILENAMELENGTH];//file for writing final order of subjects after hierarchical clustering.

   char Input_Changepoints_Filename[FILENAMELENGTH];//file which initial partition is read in from. Each row is a changepoint.
};

#endif

