#include "spline_basis_fn.h"

Spline_Basis_Fn::Spline_Basis_Fn()
      : Basis_Fn()
{
   basis_type = SPLINE_BASIS;
}

Spline_Basis_Fn::Spline_Basis_Fn( double* v, int terms, int* cmp, int* pw, int* sgn, int* actv_dim )
      : Basis_Fn( v, terms, cmp, pw, sgn, actv_dim )
{
   basis_type = SPLINE_BASIS;
}

Spline_Basis_Fn::Spline_Basis_Fn( double d, int cmp, int pw, int sgn, int* actv_dim )
      : Basis_Fn( d, cmp, pw, sgn, actv_dim )
{
   basis_type = SPLINE_BASIS;
}

Spline_Basis_Fn::~Spline_Basis_Fn()
{
}

double Spline_Basis_Fn::evaluate( double* x, int* actv_dim, int order_integral, int integral_component )
{
   if( actv_dim && active_dimension && *active_dimension != *actv_dim )
      return 0;
   double x_i, x_val = 1, x_mult;
   bool positive = true, component_matched = false;
   int i = 0;
   int power_i;
   while( positive && i < num_interactions )
   {
      x_i = (x[ component[ i ] ] - t[ i ])/scale;
      if( power[ i ] < 0 )
      {
         if( x_i )
         {
            x_val = 0;
            positive = false;
         }
      }
      else
      {
         if( !sign[ i ] || ( sign[ i ] > 0 &&  x_i >= 0 ) || ( sign[ i ] < 0 &&  x_i < 0 )  )
         {
            if( !order_integral || component[ i ] == integral_component )
            {
               component_matched = true;
               power_i = power[ i ];
               int j = 0;
               x_mult = 1;
               if( order_integral > 0 )
               {
                  while( j < order_integral )
                  {
                     j++;
                     x_mult /= power_i + j;
                  }
               }
               else if( order_integral < 0 )
               {
                  if( order_integral + power_i < 0 )
                     x_mult = 0;
                  else
                  {
                     while( j > order_integral )
                     {
                        x_mult *= power_i - j;
                        j--;
                     }
                  }
               }
               x_val *= x_mult * pow( fabs( x_i ), (double)power_i + order_integral );
            }
         }
         else
         {
            x_val = 0;
            positive = false;
         }
      }
      i++;
   }
   if( order_integral && !component_matched )
   {
      if( order_integral > 0 )
      {
         x_val *= pow( x[ integral_component ], order_integral );
         x_val /= gsl_sf_fact( order_integral );
      }
      else
         x_val = 0;
   }
   double mean_mult = 1;
   if( order_integral > 0 )
      mean_mult = pow( x[ integral_component ], order_integral ) / gsl_sf_fact( order_integral );
   return ( x_val - mean * mean_mult ) / sd;
}
